﻿------------------------------------------------------------------
--FCOItemSaver.lua
--Author: ingeniousclown, much modified by Baertram
--v0.4.5
--[[
Allows you to mark an item so you can know that you meant to save it for some reason.
-> Several icons for free usage
Proposal:
-Research icon to mark items for upcoming research
-Tank/DD/Heal icon to show currently equipped items (up to 5 different gears)
-Red lock to show locked items (do not sell them)
-Yellow coins icon for stuff you want to sell
Including filters on/off/show only marked items inside inventories, crafting stations, banks, guild banks, guild stores, player 2 player trading and sending mail
]]
------------------------------------------------------------------
-- Create the filter object for addon libFilters
local libFilters = LibStub("libFilters")
local LAM 		 = LibStub('LibAddonMenu-2.0')

--Array for all the variables
local locVars = {}

--Uncolored "FCOIS" pre chat text for the chat output
locVars.preChatText = "FCOIS"
--Green colored "FCOIS" pre text for the chat output
locVars.preChatTextGreen = "|c22DD22"..locVars.preChatText.."|r: "
--Red colored "FCOIS" pre text for the chat output
locVars.preChatTextRed = "|cDD2222"..locVars.preChatText.."|r: "
--Blue colored "FCOIS" pre text for the chat output
locVars.preChatTextBlue = "|c2222DD"..locVars.preChatText.."|r: "

--Constants for the inventory types, libFilter values
INV_BAGS 							= 1
INV_BANK 							= 2
INV_GUILDBANK 			   			= 3
INV_STORE 							= 4
INV_DECONSTRUCTION  				= 5
INV_GUILDSTORE 	 		   			= 6
INV_MAIL 							= 7
INV_TRADE 				   			= 8
INV_ENCHANTING_CREATION				= 11
INV_ENCHANTING_EXTRACTION	 		= 12
INV_IMPROVEMENT						= 13
--Count of available inventory filter types (INV_BAGS, INV_BANK, etc. -> see above)
locVars.gFCONumFilterInventoryTypes = 13

--The mapping array for INV to inventory type
local InvToInventoryType = {
	[INV_BAGS] 					= INVENTORY_BACKPACK,
	[INV_BANK] 					= INVENTORY_BANK,
	[INV_GUILDBANK] 		    = INVENTORY_GUILD_BANK,
	[INV_STORE] 				= INVENTORY_BACKPACK,
	[INV_DECONSTRUCTION]  		= INVENTORY_BACKPACK,
	[INV_GUILDSTORE] 	 		= INVENTORY_BACKPACK,
	[INV_MAIL] 					= INVENTORY_BACKPACK,
	[INV_TRADE] 				= INVENTORY_BACKPACK,
	[INV_ENCHANTING_CREATION]	= INVENTORY_BACKPACK,
	[INV_ENCHANTING_EXTRACTION]	= INVENTORY_BACKPACK,
	[INV_IMPROVEMENT]			= INVENTORY_BACKPACK,
}

--Global variable to tell where the filtering is currently needed (Inventory, Bank, Crafting Station, Guild Bank, Guild Store, Mail, Trading, Vendor, Enchanting table)
locVars.gFilterWhere		   		= INV_BAGS
--Global variable to tell which filter was clicked/used by chat command at last
locVars.gLastFilterId               = 1
--variable to override the changed "split filters" settings at function UnregisterFilters()
locVars.gSplitFilterOverride		= false

--Global value: Number of filters
locVars.gFCONumFilters			= 4
--Global value: Number of filter icons to choose by right click menu
locVars.gFCONumFilterIcons		= 8
--Global value: Number of filter panels, where the filters can be active
--As the panels for inventory, bank (deposit), guild bank (deposit), mail, trade
--and guild store (select items from inventory for selling) are the same,
--the number here differes from locVars.gFCONumFilterInventoryTypes!
--Different panels are:
--1. Inventories (all)
--2. Deconstruction
--3. Vendor
--4. Guild bank
--5. Bank
--6. Enchanting station extraction & creation
--7. Improvement
--8. Guild store
--9. Mail
--10. Trade
locVars.gFCONumFilterPanels		= 7

--Width and height for the icons
locVars.gIconWidth				= 32
locVars.gIconHeight             = 32
locVars.gEquipmentIconWidth 	= 20
locVars.gEquipmentIconHeight	= 20
--Width and height for the filter buttons
locVars.gFilterButtonWidth		= 24
locVars.gFilterButtonHeight     = 24
--Left and top of the filter buttons
locVars.gFilterButtonTop		= 6
locVars.gFilterButtonLeft	   	= {
	12,
    38,
    64,
    90
}
--Filter button offset on x axis, if InventoryGriView addon is active too
locVars.gGridViewOffsetX			= 26

--Addon variables
local addonVars = {}
addonVars.gAddonName				= "FCOItemSaver"
addonVars.addonNameMenu				= "FCO ItemSaver"
addonVars.addonVersion		   		= 0.10 -- Changing this will reset SavedVariables!
addonVars.addonVersionOptions 		= '0.4.4' -- version shown in the settings panel
addonVars.gAddonLoaded				= false

-- For the test, if the Addon "Inventory Gridview" is enabled
local otherAddons = {}
otherAddons.GRIDVIEWBUTTON    		= "ZO_PlayerInventory_GridButton"
otherAddons.inventoryGridViewActive = false
--For the test, if the addon "Chat Merchant" is enabled
otherAddons.CHATMERCHANTBUTTON 		= "ZO_PlayerInventory_CMbutton"
otherAddons.chatMerchantActive 		= false
--For the test, if the addon "Research Assistant" is enabled
otherAddons.researchAssistantActive	= false

--Control names of ZO* standrd controls etc.
local INV						= ZO_PlayerInventory
local INV_NAME					= INV:GetName()
local BACKPACK 		    		= ZO_PlayerInventoryBackpack
local BACKPACK_BAG 				= ZO_PlayerInventoryBackpackContents
local VENDOR_SELL				= ZO_StoreWindowListSellToVendorArea
local VENDOR_SELL_NAME			= VENDOR_SELL:GetName()
local VENDOR_MENUBAR_BUTTON_SELL = ZO_StoreWindowMenuBarButton2
local BANK_INV					= ZO_PlayerBank
local BANK_INV_NAME				= BANK_INV:GetName()
local BANK			    		= ZO_PlayerBankBackpack
local BANK_BAG		    		= ZO_PlayerBankBackpackContents
local BANK_MENUBAR_BUTTON_WITHDRAW	= ZO_PlayerBankMenuBarButton1
local BANK_MENUBAR_BUTTON_DEPOSIT = ZO_PlayerBankMenuBarButton2
local GUILD_BANK_INV 	    	= ZO_GuildBank
local GUILD_BANK_INV_NAME		= GUILD_BANK_INV:GetName()
local GUILD_BANK 	    		= ZO_GuildBankBackpack
local GUILD_BANK_BAG    		= ZO_GuildBankBackpackContents
local GUILD_BANK_MENUBAR_BUTTON_WITHDRAW	= ZO_GuildBankMenuBarButton1
local GUILD_BANK_MENUBAR_BUTTON_DEPOSIT = ZO_GuildBankMenuBarButton2
local GUILD_STORE				= ZO_TradingHouse
--local GUILD_STORE_MENUBAR_BUTTON_SEARCH = ZO_TradingHouseMenuBarButton1
--local GUILD_STORE_MENUBAR_BUTTON_SELL = ZO_TradingHouseMenuBarButton2
--local GUILD_STORE_MENUBAR_BUTTON_LIST = ZO_TradingHouseMenuBarButton3
--local SMITHING_TOP_LEVEL		= ZO_SmithingTopLevel
local DECONSTRUCTION_INV		= ZO_SmithingTopLevelDeconstructionPanelInventory
local DECONSTRUCTION_INV_NAME	= DECONSTRUCTION_INV:GetName()
local DECONSTRUCTION    		= ZO_SmithingTopLevelDeconstructionPanelInventoryBackpack
local DECONSTRUCTION_BAG 		= ZO_SmithingTopLevelDeconstructionPanelInventoryBackpackContents
local DECONSTRUCTION_SLOT 		= ZO_SmithingTopLevelDeconstructionPanelSlotContainerExtractionSlot
--local SMITHING_MENUBAR_BUTTON_DECONSTRUCTION 		= ZO_SmithingTopLevelModeMenuBarButton3
local IMPROVEMENT_INV  			= ZO_SmithingTopLevelImprovementPanelInventory
local IMPROVEMENT_INV_NAME		= IMPROVEMENT_INV:GetName()
local IMPROVEMENT    			= ZO_SmithingTopLevelImprovementPanelInventoryBackpack
local IMPROVEMENT_BAG  			= ZO_SmithingTopLevelImprovementPanelInventoryBackpackContents
local IMPROVEMENT_BOOSTER_CONTAINER = ZO_SmithingTopLevelImprovementPanelBoosterContainer
--local SMITHING_MENUBAR_BUTTON_IMPROVEMENT 			= ZO_SmithingTopLevelModeMenuBarButton4
local LIST_DIALOG 	    		= ZO_ListDialog1List
local MAIL_SEND					= ZO_MailSend
local MAIL_SEND_NAME			= MAIL_SEND:GetName()
--local MAIL_INBOX				= ZO_MailInbox
--local MAIL_ATTACHMENTS			= ZO_MailSendAttachments
--local MAIL_MENUBAR_BUTTON_SEND  = ZO_MainMenuSceneGroupBarButton2
local PLAYER_TRADE				= ZO_Trade
local PLAYER_TRADE_NAME			= PLAYER_TRADE:GetName()
local ENCHANTING_STATION		= ZO_EnchantingTopLevelInventoryBackpack
local ENCHANTING_STATION_NAME	= ENCHANTING_STATION:GetName()
local ENCHANTING_STATION_BAG	= ZO_EnchantingTopLevelInventoryBackpackContents
--local ENCHANTING_STATION_MENUBAR_BUTTON_CREATION    = ZO_EnchantingTopLevelModeMenuBarButton1
--local ENCHANTING_STATION_MENUBAR_BUTTON_EXTRACTION  = ZO_EnchantingTopLevelModeMenuBarButton2
local QUICKSLOT_CIRCLE  		= ZO_QuickSlotCircle
local DestroyItemDialog    		= ESO_Dialogs["DESTROY_ITEM_PROMPT"]
local CHARACTER					= ZO_Character

--The last clicked enchanting panel top menu bar button
--locVars.gLastEnchantingButton	= ENCHANTING_STATION_MENUBAR_BUTTON_CREATION
--locVars.gLastSmithingButton		= SMITHING_MENUBAR_BUTTON_DECONSTRUCTION
locVars.gLastBankButton			= BANK_MENUBAR_BUTTON_WITHDRAW
locVars.gLastGuildBankButton	= GUILD_BANK_MENUBAR_BUTTON_WITHDRAW
--locVars.gLastGuildStoreButton	= GUILD_STORE_MENUBAR_BUTTON_SEARCH
--locVars.gLastVendorButton		= VENDOR_MENUBAR_BUTTON_SELL
--locVars.gLastMailButton         = MAIL_MENUBAR_BUTTON_SEND

--The array for the texture names of each panel Id
local gFilterPanelIdToTextureName = {
	[INV_BAGS] = INV_NAME .. "_FilterButton%sTexture",
	[INV_DECONSTRUCTION] = DECONSTRUCTION_INV_NAME .. "_FilterButton%sTexture",
	[INV_IMPROVEMENT] = IMPROVEMENT_INV_NAME .. "_FilterButton%sTexture",
	[INV_STORE] = INV_NAME .. "_FilterButton%sTexture",
	[INV_GUILDBANK] = GUILD_BANK_INV_NAME .. "_FilterButton%sTexture",
	[INV_GUILDSTORE] = INV_NAME .. "_FilterButton%sTexture",
	[INV_BANK] = BANK_INV_NAME .. "_FilterButton%sTexture",
	[INV_ENCHANTING_EXTRACTION] = ENCHANTING_STATION_NAME .. "_FilterButton%sTexture",
	[INV_ENCHANTING_CREATION] = ENCHANTING_STATION_NAME .. "_FilterButton%sTexture",
	[INV_MAIL] = INV_NAME .. "_FilterButton%sTexture",
	[INV_TRADE] = INV_NAME .. "_FilterButton%sTexture",
}

--The icons to choose from
local MARKER_TEXTURES			= {
	[1]  = [[/esoui/art/campaign/campaignbrowser_fullpop.dds]],
	[2]  = [[/esoui/art/inventory/inventory_tabicon_armor_disabled.dds]],
	[3]  = [[/esoui/art/crafting/smithing_tabicon_research_disabled.dds]],
	[4]  = [[/esoui/art/tradinghouse/tradinghouse_sell_tabicon_disabled.dds]],
	[5]  = [[/esoui/art/campaign/overview_indexicon_bonus_disabled.dds]],
	[6]  = [[/esoui/art/ava/tabicon_bg_score_disabled.dds]],
	[7]  = [[/esoui/art/guild/guild_rankicon_leader_large.dds]],
	[8]  = [[/esoui/art/lfg/lfg_healer_up.dds]],
	[9]  = [[/esoui/art/mounts/timer_icon.dds]],
    [10] = [[/esoui/art/crafting/alchemy_tabicon_solvent_up.dds]],
	[11] = [[/esoui/art/buttons/cancel_up.dds]],
    [12] = [[/esoui/art/buttons/info_up.dds]],
    [13] = [[/esoui/art/buttons/pinned_normal.dds]],
    [14] = [[/esoui/art/cadwell/cadwell_indexicon_gold_up.dds]],
    [15] = [[/esoui/art/cadwell/cadwell_indexicon_silver_up.dds]],
    [16] = [[/esoui/art/campaign/campaignbonus_keepicon.dds]],
    [17] = [[/esoui/art/campaign/campaignbonus_scrollicon.dds]],
    [18] = [[/esoui/art/campaign/campaignbrowser_columnheader_ad.dds]],
    [19] = [[/esoui/art/campaign/campaignbrowser_columnheader_dc.dds]],
    [20] = [[/esoui/art/campaign/campaignbrowser_columnheader_ep.dds]],
    [21] = [[/esoui/art/campaign/campaignbrowser_guild.dds]],
    [22] = [[/esoui/art/campaign/campaignbrowser_indexicon_normal_up.dds]],
    [23] = [[/esoui/art/campaign/overview_indexicon_scoring_up.dds]],
    [24] = [[/esoui/art/charactercreate/charactercreate_bodyicon_up.dds]],
    [25] = [[/esoui/art/characterwindow/gearslot_offhand.dds]],
    [26] = [[/esoui/art/characterwindow/gearslot_mainhand.dds]],
    [27] = [[/esoui/art/characterwindow/gearslot_costume.dds]],
    [28] = [[/esoui/art/chatwindow/chat_mail_up.dds]],
    [29] = [[/esoui/art/chatwindow/chat_notification_up.dds]],
    [30] = [[/esoui/art/crafting/alchemy_tabicon_reagent_up.dds]],
    [31] = [[/esoui/art/crafting/smithing_tabicon_refine_up.dds]],
    [32] = [[/esoui/art/deathrecap/deathrecap_killingblow_icon.dds]],
    [33] = [[/esoui/art/fishing/bait_emptyslot.dds]],
    [34] = [[/esoui/art/guild/guildhistory_indexicon_guildbank_up.dds]],
    [35] = [[/esoui/art/guild/guild_indexicon_member_up.dds]],
    [36] = [[/esoui/art/guild/tabicon_roster_up.dds]],
    [37] = [[/esoui/art/icons/poi/poi_dungeon_complete.dds]],
    [38] = [[/esoui/art/icons/poi/poi_groupinstance_complete.dds]],
    [39] = [[/esoui/art/icons/servicemappins/servicepin_magesguild.dds]],
    [40] = [[/esoui/art/icons/servicemappins/servicepin_fightersguild.dds]],
	[41] = [[/esoui/art/lfg/lfg_dps_up.dds]],
	[42] = [[/esoui/art/lfg/lfg_leader_icon.dds]],
	[43] = [[/esoui/art/lfg/lfg_tank_up.dds]],
	[44] = [[/esoui/art/lfg/lfg_veterandungeon_up.dds]],
	[45] = [[/esoui/art/lfg/lfg_normaldungeon_up.dds]],
	[46] = [[/esoui/art/progression/icon_dualwield.dds]],
	[47] = [[/esoui/art/progression/icon_firestaff.dds]],
	[48] = [[/esoui/art/progression/icon_bows.dds]],
	[49] = [[/esoui/art/progression/icon_2handed.dds]],
	[50] = [[/esoui/art/progression/icon_1handed.dds]],
	[51] = [[/esoui/art/progression/progression_tabicon_backup_inactive.dds]],
	[52] = [[/esoui/art/repair/inventory_tabicon_repair_disabled.dds]],
	[53] = [[/esoui/art/worldmap/selectedquesthighlight.dds]],
	[54] = [[/esoui/art/journal/journal_tabicon_cadwell_up.dds]],
}

--Arrays for the right click/context menu entries
local lTextMark 				= {}
local lTextDemark 				= {}
local contextEntries 			= {}

local settings 					= {}
local defaultSettings			= {}

local markedItems 				= {}
for i = 1, locVars.gFCONumFilters, 1 do
	markedItems[i] = {}
end

local preventerVars = {}
preventerVars.gScanningInv	    	= false
--preventerVars.canUpdateInv 	   		= true
preventerVars.gFilteringBasics		= false
preventerVars.gActiveFilterPanel	= false
preventerVars.gNoCloseEvent 		= false
preventerVars.gAllowDestroyItem		= false
preventerVars.gItemSlotIsLocked 	= false
preventerVars.gCheckEquipmentSlots  = false
preventerVars.gUpdateMarkersNow		= false
preventerVars.gChangedGears			= false
preventerVars.gContextCreated		= {}

local eventHandlers 		= {}

--The options panel of FCO ItemSaver
local FCOSettingsPanel

--Table to map the settings filter state to the output text identifier
local settingsFilterStateToText = {
	["true"]  = "on",
    ["false"] = "off",
    ["-99"]   = "onlyfiltered",
}

--Table to map iconId to relating filterId
local iconIdToFilter = {
	[1]	= 1,
	[2]	= 2,
	[3]	= 3,
	[4]	= 2,
	[5]	= 4,
	[6]	= 2,
	[7]	= 2,
	[8]	= 2,
}

--Table to map iconId to gearId
local iconToGear = {
	[2] = 1,
	[4] = 2,
	[6] = 3,
	[7] = 4,
	[8] = 5,
}

--Table with the weapon types for the main&offhand checks
local weaponTypeCheckTable = {
   ["1hd"] = {
	[WEAPONTYPE_AXE] = true,
	[WEAPONTYPE_DAGGER] = true,
	[WEAPONTYPE_HAMMER] = true,
	[WEAPONTYPE_SWORD] = true,
   },
   ["2hd"] = {
	[WEAPONTYPE_TWO_HANDED_AXE] = true,
	[WEAPONTYPE_TWO_HANDED_HAMMER] = true,
	[WEAPONTYPE_TWO_HANDED_SWORD] = true,
   },
   ["2hdall"] = {
	[WEAPONTYPE_TWO_HANDED_AXE] = true,
	[WEAPONTYPE_TWO_HANDED_HAMMER] = true,
	[WEAPONTYPE_TWO_HANDED_SWORD] = true,
	[WEAPONTYPE_BOW] = true,
	[WEAPONTYPE_FIRE_STAFF] = true,
	[WEAPONTYPE_LIGHTNING_STAFF] = true,
	[WEAPONTYPE_FROST_STAFF] = true,
	[WEAPONTYPE_HEALING_STAFF] = true,
   },
   ["staff"] = {
	[WEAPONTYPE_FIRE_STAFF] = true,
	[WEAPONTYPE_LIGHTNING_STAFF] = true,
	[WEAPONTYPE_FROST_STAFF] = true,
	[WEAPONTYPE_HEALING_STAFF] = true,
   },
   ["sword"] = {
	[WEAPONTYPE_SWORD] = true,
   },
   ["swordall"] = {
	[WEAPONTYPE_SWORD] = true,
	[WEAPONTYPE_TWO_HANDED_SWORD] = true,
   },
   ["axe"] = {
	[WEAPONTYPE_AXE] = true,
   },
   ["axeall"] = {
	[WEAPONTYPE_AXE] = true,
	[WEAPONTYPE_TWO_HANDED_AXE] = true,
   },
   ["dagger"] = {
	[WEAPONTYPE_DAGGER] = true,
   },
   ["hammer"] = {
	[WEAPONTYPE_HAMMER] = true,
   },
   ["hammerall"] = {
	[WEAPONTYPE_HAMMER] = true,
	[WEAPONTYPE_TWO_HANDED_HAMMER] = true,
   },
   ["bow"] = {
	[WEAPONTYPE_BOW] = true,
   },
   ["shield"] = {
	[WEAPONTYPE_SHIELD] = true,
   },
   ["firestaff"] = {
	[WEAPONTYPE_FIRE_STAFF] = true,
   },
   ["lightningstaff"] = {
	[WEAPONTYPE_LIGHTNING_STAFF] = true,
   },
   ["froststaff"] = {
	[WEAPONTYPE_FROST_STAFF] = true,
   },
   ["healingstaff"] = {
	[WEAPONTYPE_HEALING_STAFF] = true,
   },
   ["rune"] = {
	[WEAPONTYPE_RUNE] = true,
   },
   }
--Table with allowed item traits for ornate items
local allowedOrnateItemTraits = {
	[ITEM_TRAIT_TYPE_ARMOR_ORNATE]   = true,
	[ITEM_TRAIT_TYPE_JEWELRY_ORNATE] = true,
	[ITEM_TRAIT_TYPE_WEAPON_ORNATE]  = true,
}
--Table with allowed item types for researching
   local allowedResearchableItemTypes = {
   	[ITEMTYPE_ARMOR]	= true,
    [ITEMTYPE_WEAPON]   = true,
   }
--Table with NOT allowed parent control names. These cannot use the new FCOItemSaver context menu entries
local notAllowedContextMenuParentControls = {
	["ZO_StoreWindowListContents"] = true,
	["ZO_BuyBackListContents"] = true,
	["ZO_PlayerInventoryQuestContents"] = true,
	["ZO_SmithingTopLevelImprovementPanel"] = true,
    ["ZO_SmithingTopLevelResearchPanelResearchLineListList"] = true,
	["ZO_SmithingTopLevelCreationPanelPatternListList"] = true,
	["ZO_SmithingTopLevelCreationPanelMaterialListList"] = true,
	["ZO_SmithingTopLevelCreationPanelStyleListList"] = true,
	["ZO_SmithingTopLevelCreationPanelTraitListList"] = true,
	["ZO_AlchemyTopLevelInventoryBackpackContents"] = true,
	["ZO_MailInboxMessage"] = true,
	["ZO_MailSend"] = true,
    ["ZO_TradingHouseItemPaneSearchResultsContents"] = true,
    ["ZO_TradingHousePostedItemsList"] = true,
--		["ZO_QuickSlotListContents"] = true,
}
--Table with NOT allowed control names. These cannot use the new FCOItemSaver context menu entries
local notAllowedContextMenuControls = {
	["ZO_SmithingTopLevelRefinementPanelSlotContainer"] = true,
	["ZO_SmithingTopLevelDeconstructionPanelSlotContainer"] = true,
	["ZO_AlchemyTopLevelSlotContainer"] = true,
    ["ZO_EnchantingTopLevelRuneSlotContainer"] = true,
    ["ZO_EnchantingTopLevelExtractionSlotContainer"] = true,
	["ZO_ApplyEnchantPanel"] = true,
	["ZO_SoulGemItemChargerPanel"] = true,
}
--Table with allowed control names for the character equipment weapon and offhand weapon slots
local allowedCharacterEquipmentWeaponControlNames = {
	['ZO_CharacterEquipmentSlotsMainHand'] = true,
	['ZO_CharacterEquipmentSlotsOffHand'] = true,
	['ZO_CharacterEquipmentSlotsBackupMain'] = true,
	['ZO_CharacterEquipmentSlotsBackupOff'] = true,
   }
--Table with allowed control names for the character equipment jewelry
local allowedCharacterEquipmentJewelryControlNames = {
	['ZO_CharacterEquipmentSlotsNeck'] = true,
	['ZO_CharacterEquipmentSlotsRing1'] = true,
	['ZO_CharacterEquipmentSlotsRing2'] = true,
   }
--Table with weapon and jewelry slot names for equipment checks
local equipmentSlotsNames = {
    ["no_auto_mark"] = {
		["ZO_CharacterEquipmentSlotsCostume"] = true,
    }
}
--Table with all equipment slot names which can be updated with markes for the icons
--The index is the relating slotIndex of the bag BAG_WORN!
local characterEquipmentSlotNameByIndex = {
	[0] = "ZO_CharacterEquipmentSlotsHead",
	[1] = "ZO_CharacterEquipmentSlotsNeck",
	[2] = "ZO_CharacterEquipmentSlotsChest",
	[3] = "ZO_CharacterEquipmentSlotsShoulder",
	[4] = "ZO_CharacterEquipmentSlotsMainHand",
	[5] = "ZO_CharacterEquipmentSlotsOffHand",
	[6] = "ZO_CharacterEquipmentSlotsBelt",
	[8] = "ZO_CharacterEquipmentSlotsLeg",
	[9] = "ZO_CharacterEquipmentSlotsFoot",
	[10] = "ZO_CharacterEquipmentSlotsCostume",
	[11] = "ZO_CharacterEquipmentSlotsRing1",
	[12] = "ZO_CharacterEquipmentSlotsRing2",
	[16] = "ZO_CharacterEquipmentSlotsGlove",
	[20] = "ZO_CharacterEquipmentSlotsBackupMain",
	[21] = "ZO_CharacterEquipmentSlotsBackupOff",
}

--libFilter values
local inventoryUpdaters = {
	BACKPACK = function()
		PLAYER_INVENTORY:UpdateList(INVENTORY_BACKPACK)
	end,
	BANK = function()
		PLAYER_INVENTORY:UpdateList(INVENTORY_BANK)
	end,
	GUILD_BANK = function()
		PLAYER_INVENTORY:UpdateList(INVENTORY_GUILD_BANK)
	end,
	DECONSTRUCTION = function()
		SMITHING.deconstructionPanel.inventory:HandleDirtyEvent()
	end,
	IMPROVEMENT = function()
		SMITHING.improvementPanel.inventory:HandleDirtyEvent()
	end,
	ENCHANTING = function()
		ENCHANTING.inventory:HandleDirtyEvent()
	end,
}

local filterTypeToUpdaterName = {
	[LAF_BAGS] = "BACKPACK",
	[LAF_BANK] = "BANK",
	[LAF_GUILDBANK] = "GUILD_BANK",
	[LAF_STORE] = "BACKPACK",
	[LAF_DECONSTRUCTION] = "DECONSTRUCTION",
	[LAF_GUILDSTORE] = "BACKPACK",
	[LAF_MAIL] = "BACKPACK",
	[LAF_TRADE] = "BACKPACK",
	[LAF_ENCHANTING_CREATION] = "ENCHANTING",
	[LAF_ENCHANTING_EXTRACTION] = "ENCHANTING",
	[LAF_IMPROVEMENT] = "IMPROVEMENT",
}

--The context menu for mark/unmark all
local contextMenu 	 = {}
local bdSelectedLine = {}
for i=1, locVars.gFCONumFilterInventoryTypes, 1 do
	preventerVars.gContextCreated[i] = false
    contextMenu[i]					 = nil
    bdSelectedLine[i]				 = nil
end

--The textures for the button context menu, selected item
local menuInfo =
{
	[MENU_TYPE_DEFAULT] =
    {
        backdropEdge       = "EsoUI/Art/miscellaneous/insethighlight_edge.dds",
        backdropCenter     = "EsoUI/Art/miscellaneous/insethighlight_center.dds",
        backdropInsets     = {16,16,-16,-16},
        backdropEdgeWidth  = 128,
        backdropEdgeHeight = 16,
    }
}

--The mapping table for context menu buttons to icon id
local buttonContextMenuToIconId = {
	["ButtonContextMenu1"] = {
	    iconId = 1,
        mark   = true,
    },
	["ButtonContextMenu2"] = {
	    iconId = 1,
        mark   = false,
    },
	["ButtonContextMenu3"] = {
	    iconId = 2,
        mark   = true,
    },
	["ButtonContextMenu4"] = {
	    iconId = 2,
        mark   = false,
    },
	["ButtonContextMenu5"] = {
	    iconId = 4,
        mark   = true,
    },
	["ButtonContextMenu6"] = {
	    iconId = 4,
        mark   = false,
    },
	["ButtonContextMenu7"] = {
	    iconId = 6,
        mark   = true,
    },
	["ButtonContextMenu8"] = {
	    iconId = 6,
        mark   = false,
    },
	["ButtonContextMenu9"] = {
	    iconId = 7,
        mark   = true,
    },
	["ButtonContextMenu10"] = {
	    iconId = 7,
        mark   = false,
    },
	["ButtonContextMenu11"] = {
	    iconId = 8,
        mark   = true,
    },
	["ButtonContextMenu12"] = {
	    iconId = 8,
        mark   = false,
    },
	["ButtonContextMenu13"] = {
	    iconId = 3,
        mark   = true,
    },
	["ButtonContextMenu14"] = {
	    iconId = 3,
        mark   = false,
    },
	["ButtonContextMenu15"] = {
	    iconId = 5,
        mark   = true,
    },
	["ButtonContextMenu16"] = {
	    iconId = 5,
        mark   = false,
    },
}

--The table with the undo entries for last changes by context menu
undoMarkedItems = {}

--===================== FUNCTIONS ==============================================

--Output debug message in chat
local function debugMessage(msg_text, deep)
	if (deep and not settings.deepDebug) then
    	return
    end
	if (settings.debug == true) then
    	if deep then
        	--Blue colored "FCOIS" at the start of the string
	        d(locVars.preChatTextBlue .. msg_text)
        else
        	--Green colored "FCOIS" at the start of the string
	        d(locVars.preChatTextGreen .. msg_text)
        end
	end
end

--==============================================================================
--			Pre-Hook Handler Methods
--==============================================================================
-- gets handler from the event handler list
local function GetEventHandler(eventName, objName)
	if( not eventHandlers or not eventHandlers[eventName] ) then return nil end
	return eventHandlers[eventName][objName]
end

-- adds handler to the event handler list
local function SetEventHandler(eventName, objName, handler)
	if ( not eventHandlers[eventName] ) then eventHandlers[eventName] = {} end
	eventHandlers[eventName][objName] = handler
end

-- puts given handler in front of the event handler of given object
local function PreHookHandler(eventName, control, handler)
	SetEventHandler(eventName, control:GetName(), control:GetHandler(eventName) )
	control:SetHandler(eventName, handler)
end


--Map filterPanelId digit to global constant value
local function mapFilterPanelId(p_filterPanel)
 if tonumber(p_filterPanel) < 1 or tonumber(p_filterPanel) > locVars.gFCONumFilterInventoryTypes then
 	return INV_BAGS
 else
 	if     tonumber(p_filterPanel) == 1 then
    	return INV_BAGS
    elseif tonumber(p_filterPanel) == 2 then
    	return INV_BANK
    elseif tonumber(p_filterPanel) == 3 then
    	return INV_GUILDBANK
    elseif tonumber(p_filterPanel) == 4 then
    	return INV_STORE
    elseif tonumber(p_filterPanel) == 5 then
    	return INV_DECONSTRUCTION
    elseif tonumber(p_filterPanel) == 6 then
    	return INV_GUILDSTORE
    elseif tonumber(p_filterPanel) == 7 then
    	return INV_MAIL
    elseif tonumber(p_filterPanel) == 8 then
    	return INV_TRADE
    elseif tonumber(p_filterPanel) == 9 then
    	return INV_ENCHANTING_EXTRACTION
    elseif tonumber(p_filterPanel) == 10 then
    	return INV_IMPROVEMENT
    elseif tonumber(p_filterPanel) == 11 then
    	return INV_ENCHANTING_CREATION
	end
 end
end

--Choose the filter type
local function getSettingsIsFilterOn(p_filterId, p_filterPanel)
    if (settings.splitFilters == true) then
        local result
		local p_filterPanelNew = p_filterPanel or locVars.gFilterWhere

		--New behaviour with filters
        result = settings.isFilterPanelOn[p_filterPanelNew][p_filterId]
        --debugMessage("[GetSettingsIsFilterOn] Filter Panel: " .. tostring(p_filterPanelNew) .. ", FilterId: " .. tostring(p_filterId) .. ", Result: " .. tostring(result), true)
        return result
	else
		--Old behaviour with filters
		result = settings.isFilterOn[p_filterId]
        debugMessage("[GetSettingsIsFilterOn] FilterId: " .. tostring(p_filterId) .. ", Result: " .. tostring(result), true)
		return result
    end
end

--Set the value of a filter type, and return it
local function setSettingsIsFilterOn(p_filterId, p_value, p_filterPanel)
    local p_filterPanelNew = p_filterPanel or locVars.gFilterWhere
    if (settings.splitFilters == true) then
		--New behaviour with filters
        settings.isFilterPanelOn[p_filterPanelNew][p_filterId] = p_value
        debugMessage("[SetSettingsIsFilterOn] Filter Panel: " .. tostring(p_filterPanelNew) .. ", FilterId: " .. tostring(p_filterId) .. ", Value: " .. tostring(p_value), true)
	else
		--Old behaviour with filters
		settings.isFilterOn[p_filterId] = p_value
        debugMessage("[SetSettingsIsFilterOn] FilterId: " .. tostring(p_filterId) .. ", Value: " .. tostring(p_value), true)
    end
    --return the value
    return p_value
end

-- Check the settings for the panels and return if enabled or disabled
local function getFilterWhereBySettings(p_filterWhere)
	--Set the resultVar and update the settings.atPanelEnabled array
    if     (p_filterWhere == INV_BAGS) then
		settings.atPanelEnabled[p_filterWhere]["filters"] = settings.allowInventoryFilter
    elseif (p_filterWhere == INV_STORE) then
		settings.atPanelEnabled[p_filterWhere]["filters"] = settings.allowVendorFilter
    elseif (p_filterWhere == INV_DECONSTRUCTION ) then
		settings.atPanelEnabled[p_filterWhere]["filters"] = settings.allowDeconstructionFilter
    elseif (p_filterWhere == INV_IMPROVEMENT ) then
		settings.atPanelEnabled[p_filterWhere]["filters"] = settings.allowImprovementFilter
    elseif (p_filterWhere == INV_ENCHANTING_EXTRACTION ) then
		settings.atPanelEnabled[p_filterWhere]["filters"] = settings.allowEnchantingFilter
    elseif (p_filterWhere == INV_ENCHANTING_CREATION ) then
		settings.atPanelEnabled[p_filterWhere]["filters"] = settings.allowEnchantingFilter
    elseif (p_filterWhere == INV_BANK) then
		settings.atPanelEnabled[p_filterWhere]["filters"] = settings.allowBankFilter
    elseif (p_filterWhere == INV_GUILDBANK) then
		settings.atPanelEnabled[p_filterWhere]["filters"] = settings.allowGuildBankFilter
    elseif (p_filterWhere == INV_GUILDSTORE) then
		settings.atPanelEnabled[p_filterWhere]["filters"] = settings.allowTradinghouseFilter
    elseif (p_filterWhere == INV_TRADE) then
		settings.atPanelEnabled[p_filterWhere]["filters"] = settings.allowTradeFilter
    elseif (p_filterWhere == INV_MAIL) then
		settings.atPanelEnabled[p_filterWhere]["filters"] = settings.allowMailFilter
	end

	--These settings are currently the same for all panels
    settings.atPanelEnabled[p_filterWhere]["anti-destroy"] 		  = settings.blockDestroying
    settings.atPanelEnabled[p_filterWhere]["anti-deconstruct"] 	  = settings.blockDeconstruction
    settings.atPanelEnabled[p_filterWhere]["anti-improve"] 		  = settings.blockImprovement
    settings.atPanelEnabled[p_filterWhere]["anti-sell"] 		  = settings.blockSelling
	settings.atPanelEnabled[p_filterWhere]["anti-sell-exception"] = settings.allowSellingForBlockedOrnate
    settings.atPanelEnabled[p_filterWhere]["anti-trade"] 		  = settings.blockTrading
    settings.atPanelEnabled[p_filterWhere]["anti-mail"] 		  = settings.blockSendingByMail

    debugMessage("[getFilterWhereBySettings] " .. tostring(p_filterWhere), false)

    return p_filterWhere
end

--Refresh the backpack list
local function RefreshBackpack()
    if not BACKPACK:IsHidden() then
		debugMessage("[RefreshBackpack] Backpack refresh", true)
    	ZO_ScrollList_RefreshVisible(BACKPACK)
    end
end

--Refresh the bank list
local function RefreshBank()
	if not BANK:IsHidden() then
		debugMessage("[RefreshBank] Bank refresh", true)
    	ZO_ScrollList_RefreshVisible(BANK)
    end
end

--Refresh the guild bank list
local function RefreshGuildBank()
	if not GUILD_BANK:IsHidden() then
		debugMessage("[RefreshGuildBank] Guild bank refresh", true)
	    ZO_ScrollList_RefreshVisible(GUILD_BANK)
    end
end

--Refresh the list dialog 1 scroll list
local function RefreshListDialog()
	if not LIST_DIALOG:IsHidden() then
		debugMessage("[RefreshListDialog] List Dialog refresh", true)
    	ZO_ScrollList_RefreshVisible(LIST_DIALOG)
    end
end

--Refresh the scroll lists
local function UpdateInventories()
    --Check if we are at a crafting station
	local craftInteractiontype = GetCraftingInteractionType()
	debugMessage("[UpdateInventories] CraftingInteractionType: " .. tostring(craftInteractiontype), true)
    --Are we not inside a crafting station? Else abort as this function will only update the player inventory
    if craftInteractiontype ~= 0 then return end

	--Update the current shown inventory
    if (locVars.gFilterWhere ~= nil and locVars.gLastFilterId ~= nil) then
		--Is the filter for this panel enabled in the settings?
		if settings.atPanelEnabled[locVars.gFilterWhere]["filters"] == true then
			--Is the filter we have added the icon for currently enabled(registered)?
           	local isFilterEnabled = getSettingsIsFilterOn(locVars.gLastFilterId, locVars.gFilterWhere)
			if isFilterEnabled == true or isFilterEnabled == -99 then
		    	debugMessage("Filter Id: " ..tostring(locVars.gLastFilterId) .. ", Filter panel Id: " .. tostring(locVars.gFilterWhere) .. ", UpdaterName: " .. filterTypeToUpdaterName[locVars.gFilterWhere], true)
			    inventoryUpdaters[filterTypeToUpdaterName[locVars.gFilterWhere]]()
            end
        end
    end

	--Update the scroll list controls for the player inventories
	RefreshBackpack()
	RefreshBank()
	RefreshGuildBank()
    RefreshListDialog()
end

local function UpdateCraftingInventory()
    --Check if we are at a crafting station
	local locCraftType = GetCraftingInteractionType()
	debugMessage("[UpdateCraftingInventory] CraftingInteractionType: " .. tostring(locCraftType), true)
	--Abort if we are not at a crafting station
	if locCraftType == 0 then return end

	--Update the current shown inventory
    if (locVars.gFilterWhere ~= nil and locVars.gLastFilterId ~= nil) then
		--Is the filter for this panel enabled in the settings?
		if settings.atPanelEnabled[locVars.gFilterWhere]["filters"] == true then
			--Is the filter we have added the icon for currently enabled(registered)?
           	local isFilterEnabled = getSettingsIsFilterOn(locVars.gLastFilterId, locVars.gFilterWhere)
			if isFilterEnabled == true or isFilterEnabled == -99 then
		    	debugMessage("Filter Id: " ..tostring(locVars.gLastFilterId) .. ", Filter panel Id: " .. tostring(locVars.gFilterWhere) .. ", UpdaterName: " .. filterTypeToUpdaterName[locVars.gFilterWhere], true)
			    inventoryUpdaters[filterTypeToUpdaterName[locVars.gFilterWhere]]()
            end
        end
    end

	--Enchanting?
	if (locCraftType == CRAFTING_TYPE_ENCHANTING) then
       	debugMessage("[UpdateCraftingInventory] Enchanting refresh", true)
		--Only refresh the scroll list
		ZO_ScrollList_RefreshVisible(ENCHANTING_STATION)
	else
	--Other crafting stations

		--Improvement
        if     (locVars.gFilterWhere == INV_IMPROVEMENT) then
	       	debugMessage("[UpdateCraftingInventory] Improvement refresh", true)
			--Are we at an improvement panel?
			--Only refresh the scroll list
			ZO_ScrollList_RefreshVisible(IMPROVEMENT)
		--Deconstruction
		elseif (locVars.gFilterWhere == INV_DECONSTRUCTION) then
	       	debugMessage("[UpdateCraftingInventory] Deconstruction refresh", true)
			--Are we at a deconstruction panel?
			--Only refresh the scroll list
			ZO_ScrollList_RefreshVisible(DECONSTRUCTION)
		end

--[[
		--Improvement
        if     (not IMPROVEMENT_INV:IsHidden()) then
	       	debugMessage("[UpdateCraftingInventory] Improvement refresh", true)
			--Are we at an improvement panel?
			--Only refresh the scroll list
			ZO_ScrollList_RefreshVisible(IMPROVEMENT)
		--Deconstruction
		elseif (not DECONSTRUCTION_INV:IsHidden()) then
	       	debugMessage("[UpdateCraftingInventory] Deconstruction refresh", true)
			--Are we at a deconstruction panel?
			--Only refresh the scroll list
			ZO_ScrollList_RefreshVisible(DECONSTRUCTION)
		end
]]
	end
end

--converts unsigned itemId to signed
local function SignItemId(itemId)
	local SIGNED_INT_MAX = 2^32 / 2 - 1
	local INT_MAX 		 = 2^32
	if(itemId and itemId > SIGNED_INT_MAX) then
		itemId = itemId - INT_MAX
	end
	return itemId
end

local function MyGetItemNameNoControl(bagId, slotIndex)
	local name = "Not found"
    local itemData
    if bagId < 1 or bagId > 4 then bagId = 1 end
	--[[
		inventoryTypes for PLAYER_INVENTORY:
		INVENTORY_BACKPACK 	 = 1
		INVENTORY_QUEST_ITEM = 2
		INVENTORY_BANK		 = 3
		INVENTORY_GUILD_BANK = 4
	]]
	itemData = PLAYER_INVENTORY.inventories[bagId].slots[slotIndex]
	if(itemData ~= nil) then
		name = itemData.name
	end
	return name
end

local function MyGetItemName(rowControl)
	local name
    if (rowControl == nil) then return end
	local dataEntry = rowControl.dataEntry

	--case to handle equiped items
	if(not dataEntry) then
		name = rowControl.name
	else
		name = dataEntry.data.name
	end
	return name
end

local function MyGetItemDetails(rowControl)
	local bagId, slotIndex

	--gotta do this in case deconstruction, or player equipment
    local dataEntry = rowControl.dataEntry

	--case to handle equiped items
	if(not dataEntry) then
    	bagId = rowControl.bagId
        slotIndex = rowControl.slotIndex
	else
		bagId = dataEntry.data.bagId
		slotIndex = dataEntry.data.slotIndex
	end

	--case to handle list dialog, list dialog uses index instead of slotIndex and bag instead of bagId...?
	if(dataEntry and not bagId and not slotIndex) then
		bagId = rowControl.dataEntry.data.bag
		slotIndex = rowControl.dataEntry.data.index
	end

	return bagId, slotIndex
end

local function MyGetItemInstanceIdNoControl(bagId, slotIndex)
	local itemId = GetItemInstanceId(bagId, slotIndex)
	return SignItemId(itemId)
end

local function MyGetItemInstanceId(rowControl)
	local bagId, slotIndex = MyGetItemDetails(rowControl)

	local itemId = GetItemInstanceId(bagId, slotIndex)
	return SignItemId(itemId)
end

-- Is the item researchable?
local function isItemResearchableNoControl(bagId, slotIndex)
	local itemType = GetItemType(bagId, slotIndex)
	if allowedResearchableItemTypes[itemType] then
		return true
	end
   	return false
end

-- Is the item researchable?
local function isItemResearchable(p_rowControl)
    local bagId, slotIndex = MyGetItemDetails(p_rowControl)
	local itemType = GetItemType(bagId, slotIndex)
	if allowedResearchableItemTypes[itemType] then
		return true
	end
   	return false
end

-- Is the item an ornate one?
local function isItemOrnate(bagId, slotIndex)
    local itemTrait = GetItemTrait(bagId, slotIndex)
	if allowedOrnateItemTraits[itemTrait] then
    	return true
	end
   	return false
end

--Get the FCOItemSaver control
local function GetItemSaverControl(parent, controlId, useParentFallback)
    --debugMessage("[GetItemSaverControl] Parent: " .. parent:GetName() .. ", ControlId: " .. tostring(controlId), true)
	local retControl = parent:GetNamedChild(addonVars.gAddonName .. tostring(controlId))

	--Use the parent control as a fallback?
	if useParentFallback == true then
	    --e.g. Inside enchanting the parent control is the correct one already
	    if (retControl == nil) then
	    	retControl = parent
	    end
    end
    return retControl
end

--The function to update the inventories and lists after an item was un/marked
local function FilterBasics(onlyPlayer)
	--Check if we are in the player inventory
	if not BACKPACK:IsHidden() then
		--we are in the player inventory (or in the banks at the deposit inventories, or at mail sending, or trading)
		onlyPlayer = true
	end

	debugMessage("[FilterBasics] onlyPlayer: " .. tostring(onlyPlayer), true)

	--Only update the lists if not currently already updating
	if (preventerVars.gFilteringBasics == false) then
		preventerVars.gFilteringBasics = true
	    if (onlyPlayer == true) then
			UpdateInventories()
	    else
			UpdateInventories()
			UpdateCraftingInventory()
	    end
		preventerVars.gFilteringBasics = false
    end
end

-- Check the weapon type
local function checkWeaponType(p_weaponType, p_check)
	if (p_check == nil or p_check == '') then
		p_check = '1hd'
	end
	if (p_weaponType ~= nil) then
		if weaponTypeCheckTable[p_check] and weaponTypeCheckTable[p_check][p_weaponType] then
			return true
		end
	end
    return false
end

--Check if the current equipment slot is the offhand or backup offhand slot
local function checkWeaponOffHand(controlName, weaponTypeName, doCheckOffHand, doCheckBackupOffHand, echo)
    if echo then
		debugMessage("[CheckWeaponOffHand] ControlName: " .. controlName ..", WeaponTypeName: " .. weaponTypeName, true)
    end
	local weaponControl
    local weaponType

	--Check the offhand?
    if doCheckOffHand == true then
		if (controlName == 'ZO_CharacterEquipmentSlotsOffHand') then
			--Check if the weapon in the main slot is a 2hd weapon/staff
			weaponControl = WINDOW_MANAGER:GetControlByName("ZO_CharacterEquipmentSlotsMainHand", "")
			if (weaponControl ~= nil) then
				weaponType = GetItemWeaponType(weaponControl.bagId, weaponControl.slotIndex)
				if (weaponType ~= nil) then
					return checkWeaponType(weaponType, weaponTypeName)
				end
			end
		end
	end

	--Check the backup offhand?
    if doCheckBackupOffHand == true then
	    if (controlName == 'ZO_CharacterEquipmentSlotsBackupOff') then
			--Check if the weapon in the backup slot is a 2hd weapon/staff
			weaponControl = WINDOW_MANAGER:GetControlByName("ZO_CharacterEquipmentSlotsBackupMain", "")
			if (weaponControl ~= nil) then
				weaponType = GetItemWeaponType(weaponControl.bagId, weaponControl.slotIndex)
				if (weaponType ~= nil) then
					return checkWeaponType(weaponType, weaponTypeName)
				end
			end
		end
    end

    return false
end

--Check if a two handed weapon is used and hide the marker for the offhand slot, or find empty slots and remove markers
function checkForEmptyOrTwoHandedWeaponMarker(slot, onlyWeapons)

	--Only check if character window is shown
    if (CHARACTER:IsHidden()) then
    	return
    end

	local control
    local offHandControlName
    local twoHandedWeaponOffHand = false
    local offHandControl
    local offHandCheck = false
    local backupOffHandCheck = false

	--Only check a specific character item slot?
    if slot ~= nil and slot ~= -1 and onlyWeapons == false then
        equipmentName = characterEquipmentSlotNameByIndex[slot]
		--Abort if equipment name was not found
        if equipmentName == nil or equipmentName == '' then return end

		debugMessage("[CheckForEmptyOrTwoHandedWeaponMarker] Slot: " ..tostring(slot) .. ", Name: " .. equipmentName, false)

		offHandControl = WINDOW_MANAGER:GetControlByName(equipmentName, "")
		if offHandControl ~= nil then
			if offHandControl.stackCount > 0 then
                if (slot==4 or slot==5 or slot==20 or slot==21) then
	            	if (slot==4 or slot==5) then offHandCheck = true end
					if (slot==20 or slot==21) then backupOffHandCheck = true end
					offHandControlName = offHandControl:GetName()
					twoHandedWeaponOffHand = checkWeaponOffHand(offHandControlName, "2hdall", offHandCheck, backupOffHandCheck, true)
                end -- if (slot==4 ...
            end -- if offHandControl.stack ...
		    --Do for each icon Id
		    for j=1, locVars.gFCONumFilterIcons, 1 do
       			control = nil
				--OffHand is a 2hd weapon or no item is equipped, so remove the markers
		        if (offHandControl.stackCount <= 0 or twoHandedWeaponOffHand) then
					--control = offHandControl:GetNamedChild(addonVars.gAddonName .. tostring(j))
                    control = GetItemSaverControl(offHandControl, j, false)
		            if control ~= nil then
						control:SetHidden(true)
		            end
		        end
			end -- for j=1 ...
		end -- 	if offHandControl ~= nil

	else
		--Check all equipment items or weapons only

		--For each equipment item
	   	for index, _ in pairs(characterEquipmentSlotNameByIndex) do
			if ( onlyWeapons == false) then
				--Recursively check all the character equipment slots
               	checkForEmptyOrTwoHandedWeaponMarker(index, onlyWeapons)
 			else
				--Only check the weapons
				if (index==4 or index==5 or index==20 or index==21) then
					--Recursively check all the character equipment slots
					checkForEmptyOrTwoHandedWeaponMarker(index, false)
				end
            end -- if ( (onlyWeapons....
		end -- for ... in pairs ...
    end

end

--Function that close the context menu (if for example the user closes the inventory without
--choosing a option on the context menu)
local function hideContextMenu(whichContextMenu)
	if whichContextMenu == nil or whichContextMenu == -1 then
        hideContextMenu(locVars.gFilterWhere)
--[[
    	for i=1, locVars.gFCONumFilterInventoryTypes, 1 do
			if contextMenu[i] ~= nil then
				if not contextMenu[i]:IsHidden() then
					contextMenu[i]:SetHidden(true)
				end
			end
	    end
]]
    else
		if contextMenu[whichContextMenu] ~= nil then
			if not contextMenu[whichContextMenu]:IsHidden() then
				contextMenu[whichContextMenu]:SetHidden(true)
			end
		end
    end
end

--Create the marker controls, holding the icon textures
local function CreateMarkerControl(parent, controlId, pWidth, pHeight, pTexture, pIsEquipmentSlot, pCreateControlIfNotThere)
    --No parent? Abort here
	if parent == nil then return nil end

    local InventoryGridViewActivated = false
	--debugMessage("[CreateMarkerControl] Control: " .. parent:GetName() .. ", EquipmentSlot: " .. tostring(pIsEquipmentSlot) .. ", ControlId: "..tostring(controlId) .. ", Create: " .. tostring(pCreateControlIfNotThere), true)

	--Preset the variable for control creation with false, if it is not given
    pCreateControlIfNotThere = pCreateControlIfNotThere or false

	if(parent:GetOwningWindow() ~= QUICKSLOT_CIRCLE) then
        if pIsEquipmentSlot == nil then pIsEquipmentSlot = false end

    	--Does the FCOItemSaver marker control exist already?
		local control = GetItemSaverControl(parent, controlId, false)
		--local control = parent:GetNamedChild(addonVars.gAddonName .. tostring(controlId))

        local doHide

		--Marker control for a disabled gear? Hide the icon then
	    local gearId = iconToGear[controlId]
	    if gearId ~= nil and not settings.isGearEnabled[gearId] then
        	doHide = true
		else
			doHide = not markedItems[controlId][MyGetItemInstanceId(parent)]
        end
        if doHide == nil then doHide = false end

		--It does not exist yet, so create it now
		if(control == parent or not control) then
			--Abort here if control should be hiden and is not created yet
            if doHide == true and pCreateControlIfNotThere == false then return end

            --If not aborted: Create the marker control now
            control = WINDOW_MANAGER:CreateControl(parent:GetName() .. addonVars.gAddonName .. tostring(controlId), parent, CT_TEXTURE)
		end
		--Control did already exist or was created now
	    if control ~= nil then
			--Hide or show the control now
			control:SetHidden(doHide)

			--Remove the sell icon and price if Inventory gridview is active
			if(parent:GetWidth() - parent:GetHeight() < 5) then
				if(parent:GetNamedChild("SellPrice")) then
					parent:GetNamedChild("SellPrice"):SetHidden(true)
				end
                InventoryGridViewActivated = true
            end

			if not doHide then
	        	control:SetDimensions(pWidth, pHeight)
				control:SetTexture(pTexture)
				control:SetColor(settings.icon[controlId].color.r, settings.icon[controlId].color.g, settings.icon[controlId].color.b, settings.icon[controlId].color.a)
				--Marker was created/updated for the character equipment slots?
				if pIsEquipmentSlot == true then
					control:ClearAnchors()
					control:SetAnchor(BOTTOMLEFT, parent, BOTTOMLEFT)
					control:SetDrawTier(1)
				else
					if InventoryGridViewActivated == true then
						control:SetDrawTier(DT_HIGH)
						control:ClearAnchors()
						control:SetAnchor(CENTER, parent, BOTTOMLEFT, 12, -12)
					else
						control:ClearAnchors()
						control:SetAnchor(LEFT, parent, LEFT)
					end
				end
            end  -- if not doHide then
			return control
	    else
	    	return nil
	    end
    else
    	return nil
    end
end

--The function to refresh the equipped items
local function RefreshEquipmentControl(equipmentControl, doCreateMarkerControl, p_markId)
	--Preset the value for "Create control if not existing yet" with false
    doCreateMarkerControl = doCreateMarkerControl or false

	--is the equipment control already given?
    if equipmentControl ~= nil then
        if p_markId == nil or p_markId == 0 then
			debugMessage("[RefreshEquipmentControl] Control: " .. equipmentControl:GetName() .. ", Create: " .. tostring(doCreateMarkerControl) .. ", MarkId: ALL", true)
	    	--Add/Update the markers for the filter icons at the equipment slot
			for j=1, locVars.gFCONumFilterIcons, 1 do
				CreateMarkerControl(equipmentControl, j, locVars.gEquipmentIconWidth, locVars.gEquipmentIconHeight, MARKER_TEXTURES[settings.icon[j].texture], true, doCreateMarkerControl)
			end
        else
			debugMessage("[RefreshEquipmentControl] Control: " .. equipmentControl:GetName() .. ", Create: " .. tostring(doCreateMarkerControl) .. ", MarkId: " .. tostring(p_markId), true)
	    	--Add/Update the marker p_markId for the filter icons at the equipment slot
			CreateMarkerControl(equipmentControl, p_markId, locVars.gEquipmentIconWidth, locVars.gEquipmentIconHeight, MARKER_TEXTURES[settings.icon[p_markId].texture], true, doCreateMarkerControl)
        end
    else
		--Check all equipment controls
		debugMessage("[RefreshEquipmentControl] ALL CONTROLS! Create: " .. tostring(doCreateMarkerControl) .. ", MarkId: " .. tostring(p_markId), true)

        local equipmentSlotControl
	    for _, equipmentSlotName in pairs(characterEquipmentSlotNameByIndex) do
			--Get the control of the equipment slot
	    	equipmentSlotControl = WINDOW_MANAGER:GetControlByName(equipmentSlotName, "")
	        if equipmentSlotControl ~= nil then
            	RefreshEquipmentControl(equipmentSlotControl, doCreateMarkerControl, p_markId)
	        end
		end
    end
end

--The callback function for the right click/context menus to mark an item
local function MarkMe(rowControl, markId, updateNow, doUnmark, refreshPopupDialog)
	--Set the last used filter Id
    locVars.gLastFilterId = iconIdToFilter[markId]

	local itemId = MyGetItemInstanceId(rowControl)
    if markId ~= nil and itemId ~= nil then
		if doUnmark == true then
			markedItems[markId][itemId] = nil
	    else
			markedItems[markId][itemId] = true
	    end

		GetItemSaverControl(rowControl, markId, true):SetHidden(doUnmark)
	    if (updateNow == true) then
			--Inventories or character equipment?
			if rowControl:GetParent() ~= CHARACTER then
	        	FilterBasics(false)
	        else
				RefreshEquipmentControl(rowControl, not doUnmark, markId)
	        end
		else
			--Refresh the ZO_ListDialog1 popup now after right click/context menu was used?
			if refreshPopupDialog then
				--To remove the active selected row again (was activated during mouse right click/context menu
				RefreshListDialog()
	            ClearCursor()
				--if researching of marked items is not allowed
	            if not settings.allowResearch then
					--Disable the "Research" button in the popup
					WINDOW_MANAGER:GetControlByName("ZO_ListDialog1Button1", ""):SetEnabled(false)
	            end
			end
	    end
    end
end

--The callback function for the right click/context menus to mark all equipment items with one click
local function MarkAllEquipment(rowControl, markId, updateNow, doHide)
	--Set the last used filter Id
    locVars.gLastFilterId = iconIdToFilter[markId]

	local equipmentControl
    local itemId
    local equipmentSlotName
    local dontChange
	local fcoMarkerControl

	--Get each equipped item
	for i=1, CHARACTER:GetNumChildren() do
        dontChange = false
    	equipmentSlotName = CHARACTER:GetChild(i):GetName()

   		debugMessage("[MarkAllEquipment] MarkId: " .. tostring(markId) .. ", EquipmentSlot: "..equipmentSlotName..", no_auto_mark: " .. tostring(equipmentSlotsNames["no_auto_mark"][equipmentSlotName]))
		--Is the current equipment slot a changeable one?
		if(string.find(equipmentSlotName, "ZO_CharacterEquipmentSlots")) then
			--Only mark weapons too if enabled in settings
			if ( (settings.autoMarkAllWeapon == false and allowedCharacterEquipmentWeaponControlNames[equipmentSlotName] == true)
			--Only mark jewelry too if enabled in settings
			  or (settings.autoMarkAllJewelry == false and allowedCharacterEquipmentJewelryControlNames[equipmentSlotName] == true) ) then
	            dontChange = true
            end
			--Still need to change the texture and un-/mark item?
            if dontChange == false then
				dontChange = equipmentSlotsNames["no_auto_mark"][equipmentSlotName] or false
            end
			--Are we allowed to change the marker?
        	if (dontChange == false) then
				--get the control of the current equipment slot
	            equipmentControl = CHARACTER:GetChild(i)
				if (equipmentControl ~= nil) then
					--get the itemid
		            itemId = MyGetItemInstanceId(equipmentControl)
                    if markId ~= nil and itemId ~= nil then
						if doHide == true then
							markedItems[markId][itemId] = nil
						else
							markedItems[markId][itemId] = true
						end
						--Update the texture, create it if not there yet
						RefreshEquipmentControl(equipmentControl, true, markId)
                    end
	            end
			end
        end
    end

    --if (updateNow == true) then
	    --FilterBasics(false)
    --end
end

--Add the icon to the item and add the right click menu entry
local function AddMark(rowControl, markId, isEquipmentSlot, refreshPopupDialog)
	local parentName = rowControl:GetParent():GetName()
    local controlName = rowControl:GetName()

    if parentName == nil or controlName == nil then return end

	--To prevent spamming only output the debug message once for markId = 1
    if markId == 1 then
	    debugMessage("[AddMark]: Parent: " .. parentName .. ", Control: " .. controlName .. ", IsEquipmentSlot: " ..tostring(isEquipmentSlot), true)
    end
    local isEquipmentSlotContextmenu = false

    --For equipment slots: Only go on if markId == 2, 4, 6, 7 or 8 (gear sets 1 to 5)
    --and an item is equipped (rowControl.stackCount ~= 0)
    if (isEquipmentSlot == true) then
    	if (markId == 1 or markId == 3 or markId == 5) then
        	if (rowControl.stackCount == 0) then
        		return
            end
        end

	    --For 2handed weapons/staffs: Only go on if the current equipment slot is not
        --the 1st weapon set's backup or the 2nd weapon set's backup
		if checkWeaponOffHand(controlName, "2hdall", true, true, markId==1) == true then return end
    end

	--===========================================================================================================
	--Check if the right click menu should be updated. Only allowed panels and menus apply
    -- NEW
    -- Check two tables for parent and control names. If current control and parent are not in the relating table
    -- the contextmenu will be enhanced with FCOItemSaver entries
    if (notAllowedContextMenuParentControls[parentName] or notAllowedContextMenuControls[controlName]) then
		--Not allowed parent or control is given -> Abort here
    	return
    end

	-- Equipment gear 1, 2, 3, 4, 5 or Research
	if (markId == 2 or markId == 3 or markId == 4 or markId == 6 or markId == 7 or markId == 8) then
		-- Check if item is researchable (as only researchable items can work as equipment too)
		if (not isItemResearchable(rowControl)) then
			return
		end
	end

	--Is the setting for "mark all equipped items at once" activated and we are handling an equipment slot
       --and are we trying to add options for gear 1, 2, 3, 4, 5?
	if (settings.autoMarkAllEquipment == true) then
		if ( isEquipmentSlot == true and (markId == 2 or markId == 4 or markId == 6 or markId == 7 or markId == 8) ) then
			isEquipmentSlotContextmenu = true
			if (settings.autoMarkAllWeapon == false) then
				if ( allowedCharacterEquipmentWeaponControlNames[controlName] ) then
					isEquipmentSlotContextmenu = false
				end
			end
			if (settings.autoMarkAllJewelry == false and isEquipmentSlotContextmenu == true) then
				if ( allowedCharacterEquipmentJewelryControlNames[controlName] ) then
					isEquipmentSlotContextmenu = false
				end
			end
		end
	else
		isEquipmentSlotContextmenu = false
	end

	--Update the list / popup dialog list?
    local refreshList
    if refreshPopupDialog == true then
    	refreshList = false
    else
    	refreshList = true
    end

	--Add the equipment right click / context menu entries
	if (isEquipmentSlotContextmenu == true) then
		-- Add/Update the right click menu item for character slot now
		if(not markedItems[markId][MyGetItemInstanceId(rowControl)]) then
			AddMenuItem(lTextEquipmentMark[markId], function() MarkAllEquipment(rowControl, markId, refreshList, false) end, MENU_ADD_OPTION_LABEL)
		else
			AddMenuItem(lTextEquipmentDemark[markId], function() MarkAllEquipment(rowControl, markId, refreshList, true) end, MENU_ADD_OPTION_LABEL)
		end

	--Add the right click / context menu entries
	else
		-- Add/Update the right click menu item now
		if(not markedItems[markId][MyGetItemInstanceId(rowControl)]) then
			AddMenuItem(lTextMark[markId], function() MarkMe(rowControl, markId, refreshList, false, refreshPopupDialog) end, MENU_ADD_OPTION_LABEL)
		else
			AddMenuItem(lTextDemark[markId], function() MarkMe(rowControl, markId, refreshList, true, refreshPopupDialog) end, MENU_ADD_OPTION_LABEL)
		end
	end
    --Show the new added menu entries inside the context menu borders
    ShowMenu(rowControl)
end

--Filter callBack function for deconstruction & enchanting panels
--local function FilterSavedItems(bagId, slotIndex, ...)
local function FilterSavedItems(bagId, slotIndex, ...)
	--This function will be executed once for EACH ITEM SLOT in your inventory/bank,
    --if you open a crafting station
	--The returned value of this filter function must be either
    --true  - Show the slot
    --false - Hide the slot

	-- Return value variable initalization: Show the slot
	local result = true
    local isFilterActivated
    local slotItemInstanceId = MyGetItemInstanceIdNoControl(bagId, slotIndex)

	--Check for each filter the marked items
	for i=1, locVars.gFCONumFilters, 1 do
		--Check if filter is activated for current slot
        isFilterActivated = getSettingsIsFilterOn(i)

        --Special treatment for filter type 2 as it handels "gear sets" marked items arrays 2, 4, 6, 7 and 8
		if (i == 2) then
			--Filter 2 on
	    	if(isFilterActivated == true
               and (   markedItems[i][slotItemInstanceId]
               	    or markedItems[4][slotItemInstanceId]
               	    or markedItems[6][slotItemInstanceId]
               	    or markedItems[7][slotItemInstanceId]
               	    or markedItems[8][slotItemInstanceId]
                    )
            ) then
            	return false
			--Filter 2 "show only marked"
		    elseif(isFilterActivated == -99) then
		    	if (    markedItems[i][slotItemInstanceId]
                	 or markedItems[4][slotItemInstanceId]
               	     or markedItems[6][slotItemInstanceId]
               	     or markedItems[7][slotItemInstanceId]
               	     or markedItems[8][slotItemInstanceId]
                ) then
			    	return true
		        else
					result = false
		        end
			--Filter 2 off
            else
				if (result ~= false) then
					result = true
                end
			end

        --Special treatment for filter type 4, as the marked items are stored in array 5
		elseif (i == 4) then
			--Filter 4 on
	    	if(isFilterActivated == true and markedItems[i+1][slotItemInstanceId]) then
            	return false
			--Filter 4 "show only marked"
		    elseif(isFilterActivated == -99) then
		    	if (markedItems[i+1][slotItemInstanceId]) then
                	return true
		        else
                	result = false
		        end
			--Filter 4 off
            else
				if (result ~= false) then
					result = true
                end
            end


		-- Normal treatment here
        else
			--Filter 1 (or 3, or others) on
	    	if(isFilterActivated == true and markedItems[i][slotItemInstanceId]) then
            	return false
			--Filter 1 (or 3, or others) "show only marked"
		    elseif(isFilterActivated == -99) then
		    	if (markedItems[i][slotItemInstanceId]) then
			    	return true
		        else
                	result = false
		        end
			--Filter 1 (or 3, or others) off
            else
				if (result ~= false) then
					result = true
                end
			end
        end

        --debugMessage("[FilterSavedItems]: Bag: " .. tostring(bagId) .. ", Slot: " .. tostring(slotIndex) .. ", Name: " .. MyGetItemNameNoControl(bagId, slotIndex) .. ", Filter: " .. i .. ", Result: " .. tostring(result), true)
	end

	-- Return the result if all filters were cross-checked and last filter is reached
	return result
end

--filter callBack function for bags, bank, mail, trade, guild bank, guild store, etc.
local function FilterSavedItemsForShop(slot)
	--This function will be executed once for EACH ITEM SLOT in your inventory
	--The returned value of this filter function must be either
    --true  - Show the slot
    --false - Hide the slot

	-- Return value variable initalization: Show the slot
	local result = true
	local isFilterActivated
    local slotItemInstanceId = MyGetItemInstanceId(slot)

	--Check for each filter the marked items
	for i=1, locVars.gFCONumFilters, 1 do
    	--Check if filter is activated for current slot
        isFilterActivated = getSettingsIsFilterOn(i)

        --Special treatment for filter type 2 as it handels arrays 2, 4, 6, 7 and 8
		if (i == 2) then
			--Filter 2 on
	    	if(isFilterActivated == true
               and (   markedItems[i][slotItemInstanceId]
               	    or markedItems[4][slotItemInstanceId]
               	    or markedItems[6][slotItemInstanceId]
               	    or markedItems[7][slotItemInstanceId]
               	    or markedItems[8][slotItemInstanceId]
                    )
            ) then
            	return false
			--Filter 2 "show only marked"
		    elseif(isFilterActivated == -99) then
		    	if (    markedItems[i][slotItemInstanceId]
                	 or markedItems[4][slotItemInstanceId]
                	 or markedItems[6][slotItemInstanceId]
                	 or markedItems[7][slotItemInstanceId]
                	 or markedItems[8][slotItemInstanceId]
                ) then
			    	return true
		        else
					result = false
		        end
			--Filter 2 off
            else
				if (result ~= true) then
					result = false
                end
			end


        --Special treatment for filter type 4, as the marked items are stored in array 5
		elseif (i == 4) then
			--Filter 4 on
	    	if(isFilterActivated == true and markedItems[i+1][slotItemInstanceId]) then
            	return false
			--Filter 4 "show only marked"
		    elseif(isFilterActivated == -99) then
		    	if (markedItems[i+1][slotItemInstanceId]) then
                	return true
		        else
                	result = false
		        end
			--Filter 4 off
            else
				if (result ~= true) then
					result = false
                end
            end


		-- Normal treatment here
        else
			--Filter 1 (or 3, or others) on
	    	if(isFilterActivated == true and markedItems[i][slotItemInstanceId]) then
            	return false
			--Filter 1 (or 3, or others) "show only marked"
		    elseif(isFilterActivated == -99) then
		    	if (markedItems[i][slotItemInstanceId]) then
			    	return true
		        else
                	result = false
		        end
			--Filter 1 (or 3, or others) off
            else
				if (result ~= true) then
					result = false
                end
			end
        end

        --debugMessage("[FilterSavedItemsForShop]: " ..MyGetItemName(slot) .. ", Filter: " .. i .. ", Result: " .. tostring(result), true)
	end

	-- Return the result if all filters were cross-checked and last filter is reached
	return result
end

--Update the button colors and textures depending on the filters
local function UpdateButtonColors(p_buttonId, p_button, p_status, p_filterPanelId)
	local btnName = "No button"
    if p_button ~= nil then
    	btnName = p_button:GetName()
    end
	--If the filterPanelId parameter is empty use the global filter panel Id from events and button clicks
    p_filterPanelId = p_filterPanelId or locVars.gFilterWhere

    debugMessage("[UpdateButtonColors] Button: " .. btnName .. ", ButtonId: " .. tostring(p_buttonId) .. ", Status: " .. tostring(p_status) .. ", FilterPanelId: " .. tostring(p_filterPanelId), true)
   	local texture
	local offset

--------------------------------------------------------------------------------
-- New filter method
--------------------------------------------------------------------------------
	if(settings.splitFilters == true and p_filterPanelId ~= nil) then
		-- New filter method
		if (p_buttonId == -1 or p_status == -1) then
	    -- for Schleife für 4 Buttons -> Initialisierung beim Laden des Addons
	    -- Läd für die Ansicht p_filterPanelId die Farben der 4 Filter Buttons
		    for i=1, locVars.gFCONumFilters, 1 do
		    	UpdateButtonColors(i, nil, getSettingsIsFilterOn(i, p_filterPanelId), p_filterPanelId)
		    end
	    else

	        if (p_button == nil) then
	        	-- Die 4 Filter Buttons einer Ansicht updaten!
	    		-- Ansichten = (Inventar, Guild shop, Verkäufer, Handwerksstation, Guild bank , Player bank, Verzauberungen erstellen, Verzauberungen extrahieren, Mail, Trade, Improvement)
--			    for j=1, locVars.gFCONumFilterPanels, 1 do
			        texture  = WINDOW_MANAGER:GetControlByName(string.format(gFilterPanelIdToTextureName[p_filterPanelId], p_buttonId), "")
                    --Does texture exist now?
                    if(texture ~= nil) then

						-- Fix to show icon 5 for filter 4
				        offset = 0
				        if (p_buttonId == 4) then
				        	offset = 1
				        end
				        --Workaround to show at least a default texure, if none is found
				  		if (MARKER_TEXTURES[settings.icon[p_buttonId + offset].texture] ~= nil) then
					        --Set the texture now
					        texture:SetTexture(MARKER_TEXTURES[settings.icon[p_buttonId + offset].texture])
						else
					        --Set fallback texture now
					        texture:SetTexture(MARKER_TEXTURES[p_buttonId])
							--Set the fallback texture to the settings menu
				            settings.icon[p_buttonId + offset].texture = p_buttonId
				        end

						--Only update colors if wished
	                    if (p_status ~= -999) then
							-- Special state between true and false (only shows marked items)
		                    if(p_status == -99) then
								-- Yellow button
					    		texture:SetColor(1, 1, 0, 1)
		                    else
			                	if(not p_status) then
		                        	-- red button
						    		texture:SetColor(1, 0, 0, 1)
						        else
		                        	-- Green button
						            texture:SetColor(0, 1, 0, 1)
						        end
		                    end
	                    end
					end

--	            end -- for j ...

	        else -- if (p_button == nil) then

				-- Einen ganz speziellen Button nur updaten
                texture = WINDOW_MANAGER:GetControlByName(string.format(gFilterPanelIdToTextureName[p_filterPanelId], p_buttonId), "")
				if(texture ~= nil) then

					-- Fix to show icon 5 for filter 4
			        offset = 0
			        if (p_buttonId == 4) then
			        	offset = 1
			        end
			        --Workaround to show at least a default texure, if none is found
			  		if (MARKER_TEXTURES[settings.icon[p_buttonId + offset].texture] ~= nil) then
				        --Set the texture now
				        texture:SetTexture(MARKER_TEXTURES[settings.icon[p_buttonId + offset].texture])
					else
				        --Set fallback texture now
				        texture:SetTexture(MARKER_TEXTURES[p_buttonId])
						--Set the fallback texture to the settings menu
			            settings.icon[p_buttonId + offset].texture = p_buttonId
			        end

					--Only update colors if wished
					if (p_status ~= -999) then
						-- Special state between true and false (only shows marked items)
						if(p_status == -99) then
						-- Yellow button
							texture:SetColor(1, 1, 0, 1)
						else
							if(not p_status) then
								-- red button
								texture:SetColor(1, 0, 0, 1)
							else
								-- Green button
								texture:SetColor(0, 1, 0, 1)
							end
						end
					end
				end
		    end

	    end

--------------------------------------------------------------------------------
-- Old filter method
--------------------------------------------------------------------------------
	else

        if settings.splitFilters == false then
	    -- Old filter method

			if (p_buttonId == -1 or p_status == -1) then

            -- for Schleife für 4 Buttons -> Initialisierung beim Laden des Addons
		    -- Läd für alle Ansichten die Farben der 4 Filter Buttons
			    for i=1, locVars.gFCONumFilters, 1 do
			    	UpdateButtonColors(i, nil, getSettingsIsFilterOn(i))
			    end
		    else

		        if (p_button == nil) then

		        	-- Die gleichen Filter Buttons aller Ansichten updaten!
		    		-- Ansichten = (Inventar, Guild shop, Verkäufer, Handwerksstation, Guild bank , Player bank, Verzauberungen erstellen, Verzauberungen extrahieren, Mail, Trade, Improvement)
				    --for j=1, locVars.gFCONumFilterPanels, 1 do

				        texture  = WINDOW_MANAGER:GetControlByName(string.format(gFilterPanelIdToTextureName[p_filterPanelId], p_buttonId), "")
						if(texture ~= nil) then

							-- Fix to show icon 5 for filter 4
					        offset = 0
					        if (p_buttonId == 4) then
					        	offset = 1
					        end
					        --Workaround to show at least a default texure, if none is found
					  		if (MARKER_TEXTURES[settings.icon[p_buttonId + offset].texture] ~= nil) then
						        --Set the texture now
						        texture:SetTexture(MARKER_TEXTURES[settings.icon[p_buttonId + offset].texture])
							else
						        --Set fallback texture now
						        texture:SetTexture(MARKER_TEXTURES[p_buttonId])
								--Set the fallback texture to the settings menu
					            settings.icon[p_buttonId + offset].texture = p_buttonId
					        end

							--Only update colors if wished
		                    if (p_status ~= -999) then
								-- Special state between true and false (only shows marked items)
			                    if(p_status == -99) then
									-- Yellow button
						    		texture:SetColor(1, 1, 0, 1)
			                    else
				                	if(not p_status) then
			                        	-- red button
							    		texture:SetColor(1, 0, 0, 1)
							        else
			                        	-- Green button
							            texture:SetColor(0, 1, 0, 1)
							        end
			                    end
		                    end
						end

		            --end -- for j=1, locVars.gFCONumFilterPanels, 1 do

		        else -- if (p_button == nil) then

					-- Einen ganz speziellen Button nur updaten
				    texture = nil
	                texture = WINDOW_MANAGER:GetControlByName(string.format(gFilterPanelIdToTextureName[p_filterPanelId], p_buttonId), "")
					if(texture ~= nil) then

						-- Fix to show icon 5 for filter 4
				        offset = 0
				        if (p_buttonId == 4) then
				        	offset = 1
				        end
				        --Workaround to show at least a default texure, if none is found
				  		if (MARKER_TEXTURES[settings.icon[p_buttonId + offset].texture] ~= nil) then
					        --Set the texture now
					        texture:SetTexture(MARKER_TEXTURES[settings.icon[p_buttonId + offset].texture])
						else
					        --Set fallback texture now
					        texture:SetTexture(MARKER_TEXTURES[p_buttonId])
							--Set the fallback texture to the settings menu
				            settings.icon[p_buttonId + offset].texture = p_buttonId
				        end

						--Only update colors if wished
						if (p_status ~= -999) then
							-- Special state between true and false (only shows marked items)
							if(p_status == -99) then
							-- Yellow button
								texture:SetColor(1, 1, 0, 1)
							else
								if(not p_status) then
									-- red button
									texture:SetColor(1, 0, 0, 1)
								else
									-- Green button
									texture:SetColor(0, 1, 0, 1)
								end
							end
						end
					end
			    end

		    end
        end -- if settings.splitFilters == false then

    end -- if filterPanelId ~= nil

end

--Filter the player inventory
local function FilterPlayerInventory(filterId, panelId)
    local newFilterMethod = settings.splitFilters
    panelId = panelId or locVars.gFilterWhere

    --Filtering inside inventory is enabled in the settings?
	if (settings.allowInventoryFilter == true) then
        --Register only 1 filter in the player inventory
    	if (filterId ~= -1) then
            if (newFilterMethod == true) then
	        	if(not libFilters:IsFilterRegistered("FCOItemSaver_PlayerInventoryFilterNew" .. tostring(panelId) .. "_" .. tostring(filterId))) then
					libFilters:RegisterFilter("FCOItemSaver_PlayerInventoryFilterNew" .. tostring(panelId) .. "_" .. tostring(filterId), LAF_BAGS, FilterSavedItemsForShop)
				end
            else
	        	if(not libFilters:IsFilterRegistered("FCOItemSaver_PlayerInventoryFilter" .. tostring(filterId))) then
					libFilters:RegisterFilter("FCOItemSaver_PlayerInventoryFilter" .. tostring(filterId), LAF_BAGS, FilterSavedItemsForShop)
				end
            end
        else
        	--Register all the filters in the player inventory
            if (newFilterMethod) then
	        	for i=1, locVars.gFCONumFilters, 1 do
					if(not libFilters:IsFilterRegistered("FCOItemSaver_PlayerInventoryFilterNew" .. tostring(panelId) .. "_" .. tostring(i))) then
						libFilters:RegisterFilter("FCOItemSaver_PlayerInventoryFilterNew" .. tostring(panelId) .. "_" .. tostring(i), LAF_BAGS, FilterSavedItemsForShop)
					end
	            end
            else
	        	for i=1, locVars.gFCONumFilters, 1 do
					if(not libFilters:IsFilterRegistered("FCOItemSaver_PlayerInventoryFilter" .. tostring(i))) then
						libFilters:RegisterFilter("FCOItemSaver_PlayerInventoryFilter" .. tostring(i), LAF_BAGS, FilterSavedItemsForShop)
					end
	            end
            end
        end

    else
    --Filtering inside inventory is NOT enabled in the settings
		if (newFilterMethod == true) then
			if (filterId ~= -1) then
				libFilters:UnregisterFilter("FCOItemSaver_PlayerInventoryFilterNew" .. tostring(panelId) .. "_" .. tostring(filterId), LAF_BAGS)
	        else
				for i=1, locVars.gFCONumFilters, 1 do
					libFilters:UnregisterFilter("FCOItemSaver_PlayerInventoryFilterNew" .. tostring(panelId) .. "_" .. tostring(i), LAF_BAGS)
	            end
	        end
        else
			if (filterId ~= -1) then
				libFilters:UnregisterFilter("FCOItemSaver_PlayerInventoryFilter" .. tostring(filterId))
	        else
				for i=1, locVars.gFCONumFilters, 1 do
					libFilters:UnregisterFilter("FCOItemSaver_PlayerInventoryFilter" .. tostring(i))
	            end
	        end
        end
    end
end

--Unregister all filters
local function unregisterFilters(filterId, onlyPlayerInvFilter, filterPanelId)
	--Only remove filters for player inventory?
    if (onlyPlayerInvFilter == nil) then
    	onlyPlayerInvFilter = false
    end
	--Only update a special panel, or all?
	local forVar, maxVar
	if (filterPanelId ~= nil) then
    	forVar  = filterPanelId
        maxVar  = filterPanelId
		debugMessage("[UnregisterFilters] FilterPanelId: " .. tostring(filterPanelId) .. ", filterId: " .. tostring(filterId) .. ", OnlyPlayerInvFilter: " .. tostring(onlyPlayerInvFilter), false)
    else
    	forVar  = 1
        maxVar  = locVars.gFCONumFilterInventoryTypes
		debugMessage("[UnregisterFilters] From panel Id: " .. tostring(forVar) .. ", To panel Id: " .. tostring(maxVar) .. ", filterId: " .. tostring(filterId) .. ", OnlyPlayerInvFilter: " .. tostring(onlyPlayerInvFilter), false)
    end

    local unregisterArrayNew = {}
    --Unregister only 1 filter ID?
	if (filterId ~= nil and filterId ~= -1) then
    	if (settings.splitFilters == true) then
			--New filter method
			for lFilterWhere=forVar, maxVar , 1 do
				if (onlyPlayerInvFilter == true) then
					libFilters:UnregisterFilter("FCOItemSaver_PlayerInventoryFilterNew" .. tostring(lFilterWhere) .. "_" .. tostring(filterId))
	            else
                	unregisterArrayNew = {
                       	[INV_BAGS] = "FCOItemSaver_PlayerInventoryFilterNew" .. tostring(lFilterWhere) .. "_" .. tostring(filterId),
	                    [INV_BANK] = "FCOItemSaver_PlayerBankFilterNew" .. tostring(lFilterWhere) .. "_" .. tostring(filterId),
						[INV_GUILDBANK] = "FCOItemSaver_GuildBankFilterNew" .. tostring(lFilterWhere) .. "_" .. tostring(filterId),
						[INV_STORE] = "FCOItemSaver_VendorFilterNew" .. tostring(lFilterWhere) .. "_" .. tostring(filterId),
						[INV_DECONSTRUCTION] = "FCOItemSaver_DeconstructionFilterNew" .. tostring(lFilterWhere) .. "_" .. tostring(filterId),
						[INV_IMPROVEMENT] = "FCOItemSaver_ImprovementFilterNew" .. tostring(lFilterWhere) .. "_" .. tostring(filterId),
						[INV_GUILDSTORE] = "FCOItemSaver_GuildStoreFilterNew" .. tostring(lFilterWhere) .. "_" .. tostring(filterId),
						[INV_MAIL] = "FCOItemSaver_MailFilterNew" .. tostring(lFilterWhere) .. "_" .. tostring(filterId),
						[INV_TRADE] = "FCOItemSaver_Player2PlayerFilterNew" .. tostring(lFilterWhere) .. "_" .. tostring(filterId),
						[INV_ENCHANTING_EXTRACTION] = "FCOItemSaver_EnchantingExtractionFilterNew" .. tostring(lFilterWhere) .. "_" .. tostring(filterId),
						[INV_ENCHANTING_CREATION] = "FCOItemSaver_EnchantingCreationFilterNew" .. tostring(lFilterWhere) .. "_" .. tostring(filterId),
					}
                	--Unregister the registered filters for each panel
					libFilters:UnregisterFilter(unregisterArrayNew[lFilterWhere], lFilterWhere)
				end
            end

		else -- if (settings.splitFilters == true) then
			--Old filter method
			if (onlyPlayerInvFilter == true) then
				libFilters:UnregisterFilter("FCOItemSaver_PlayerInventoryFilter" .. tostring(filterId))
            else
				--Unregister the registered filters
				libFilters:UnregisterFilter("FCOItemSaver_VendorFilter" .. tostring(filterId))
				libFilters:UnregisterFilter("FCOItemSaver_DeconstructionFilter" .. tostring(filterId))
				libFilters:UnregisterFilter("FCOItemSaver_ImprovementFilter" .. tostring(filterId))
				libFilters:UnregisterFilter("FCOItemSaver_GuildStoreFilter" .. tostring(filterId))
				libFilters:UnregisterFilter("FCOItemSaver_GuildBankFilter" .. tostring(filterId))
				libFilters:UnregisterFilter("FCOItemSaver_PlayerBankFilter" .. tostring(filterId))
				libFilters:UnregisterFilter("FCOItemSaver_PlayerInventoryFilter" .. tostring(filterId))
				libFilters:UnregisterFilter("FCOItemSaver_Player2PlayerFilter" .. tostring(filterId))
				libFilters:UnregisterFilter("FCOItemSaver_MailFilter" .. tostring(filterId))
				libFilters:UnregisterFilter("FCOItemSaver_EnchantingCreationFilter" .. tostring(filterId))
				libFilters:UnregisterFilter("FCOItemSaver_EnchantingExtractionFilter" .. tostring(filterId))
			end
        end

    else

	-- Unregister all filter IDs
		if (settings.splitFilters == true or locVars.gSplitFilterOverride == true) then
			--New filter method
			for filterId=1, locVars.gFCONumFilters, 1 do
				for lFilterWhere=forVar, maxVar , 1 do
					if (onlyPlayerInvFilter == true) then
						libFilters:UnregisterFilter("FCOItemSaver_PlayerInventoryFilterNew" .. tostring(lFilterWhere) .. "_" .. tostring(filterId))
		            else
	                	unregisterArrayNew = {
                        	[INV_BAGS] = "FCOItemSaver_PlayerInventoryFilterNew" .. tostring(lFilterWhere) .. "_" .. tostring(filterId),
		                    [INV_BANK] = "FCOItemSaver_PlayerBankFilterNew" .. tostring(lFilterWhere) .. "_" .. tostring(filterId),
							[INV_GUILDBANK] = "FCOItemSaver_GuildBankFilterNew" .. tostring(lFilterWhere) .. "_" .. tostring(filterId),
							[INV_STORE] = "FCOItemSaver_VendorFilterNew" .. tostring(lFilterWhere) .. "_" .. tostring(filterId),
							[INV_DECONSTRUCTION] = "FCOItemSaver_DeconstructionFilterNew" .. tostring(lFilterWhere) .. "_" .. tostring(filterId),
							[INV_IMPROVEMENT] = "FCOItemSaver_ImprovementFilterNew" .. tostring(lFilterWhere) .. "_" .. tostring(filterId),
							[INV_GUILDSTORE] = "FCOItemSaver_GuildStoreFilterNew" .. tostring(lFilterWhere) .. "_" .. tostring(filterId),
							[INV_MAIL] = "FCOItemSaver_MailFilterNew" .. tostring(lFilterWhere) .. "_" .. tostring(filterId),
							[INV_TRADE] = "FCOItemSaver_Player2PlayerFilterNew" .. tostring(lFilterWhere) .. "_" .. tostring(filterId),
							[INV_ENCHANTING_EXTRACTION] = "FCOItemSaver_EnchantingExtractionFilterNew" .. tostring(lFilterWhere) .. "_" .. tostring(filterId),
							[INV_ENCHANTING_CREATION] = "FCOItemSaver_EnchantingCreationFilterNew" .. tostring(lFilterWhere) .. "_" .. tostring(filterId),
						}
	                	--Unregister the registered filters for each panel
						libFilters:UnregisterFilter(unregisterArrayNew[lFilterWhere], lFilterWhere)
					end
		        end
	        end

        elseif (settings.splitFilters == false or locVars.gSplitFilterOverride == true) then
  			--Old filter method
        	for filterId=1, locVars.gFCONumFilters, 1 do
				if (onlyPlayerInvFilter == true) then
					libFilters:UnregisterFilter("FCOItemSaver_PlayerInventoryFilter" .. tostring(filterId))
	            else
					--Unregister the registered filters
					libFilters:UnregisterFilter("FCOItemSaver_VendorFilter" .. tostring(filterId))
					libFilters:UnregisterFilter("FCOItemSaver_DeconstructionFilter" .. tostring(filterId))
					libFilters:UnregisterFilter("FCOItemSaver_GuildStoreFilter" .. tostring(filterId))
					libFilters:UnregisterFilter("FCOItemSaver_GuildBankFilter" .. tostring(filterId))
					libFilters:UnregisterFilter("FCOItemSaver_PlayerBankFilter" .. tostring(filterId))
					libFilters:UnregisterFilter("FCOItemSaver_PlayerInventoryFilter" .. tostring(filterId))
					libFilters:UnregisterFilter("FCOItemSaver_Player2PlayerFilter" .. tostring(filterId))
					libFilters:UnregisterFilter("FCOItemSaver_MailFilter" .. tostring(filterId))
					libFilters:UnregisterFilter("FCOItemSaver_EnchantingCreationFilter" .. tostring(filterId))
					libFilters:UnregisterFilter("FCOItemSaver_EnchantingExtractionFilter" .. tostring(filterId))
				end
	        end
		end
	end
end

--Helper function for method registerFilters when settings.splitFilters == true
local function registerFilterId(p_onlyPlayerInvFilter, p_filterId, p_panelId)
	--Get the current LAF (filter type, e.g. LAF_BAGS, LAF_BANK, etc.)
	--local laf = libFilters:GetCurrentLAF()
    --if laf == nil then laf = LAF_BAGS end

	if (p_onlyPlayerInvFilter == true) then
		--Player inventory -> Only if activated in settings
		FilterPlayerInventory(p_filterId, p_panelId)
	else
		--(Re)register the filters for each panel
		if     (p_panelId == INV_BAGS) then
			FilterPlayerInventory(p_filterId, p_panelId)

		elseif (p_panelId == INV_BANK) then
        	-- Filter player bank panels
			if (settings.allowBankFilter == true) then
				if(not libFilters:IsFilterRegistered("FCOItemSaver_PlayerBankFilterNew" .. tostring(p_panelId) .. "_" .. tostring(p_filterId))) then
					libFilters:RegisterFilter("FCOItemSaver_PlayerBankFilterNew" .. tostring(p_panelId) .. "_" .. tostring(p_filterId), LAF_BANK, FilterSavedItemsForShop)
				end
			end

		elseif (p_panelId == INV_GUILDBANK) then
		-- Filter guildbank panels
			if (settings.allowGuildBankFilter == true) then
				if(not libFilters:IsFilterRegistered("FCOItemSaver_GuildBankFilterNew" .. tostring(p_panelId) .. "_" .. tostring(p_filterId))) then
					libFilters:RegisterFilter("FCOItemSaver_GuildBankFilterNew" .. tostring(p_panelId) .. "_" .. tostring(p_filterId), LAF_GUILDBANK, FilterSavedItemsForShop)
				end
			end

		elseif (p_panelId == INV_STORE) then
		-- Filter stores
			if (settings.allowVendorFilter == true) then
				if(not libFilters:IsFilterRegistered("FCOItemSaver_VendorFilterNew" .. tostring(p_panelId) .. "_" .. tostring(p_filterId))) then
					-- Register the before chosen filter type and function
					libFilters:RegisterFilter("FCOItemSaver_VendorFilterNew" .. tostring(p_panelId) .. "_" .. tostring(p_filterId), LAF_STORE, FilterSavedItemsForShop)
				end
			end
		elseif (p_panelId == INV_DECONSTRUCTION) then
		-- Filter deconstruction panels
			if (settings.allowDeconstructionFilter == true) then
				if(not libFilters:IsFilterRegistered("FCOItemSaver_DeconstructionFilterNew" .. tostring(p_panelId) .. "_" .. tostring(p_filterId))) then
					libFilters:RegisterFilter("FCOItemSaver_DeconstructionFilterNew" .. tostring(p_panelId) .. "_" .. tostring(p_filterId), LAF_DECONSTRUCTION, FilterSavedItems)
				end
			end
		elseif (p_panelId == INV_IMPROVEMENT) then
		-- Filter improvement panels
			if (settings.allowImprovementFilter == true) then
				if(not libFilters:IsFilterRegistered("FCOItemSaver_ImprovementFilterNew" .. tostring(p_panelId) .. "_" .. tostring(p_filterId))) then
					libFilters:RegisterFilter("FCOItemSaver_ImprovementFilterNew" .. tostring(p_panelId) .. "_" .. tostring(p_filterId), LAF_IMPROVEMENT, FilterSavedItems)
				end
            end

		elseif (p_panelId == INV_GUILDSTORE) then
		-- Filter guildstore panels
			if (settings.allowTradinghouseFilter == true) then
				if(not libFilters:IsFilterRegistered("FCOItemSaver_GuildStoreFilterNew" .. tostring(p_panelId) .. "_" .. tostring(p_filterId))) then
					libFilters:RegisterFilter("FCOItemSaver_GuildStoreFilterNew" .. tostring(p_panelId) .. "_" .. tostring(p_filterId), LAF_GUILDSTORE, FilterSavedItemsForShop)
				end
			end

		elseif (p_panelId == INV_MAIL) then
		--Mail
			if (settings.allowMailFilter == true) then
				if(not libFilters:IsFilterRegistered("FCOItemSaver_MailFilterNew" .. tostring(p_panelId) .. "_" .. tostring(p_filterId))) then
					libFilters:RegisterFilter("FCOItemSaver_MailFilterNew" .. tostring(p_panelId) .. "_" .. tostring(p_filterId), LAF_MAIL, FilterSavedItemsForShop)
				end
			end

		elseif (p_panelId == INV_TRADE) then
			--Player2player trading
			if (settings.allowTradeFilter == true) then
				if(not libFilters:IsFilterRegistered("FCOItemSaver_Player2PlayerFilterNew" .. tostring(p_panelId) .. "_" .. tostring(p_filterId))) then
					libFilters:RegisterFilter("FCOItemSaver_Player2PlayerFilterNew" .. tostring(p_panelId) .. "_" .. tostring(p_filterId), LAF_TRADE, FilterSavedItemsForShop)
				end
			end

		elseif (p_panelId == INV_ENCHANTING_EXTRACTION) then
		-- Filter enchanting panels
			if (settings.allowEnchantingFilter == true) then
				if(not libFilters:IsFilterRegistered("FCOItemSaver_EnchantingExtractionFilterNew" .. tostring(filterId))) then
					libFilters:RegisterFilter("FCOItemSaver_EnchantingExtractionFilterNew" .. tostring(p_panelId) .. "_" .. tostring(p_filterId), LAF_ENCHANTING_EXTRACTION, FilterSavedItems)
				end
			end

		elseif (p_panelId == INV_ENCHANTING_CREATION) then
		-- Filter enchanting panels
			if (settings.allowEnchantingFilter == true) then
				if(not libFilters:IsFilterRegistered("FCOItemSaver_EnchantingCreationFilterNew" .. tostring(p_panelId) .. "_" .. tostring(p_filterId))) then
					libFilters:RegisterFilter("FCOItemSaver_EnchantingCreationFilterNew" .. tostring(p_panelId) .. "_" .. tostring(p_filterId), LAF_ENCHANTING_CREATION, FilterSavedItems)
				end
			end

		end

	end
end

--Register the filters by help of library libFilters
local function registerFilters(filterId, onlyPlayerInvFilter, p_FilterPanelId)
	--Only register filters for player inventory?
    if (onlyPlayerInvFilter == nil) then
    	onlyPlayerInvFilter = false
    end

	--Register only 1 filter ID?
	if (filterId ~= nil and filterId ~= -1) then
		--Register only one filter ID

		--New or old behaviour of filtering?
		if (settings.splitFilters == true) then
			--New filtering using panels
			debugMessage("[RegisterFilters] Panel: " .. tostring(p_FilterPanelId) .. ", OnlyPlayerInv: " .. tostring(onlyPlayerInvFilter) .. ", filterId: " .. tostring(filterId), false)

			--Register the filter for the given panel ID and filter ID
			registerFilterId(onlyPlayerInvFilter, filterId, p_FilterPanelId)

		else
	        -- Old filter behaviour without panels
			debugMessage("[RegisterFilters] filterId: " .. tostring(filterId) .. ", OnlyPlayerInv: " .. tostring(onlyPlayerInvFilter), false)

	        if (onlyPlayerInvFilter == true) then
				--Player inventory -> Only if activated in settings
				FilterPlayerInventory(filterId)
	        else

	        	-- Filter stores
	            if (settings.allowVendorFilter == true) then
					if(not libFilters:IsFilterRegistered("FCOItemSaver_VendorFilter" .. tostring(filterId))) then
			   			-- Register the before chosen filter type and function
		                libFilters:RegisterFilter("FCOItemSaver_VendorFilter" .. tostring(filterId), LAF_STORE, FilterSavedItemsForShop)
		            end
				end

				-- Filter deconstruction panels
	            if (settings.allowDeconstructionFilter == true) then
					if(not libFilters:IsFilterRegistered("FCOItemSaver_DeconstructionFilter" .. tostring(filterId))) then
		                libFilters:RegisterFilter("FCOItemSaver_DeconstructionFilter" .. tostring(filterId), LAF_DECONSTRUCTION, FilterSavedItems)
		            end
				end

				-- Filter improvement panels
	            if (settings.allowImprovementFilter == true) then
					if(not libFilters:IsFilterRegistered("FCOItemSaver_ImprovementFilter" .. tostring(filterId))) then
		                libFilters:RegisterFilter("FCOItemSaver_ImprovementFilter" .. tostring(filterId), LAF_IMPROVEMENT, FilterSavedItems)
		            end
				end

				-- Filter enchanting panels
	            if (settings.allowEnchantingFilter == true) then
                    if      ENCHANTING.enchantingMode == 1 then
	                	if(not libFilters:IsFilterRegistered("FCOItemSaver_EnchantingCreationFilter" .. tostring(filterId))) then
			                libFilters:RegisterFilter("FCOItemSaver_EnchantingCreationFilter" .. tostring(filterId), LAF_ENCHANTING_CREATION, FilterSavedItems)
			            end
                    elseif ENCHANTING.enchantingMode == 2 then
						if(not libFilters:IsFilterRegistered("FCOItemSaver_EnchantingExtractionFilter" .. tostring(filterId))) then
			                libFilters:RegisterFilter("FCOItemSaver_EnchantingExtractionFilter" .. tostring(filterId), LAF_ENCHANTING_EXTRACTION, FilterSavedItems)
			            end
                   	end
				end

	            -- Filter player bank panels
	            if (settings.allowBankFilter == true) then
					if(not libFilters:IsFilterRegistered("FCOItemSaver_PlayerBankFilter" .. tostring(filterId))) then
		                libFilters:RegisterFilter("FCOItemSaver_PlayerBankFilter" .. tostring(filterId), LAF_BANK, FilterSavedItemsForShop)
	                end
				end

	 			-- Filter guildbank panels
	            if (settings.allowGuildBankFilter == true) then
					if(not libFilters:IsFilterRegistered("FCOItemSaver_GuildBankFilter" .. tostring(filterId))) then
		                libFilters:RegisterFilter("FCOItemSaver_GuildBankFilter" .. tostring(filterId), LAF_GUILDBANK, FilterSavedItemsForShop)
	                end
				end

				-- Filter guildstore panels
	            if (settings.allowTradinghouseFilter == true) then
					if(not libFilters:IsFilterRegistered("FCOItemSaver_GuildStoreFilter" .. tostring(filterId))) then
		                libFilters:RegisterFilter("FCOItemSaver_GuildStoreFilter" .. tostring(filterId), LAF_GUILDSTORE, FilterSavedItemsForShop)
					end
	            end

			    -- !!!!!												   !!!!!
			    -- !!!!! In addition always activate the following filters !!!!!
			    -- !!!!!												   !!!!!
				--Player inventory -> Only if activated in settings
				FilterPlayerInventory(filterId)

				--Player2player trading
		        if (settings.allowTradeFilter == true) then
					if(not libFilters:IsFilterRegistered("FCOItemSaver_Player2PlayerFilter" .. tostring(filterId))) then
						libFilters:RegisterFilter("FCOItemSaver_Player2PlayerFilter" .. tostring(filterId), LAF_TRADE, FilterSavedItemsForShop)
					end
				end
			    --Mail
		        if (settings.allowMailFilter == true) then
					if(not libFilters:IsFilterRegistered("FCOItemSaver_MailFilter" .. tostring(filterId))) then
						libFilters:RegisterFilter("FCOItemSaver_MailFilter" .. tostring(filterId), LAF_MAIL, FilterSavedItemsForShop)
					end
				end
            end
		end  --new filter method?

	else
		--Register all filter IDs

		if (settings.splitFilters == true) then
			--Using panels

			--Only update a special panel, or all?
			local forVar, maxVar
			if (p_FilterPanelId == nil) then
		    	forVar  = 1
		        maxVar  = locVars.gFCONumFilterInventoryTypes
		    end

			for filterId=1, locVars.gFCONumFilters, 1 do
				for lFilterWhere=forVar, maxVar , 1 do
					debugMessage("[registerFilters] Panel: " .. tostring(forVar) .. ", filterId: " .. tostring(filterId) .. ", OnlyPlayerInv: " .. tostring(onlyPlayerInvFilter), false)
                    --register the filters for the given panels
                	registerFilterId(onlyPlayerInvFilter, filterId, lFilterWhere)
		        end
	        end

		else
			--NOT using panels (old method)
			debugMessage("[RegisterFilters] All filters!, OnlyPlayerInv: " .. tostring(onlyPlayerInvFilter), false)

			for filterId=1, locVars.gFCONumFilters, 1 do
		        if (onlyPlayerInvFilter == true) then
					--Player inventory -> Only if activated in settings
					FilterPlayerInventory(filterId)
		        else

		        	-- Filter stores
		            if (settings.allowVendorFilter == true) then
						if(not libFilters:IsFilterRegistered("FCOItemSaver_VendorFilter" .. tostring(filterId))) then
				   			-- Register the before chosen filter type and function
			                libFilters:RegisterFilter("FCOItemSaver_VendorFilter" .. tostring(filterId), LAF_STORE, FilterSavedItemsForShop)
			            end
                    end

					-- Filter deconstruction panels
		            if (settings.allowDeconstructionFilter == true) then
						if(not libFilters:IsFilterRegistered("FCOItemSaver_DeconstructionFilter" .. tostring(filterId))) then
			                libFilters:RegisterFilter("FCOItemSaver_DeconstructionFilter" .. tostring(filterId), LAF_DECONSTRUCTION, FilterSavedItems)
			            end
                    end

					-- Filter improvement panels
		            if (settings.allowImprovementFilter == true) then
						if(not libFilters:IsFilterRegistered("FCOItemSaver_ImprovementFilter" .. tostring(filterId))) then
			                libFilters:RegisterFilter("FCOItemSaver_ImprovementFilter" .. tostring(filterId), LAF_IMPROVEMENT, FilterSavedItems)
			            end
                    end

					-- Filter enchanting panels
		            if (settings.allowEnchantingFilter == true) then
	                    if      ENCHANTING.enchantingMode == 1 then
		                	if(not libFilters:IsFilterRegistered("FCOItemSaver_EnchantingCreationFilter" .. tostring(filterId))) then
				                libFilters:RegisterFilter("FCOItemSaver_EnchantingCreationFilter" .. tostring(filterId), LAF_ENCHANTING_CREATION, FilterSavedItems)
				            end
	                    elseif ENCHANTING.enchantingMode == 2 then
							if(not libFilters:IsFilterRegistered("FCOItemSaver_EnchantingExtractionFilter" .. tostring(filterId))) then
				                libFilters:RegisterFilter("FCOItemSaver_EnchantingExtractionFilter" .. tostring(filterId), LAF_ENCHANTING_EXTRACTION, FilterSavedItems)
				            end
	                   	end
					end

		            -- Filter player bank panels
		            if (settings.allowBankFilter == true) then
						if(not libFilters:IsFilterRegistered("FCOItemSaver_PlayerBankFilter" .. tostring(filterId))) then
			                libFilters:RegisterFilter("FCOItemSaver_PlayerBankFilter" .. tostring(filterId), LAF_BANK, FilterSavedItemsForShop)
		                end
					end

		 			-- Filter guildbank panels
		            if (settings.allowGuildBankFilter == true) then
						if(not libFilters:IsFilterRegistered("FCOItemSaver_GuildBankFilter" .. tostring(filterId))) then
			                libFilters:RegisterFilter("FCOItemSaver_GuildBankFilter" .. tostring(filterId), LAF_GUILDBANK, FilterSavedItemsForShop)
		                end
					end

					-- Filter guildstore panels
		            if (settings.allowTradinghouseFilter == true) then
						if(not libFilters:IsFilterRegistered("FCOItemSaver_GuildStoreFilter" .. tostring(filterId))) then
			                libFilters:RegisterFilter("FCOItemSaver_GuildStoreFilter" .. tostring(filterId), LAF_GUILDSTORE, FilterSavedItemsForShop)
						end
		            end

				    -- !!!!!												   !!!!!
				    -- !!!!! In addition always activate the following filters !!!!!
				    -- !!!!!												   !!!!!
					--Player inventory -> Only if activated in settings
					FilterPlayerInventory(filterId)

					--Player2player trading
			        if (settings.allowTradeFilter == true) then
						if(not libFilters:IsFilterRegistered("FCOItemSaver_Player2PlayerFilter" .. tostring(filterId))) then
							libFilters:RegisterFilter("FCOItemSaver_Player2PlayerFilter" .. tostring(filterId), LAF_TRADE, FilterSavedItemsForShop)
						end
					end
				    --Mail
			        if (settings.allowMailFilter == true) then
						if(not libFilters:IsFilterRegistered("FCOItemSaver_MailFilter" .. tostring(filterId))) then
							libFilters:RegisterFilter("FCOItemSaver_MailFilter" .. tostring(filterId), LAF_MAIL, FilterSavedItemsForShop)
						end
					end
	            end
	        end
		end

    end -- only 1 filter ID or all filter IDs?

end

--Write the info about the current filter state into the chat
local function outputFilterState(p_outputToChat, p_newFilterMethod, p_panelId, p_filterId, p_stateText)
	local filterText
    local preChatText
    local outputText

    --Abort given because filter is/will be unregistered?
    if p_stateText == 'ABORT' then
    	filterText = "_"
        preChatText = locVars.preChatTextRed
    else
		--Not aborted
        filterText = tostring(p_filterId) .. "_"
        preChatText = locVars.preChatTextGreen
    end

    p_newFilterMethod = p_newFilterMethod or settings.splitFilters

	debugMessage("[OutputFilterState] OutputToChat: " .. tostring(p_outputToChat) ..", splitFilters: " ..tostring(p_newFilterMethod)..", filterPanelId: " ..tostring(p_panelId)..", filterId: "..tostring(p_filterId)..", stateText: " .. p_stateText, true)

	--New filter method activated?
	if (p_newFilterMethod == true) then
    	if     (p_panelId == INV_BAGS) then
			outputText = preChatText .. fco_loc["filter_inventory"] .. fco_loc["filter" .. filterText .. p_stateText]
		elseif (p_panelId == INV_GUILDBANK) then
		   outputText = preChatText .. fco_loc["filter_guildbank"] .. fco_loc["filter" .. filterText .. p_stateText]
		elseif (p_panelId == INV_GUILDSTORE) then
			outputText = preChatText .. fco_loc["filter_guildstore"] .. fco_loc["filter" .. filterText .. p_stateText]
		elseif (p_panelId == INV_BANK) then
			outputText = preChatText .. fco_loc["filter_bank"] .. fco_loc["filter" .. filterText .. p_stateText]
		elseif (p_panelId == INV_DECONSTRUCTION) then
			outputText = preChatText .. fco_loc["filter_deconstruction"] .. fco_loc["filter" .. filterText .. p_stateText]
		elseif (p_panelId == INV_IMPROVEMENT) then
			outputText = preChatText .. fco_loc["filter_improvement"] .. fco_loc["filter" .. filterText .. p_stateText]
		elseif (p_panelId == INV_ENCHANTING_EXTRACTION) then
			outputText = preChatText .. fco_loc["filter_enchantingstation_extraction"] .. fco_loc["filter" .. filterText .. p_stateText]
		elseif (p_panelId == INV_ENCHANTING_CREATION) then
			outputText = preChatText .. fco_loc["filter_enchantingstation_creation"] .. fco_loc["filter" .. filterText .. p_stateText]
		elseif (p_panelId == INV_STORE) then
			outputText = preChatText .. fco_loc["filter_store"] .. fco_loc["filter" .. filterText .. p_stateText]
		elseif (p_panelId == INV_MAIL) then
			outputText = preChatText .. fco_loc["filter_mail"] .. fco_loc["filter" .. filterText .. p_stateText]
		elseif (p_panelId == INV_TRADE) then
			outputText = preChatText .. fco_loc["filter_trade"] .. fco_loc["filter" .. filterText .. p_stateText]
		end

	--Old filter method activated?
	else
		outputText = preChatText .. fco_loc["filter" .. filterText .. p_stateText]
	end

    --Output to chat or return text only?
    if p_outputToChat == true then
    	d(outputText)
    else
    	return outputText
    end
end

--Get the "real" active panel.
--If you are at the bank e.g. panelId is 2 (locVars.gFilterWhere,set in event BANK_OPEN), but you could also be at the deposit tab which uses
--the normal inventory filters of panelId = 1. The same applies for mail, trade, and others
local function checkActivePanel(comingFrom)
	local inventoryName
	local oldFilterWhere
	if comingFrom == 0 or comingFrom == nil then
	    --Get the current filter panel id
	    oldFilterWhere = locVars.gFilterWhere
    else
	    oldFilterWhere = comingFrom
    end

	--Player bank
    if     not BANK:IsHidden() or comingFrom == INV_BANK then
		--Update the filterPanelId
		locVars.gFilterWhere = getFilterWhereBySettings(INV_BANK)
        inventoryName = BANK_INV
	--Guild bank
    elseif not GUILD_BANK:IsHidden() or comingFrom == INV_GUILDBANK then
		--Update the filterPanelId
		locVars.gFilterWhere = getFilterWhereBySettings(INV_GUILDBANK)
        inventoryName = GUILD_BANK_INV
	--Guild store
    elseif not GUILD_STORE:IsHidden() or comingFrom == INV_GUILDSTORE then
		--Update the filterPanelId
		locVars.gFilterWhere = getFilterWhereBySettings(INV_GUILDSTORE)
        inventoryName = INV
	--Vendor
    elseif not VENDOR_SELL:IsHidden() or comingFrom == INV_STORE then
		--Update the filterPanelId
		locVars.gFilterWhere = getFilterWhereBySettings(INV_STORE)
        --inventoryName = VENDOR_SELL
        inventoryName = INV
	--Mail
    elseif not MAIL_SEND:IsHidden() or comingFrom == INV_MAIL then
		--Update the filterPanelId
		locVars.gFilterWhere = getFilterWhereBySettings(INV_MAIL)
        --inventoryName = MAIL_SEND
        inventoryName = INV
	--Trade
    elseif not PLAYER_TRADE:IsHidden() or comingFrom == INV_TRADE then
		--Update the filterPanelId
		locVars.gFilterWhere = getFilterWhereBySettings(INV_TRADE)
        --inventoryName = PLAYER_TRADE
        inventoryName = INV
    --Enchanting creation mode
    elseif comingFrom == INV_ENCHANTING_CREATION then
		--Update the filterPanelId
		locVars.gFilterWhere = getFilterWhereBySettings(INV_ENCHANTING_CREATION)
        inventoryName = ENCHANTING_STATION
	--Enchanting extraction mode
    elseif comingFrom == INV_ENCHANTING_EXTRACTION then
		--Update the filterPanelId
		locVars.gFilterWhere = getFilterWhereBySettings(INV_ENCHANTING_EXTRACTION)
        inventoryName = ENCHANTING_STATION
--[[
    --Enchanting creation mode
    elseif ENCHANTING.enchantingMode == 1 then
		--Update the filterPanelId
		locVars.gFilterWhere = getFilterWhereBySettings(INV_ENCHANTING_CREATION)
        inventoryName = ENCHANTING_STATION
	--Enchanting extraction mode
    elseif ENCHANTING.enchantingMode == 2 then
		--Update the filterPanelId
		locVars.gFilterWhere = getFilterWhereBySettings(INV_ENCHANTING_EXTRACTION)
        inventoryName = ENCHANTING_STATION
]]
   	--Deconstruction
    elseif comingFrom == INV_DECONSTRUCTION then
    	--Update the filterPanelId
		locVars.gFilterWhere = getFilterWhereBySettings(INV_DECONSTRUCTION)
        inventoryName = DECONSTRUCTION_INV
       --Improvement
    elseif comingFrom == INV_IMPROVEMENT then
		--Update the filterPanelId
		locVars.gFilterWhere = getFilterWhereBySettings(INV_IMPROVEMENT)
        inventoryName = IMPROVEMENT_INV
	--Deconstruction
    elseif not DECONSTRUCTION_INV:IsHidden() then
    	--Update the filterPanelId
		locVars.gFilterWhere = getFilterWhereBySettings(INV_DECONSTRUCTION)
        inventoryName = DECONSTRUCTION_INV
    --Improvement
    elseif not IMPROVEMENT_INV:IsHidden() then
		--Update the filterPanelId
		locVars.gFilterWhere = getFilterWhereBySettings(INV_IMPROVEMENT)
        inventoryName = IMPROVEMENT_INV
	--Player inventory
    elseif not INV:IsHidden() then
		--Update the filterPanelId
		locVars.gFilterWhere = getFilterWhereBySettings(INV_BAGS)
        inventoryName = INV
    else
		--Update the filterPanelId with a standard value
		locVars.gFilterWhere = getFilterWhereBySettings(INV_BAGS)
		inventoryName = BAGS_INV
    end

    debugMessage("[CheckActivePanel] Before: " .. tostring(oldFilterWhere) .. ", after: " .. tostring(locVars.gFilterWhere), true)

	--Return the found inventory type (e.g. INV_BANK)
 	return inventoryName
end

--Enable/Disable a filter/all filters
local function doFilter(onoff, p_button, filterId, beQuiet, doFilterBasicsPlayer, doUpdateButtonColors, onlyPlayerInvFilter, p_FilterPanelId)
	--Set the last used filter Id
	locVars.gLastFilterId = filterId

	--Hide the button contextMenu if shown
	hideContextMenu()

	--Check if the current filter panel Id is given
    if (p_FilterPanelId == nil) then
		--For initialization, called from function EnableFilters()
		if (onoff == -100) then
        	p_FilterPanelId = INV_BAGS
	    else
	    	p_FilterPanelId = locVars.gFilterWhere
        end
    end
    --Check if the settings are enabled for the current panel
    p_FilterPanelId = getFilterWhereBySettings(p_FilterPanelId)

    --Only perform the check if we are not initializing
    if (onoff ~= -100) then
        --Check if we are in the player inventory
        if ( p_FilterPanelId == INV_BAGS ) then
        	--we are in the player inventory (or in the banks at the deposit inventories, or at mail sending, or trading)
            onlyPlayerInvFilter 	= true
            doFilterBasicsPlayer	= true
        end
    end

	--============================================================================
	-- ABORT HERE IF the settings for the current filter panel Id is not enabled
	--============================================================================
	if settings.atPanelEnabled[p_FilterPanelId]["filters"] == false then
        --Unregister the filter
    	unregisterFilters(filterId, onlyPlayerInvFilter, p_FilterPanelId)

		debugMessage("doFilter: !!! ABORT !!! " .. onoff .. ", filterId: " ..filterId .. ", FilterPanelId: " .. tostring(p_FilterPanelId) .. ", beQuiet: " .. tostring(beQuiet), false)

		--Give chat output if beQuiet is false
        if beQuiet == false then
        	outputFilterState(true, newFilterMethod, p_FilterPanelId, filterId, 'ABORT')
        end

		--Abort function here now
    	return
    end
	--============================================================================

	debugMessage("doFilter: " .. onoff .. ", filterId: " ..filterId .. ", beQuiet: " .. tostring(beQuiet) .. ", doFilterBasiscsPlayer: " .. tostring(doFilterBasicsPlayer) .. ", doUpdateButtonColors: " .. tostring(doUpdateButtonColors) .. ", onlyPlayerInvFilter: " .. tostring(onlyPlayerInvFilter) .. ", FilterPanelId: " .. tostring(p_FilterPanelId), false)

    --Fallback solution if filterPanelId is still empty
    if ( settings.splitFilters == true and (p_FilterPanelId == nil or p_FilterPanelId == 0) ) then
        d("[FCOIS] doFilter("..tostring(filterId).."): No filter panel ID given!")
    	return
    end

    local isFilterActive
	-- change the filter value in the settings
    if (onoff == 1) then
		isFilterActive = setSettingsIsFilterOn(filterId, true, p_FilterPanelId)
	elseif (onoff == 2) then
		isFilterActive = setSettingsIsFilterOn(filterId, false, p_FilterPanelId)
	elseif (onoff == -99) then
		isFilterActive = setSettingsIsFilterOn(filterId, -99, p_FilterPanelId)
    else
		-- Should the filter be changed to next state?
    	if (onoff == -1) then
            isFilterActive = getSettingsIsFilterOn(filterId, p_FilterPanelId)

        	--Filter is on? Turn it off
			if (isFilterActive == true) then
				isFilterActive = setSettingsIsFilterOn(filterId, false, p_FilterPanelId)
			--Filter is off? Only show filtered
			elseif (isFilterActive == false) then
				isFilterActive = setSettingsIsFilterOn(filterId, -99, p_FilterPanelId)
			--Filter only shows filtered? Turn it on
			else
				isFilterActive = setSettingsIsFilterOn(filterId, true, p_FilterPanelId)
			end
		--elseif (onoff == -100) then
            --For initialization (onoff = -100) the filter will be kept as read from the settings
        end
    end

	--Register / Unregister the libFilter filters
    local registerFiltersNow = false

   	--Are we using the old (only 1 panel) or new (different panels) filtering?
   	local newFilterMethod = settings.splitFilters

--------------------------------------------------------------------------------
-- Initializing from function EnableFilters() at e.g. addon loading
--------------------------------------------------------------------------------
	--Are we initializing from function Enablefilters() ?
	if (onoff == -100) then

		--Unregister all old filters if the addon is already loaded and filters have been registered before.
        --This happens only by function Enablefilters() called after addon has been fully loaded (e.g. the settings menu "Split filters")
        if addonVars.gAddonLoaded == true then
			--Did we change the settings "Split filters" and have to unregister all old "split filter" panel filters?
            if locVars.gSplitFilterOverride == true then
            	unregisterFilters(-1)
			else
            	unregisterFilters(filterId)
			end
        end

		--We are using filters with splitted panels
        if (newFilterMethod == true) then
			--Only update panel INV_BAGS (player inventory)
            local panels = INV_BAGS

			--Check for each panel if filters are enabled
--[[        for panels=1, locVars.gFCONumFilterInventoryTypes, 1 do

				if   (panels==INV_BAGS or panels==INV_BANK or panels==INV_STORE
	            	 or panels==INV_GUILDBANK or panels==INV_GUILDSTORE
	                 or panels==INV_DECONSTRUCTION or panels==INV_IMPROVEMENT
	                 or panels==INV_ENCHANTING_EXTRACTION or panels==INV_ENCHANTING_CREATION
	                 or panels==INV_MAIL or panels==INV_TRADE
	            	) then
]]
	                isFilterActive = nil
	                isFilterActive = getSettingsIsFilterOn(filterId, panels)

					if    (isFilterActive == true) then
		                --Filter is ON
						-- Set the filters to be registered
				        registerFiltersNow = true
	                elseif(isFilterActive == false) then
		                --Filter is OFF
						-- Set the filters to stay unregistered
				        registerFiltersNow = false
	                else
	                    --Filter is ONLY SHOWING MARKED ITEMS (-99)
						-- Set the filters to be registered again
				        registerFiltersNow = true
	               	end

			        --Register the filters now
			        if (registerFiltersNow == true) then
				     	registerFilters(filterId, onlyPlayerInvFilter, panels)
			        end

					-- Update the colors of the 4 "player inventory" filter buttons. All others will be updated upon opening (on event)
				    if (filterId == 1 and panels == INV_BAGS and doUpdateButtonColors == true) then
				    	UpdateButtonColors(-1, nil, -1, INV_BAGS)
				    end

--            	end	-- if panels == INV_BAGS, INV_BANK, ...

--	        end -- for panels...

		else -- if (newFilterMethod == true) then

			--We are using filters for only 1 panel
			isFilterActive = nil
			isFilterActive = getSettingsIsFilterOn(filterId)

			if    (isFilterActive == true) then
				--Filter is ON
				-- Set the filters to be registered
				registerFiltersNow = true
			elseif(isFilterActive == false) then
				--Filter is OFF
				-- Set the filters to stay unregistered
				registerFiltersNow = false
			else
				--Filter is ONLY SHOWING MARKED ITEMS (-99)
				-- Set the filters to be registered again
				registerFiltersNow = true
			end

			--Register the filters now
			if (registerFiltersNow == true) then
				registerFilters(filterId, onlyPlayerInvFilter)
			end

			-- Update the colors of the 4 "player inventory" filter buttons. All others will be updated upon opening (on event)
			if (filterId == 1 and doUpdateButtonColors == true) then
				UpdateButtonColors(-1, nil, -1)
			end

        end


--------------------------------------------------------------------------------
-- Responding to a filter button OnClicked() event or a chat command
--------------------------------------------------------------------------------
	else -- if (onoff == -100) then


	--We are not initializing -> Check filter state
        --The NEW filter status was set by function setSettingsIsFilterOn() at the beginning of this function
        --be determined here once again
        if isFilterActive == nil then
			isFilterActive = getSettingsIsFilterOn(filterId, p_FilterPanelId)
			--Is the new filter status still not set initialize it with "false"
            if isFilterActive == nil then
				isFilterActive = setSettingsIsFilterOn(filterId, false, p_FilterPanelId)
            end
        end

		-- Filter is "OFF"
	    if(isFilterActive == false) then

	    	-- Output "filter off" text
			if (settings.deepDebug or (beQuiet == false and settings.showFilterStatusInChat == true)) then
				outputFilterState(true, newFilterMethod, p_FilterPanelId, filterId, 'off')
	        end

			-- Set the filters to stay unregistered
	        registerFiltersNow = false

            -- Unregister all old filters for the given filterId and panel
       		unregisterFilters(filterId, onlyPlayerInvFilter, p_FilterPanelId)

		-- Filter is "ON"
		elseif (isFilterActive == true) then

			-- Output "filter on" text
	        if (settings.deepDebug or (beQuiet == false and settings.showFilterStatusInChat == true)) then
				outputFilterState(true, newFilterMethod, p_FilterPanelId, filterId, 'on')
	        end

            -- Unregister all old filters for the given filterId and panel
       		unregisterFilters(filterId, onlyPlayerInvFilter, p_FilterPanelId)

			-- Set the filters to be registered
	        registerFiltersNow = true

		--Filter got value "-99"
		--Special treatment for filter to show only marked items
	    else

			-- Output "show only marked" text
	        if (settings.deepDebug or (beQuiet == false and settings.showFilterStatusInChat == true)) then
				outputFilterState(true, newFilterMethod, p_FilterPanelId, filterId, 'onlyfiltered')
	        end

            -- Unregister all old filters for the given filterId and panel
       		unregisterFilters(filterId, onlyPlayerInvFilter, p_FilterPanelId)

			-- Set the filters to be registered again
	        registerFiltersNow = true

	    end

--=====================================================================================================
		--Register filters now?
		if(registerFiltersNow == true) then
			--(Re)register the filter (for the given panel again)
			registerFilters(filterId, onlyPlayerInvFilter, p_FilterPanelId)
		end

    end -- if (onoff == -100) then

    --Refresh the inventories scroll lists (the inventories itsself are updated within libFilters)
    --Only updte if button was clicked manually or this is the last call to this function dofilter()
    --from function enableFilters() (at initialization of this addon e.g.)
    if ( onoff ~= -100 or (onoff == -100 and filterId == locVars.gFCONumFilters) ) then
		--Update all inventories (false) / only the player inventory (true)
		FilterBasics(doFilterBasicsPlayer)
    end

	--Update the colors of the changed buttons inside the inventory panels, but only
    --if we are not coming from addon initialization
    if (onoff ~= - 100 and doUpdateButtonColors == true) then
    	UpdateButtonColors(filterId, nil, isFilterActive, p_FilterPanelId)
    end

end

-- Scan inventory for ornate items and mark them as "to sell"
local function scanInventoryForOrnate()
	if (settings.autoMarkOrnate == true and preventerVars.gScanningInv == false) then
    	preventerVars.gScanningInv = true

		local atLeastOneOrnateFound = false
	    local bagId = BAG_BACKPACK
		local itemData
        local itemId
        --for each item in your player inventory
		--for slotIndex = 1, GetBagSize(bagId) do
        for slotIndex = 0, GetBagSize(bagId) - 1 do
			--Initialize variables again
			itemId = nil
			itemData = nil
			itemId = SignItemId(GetItemInstanceId(bagId, slotIndex))
            if (markedItems[5][itemId] == false or markedItems[5][itemId] == nil) then
			    if (isItemOrnate(bagId, slotIndex)) then
	        		itemData = PLAYER_INVENTORY.inventories[INVENTORY_BACKPACK].slots[slotIndex]
			    	if (itemData ~= nil) then
					   atLeastOneOrnateFound = true
		            	markedItems[5][itemId] = true
						if (settings.showOrnateItemsInChat == true) then
							d(fco_loc["marked"] .. itemData.name .. fco_loc["ornate_item_found"])
		                else
		    				debugMessage("[ScanInventoryForOrnate]: " .. fco_loc["marked"] .. itemData.name .. fco_loc["ornate_item_found"], false)
						end
					end
				end
            end
		end
        preventerVars.gScanningInv = false

        --Was at least one ornate item found?
        if atLeastOneOrnateFound == true then
		    FilterBasics(true)
        end
	end
end

--Scan inventory for researchable items, marked by addon ResearchAssistant
local function scanInventoryForResearch()
    --If the addon ResearchAssistant is active
	if (otherAddons.researchAssistantActive == true and settings.autoMarkResearch == true and preventerVars.gScanningInv == false) then
    	preventerVars.gScanningInv = true

    	local atLeastOneResearchableFound = false
	    local bagId = BAG_BACKPACK
        local itemData
        local itemId
        --for each item in your player inventory
		--for slotIndex = 1, GetBagSize(bagId) do
		for slotIndex = 0, GetBagSize(bagId) - 1 do
			--Initialize variables again
			itemId = nil
			itemData = nil
          	itemId = SignItemId(GetItemInstanceId(bagId, slotIndex))
			if (markedItems[3][itemId] == false or markedItems[3][itemId] == nil) then
	        	itemData = PLAYER_INVENTORY.inventories[INVENTORY_BACKPACK].slots[slotIndex]
			    if (itemData ~= nil) then
	            	if (itemData.researchAssistant ~= nil and itemData.researchAssistant == 'researchable') then
                        atLeastOneResearchableFound = true
                    	markedItems[3][itemId] = true
	                    if (settings.showResearchItemsInChat == true) then
		                	d(fco_loc["marked"] .. itemData.name .. fco_loc["research_item_found"])
	                    else
		    				debugMessage("[ScanInventoryForResearch]: " .. fco_loc["marked"] .. itemData.name .. fco_loc["research_item_found"], false)
	                    end
	                end
		        end
            end
		end

        preventerVars.gScanningInv = false

        --Was at least one researchable item found?
        if atLeastOneResearchableFound == true then
	        FilterBasics(true)
        end

	end
end

--Scan the inventory for ornate and/or researchable items
local function scanInventory(p_bagId, p_slotIndex)
    local updateInv = false

	--Automatic marking of ornate and/or researchable items (researchAssistant addon needed!) is activated?
	if ( (settings.autoMarkOrnate == true and preventerVars.gScanningInv == false) or
         (otherAddons.researchAssistantActive == true and settings.autoMarkResearch == true and preventerVars.gScanningInv == false) ) then

		debugMessage("[ScanInventory] Start", false)

		-- Scan the whole inventory because no bagId and slotIndex are given
    	if (p_bagId == nil and p_slotIndex == nil) then

			--Start Inventory scan here
	    	preventerVars.gScanningInv = true

			--Update all items in inventory
			local bagId = BAG_BACKPACK
            local itemData
    		local itemId

		    --for each item in your player inventory
			--for slotIndex = 1, GetBagSize(bagId) do
			for slotIndex = 0, GetBagSize(bagId) - 1 do
				--Initialize variables again
				itemId = MyGetItemInstanceIdNoControl(bagId, slotIndex)

                if (itemId ~= nil) then

                	--Check for ornate items
					if (settings.autoMarkOrnate == true) then
						if (markedItems[5][itemId] == false or markedItems[5][itemId] == nil) then
							if (isItemOrnate(bagId, slotIndex)) then
								itemData = PLAYER_INVENTORY.inventories[INVENTORY_BACKPACK].slots[slotIndex]
								if (itemData ~= nil) then
									markedItems[5][itemId] = true
									if (settings.showOrnateItemsInChat == true) then
										d(fco_loc["marked"] .. itemData.name .. fco_loc["ornate_item_found"])
									else
										debugMessage("[ScanInventoryAll]: Ornate" .. fco_loc["marked"] .. itemData.name .. fco_loc["ornate_item_found"], false)
									end
									--Update the inventory
		                            updateInv = true
								end
							end
						end
			    	end

			        --Check researchable
		        	if (otherAddons.researchAssistantActive == true and settings.autoMarkResearch == true) then
						--itemId = MyGetItemInstanceIdNoControl(bagId, slotIndex)
						if (markedItems[3][itemId] == false or markedItems[3][itemId] == nil) then
			        		if (itemdata == nil) then
	                        	itemData = PLAYER_INVENTORY.inventories[INVENTORY_BACKPACK].slots[slotIndex]
	                        end
							if (itemData ~= nil) then
								if (itemData.researchAssistant ~= nil and itemData.researchAssistant == 'researchable') then
									markedItems[3][itemId] = true
									if (settings.showResearchItemsInChat == true) then
										d(fco_loc["marked"] .. itemData.name .. fco_loc["research_item_found"])
									else
										debugMessage("[ScanInventoryAll]: Research" .. fco_loc["marked"] .. itemData.name .. fco_loc["research_item_found"], false)
									end
									--Update the inventory
		                            updateInv = true
								end
							end
						end
			    	end

                end

        	end -- for slotindex = 1 ...

			--Inventory scan is finished here
			preventerVars.gScanningInv = false

        else --if (p_bagId == nil and p_slotIndex == nil) then

	       	-- Update only one item in inventory
			-- bagId AND slotIndex are given?
	        if (p_bagId ~= nil and p_slotIndex ~= nil) then
                local itemData
            	local itemId = MyGetItemInstanceIdNoControl(p_bagId, p_slotIndex)

                if (itemId ~= nil) then
					--Update ornate items
					if (settings.autoMarkOrnate == true) then
	                	if (markedItems[5][itemId] == false or markedItems[5][itemId] == nil) then
				        	if (p_bagId == BAG_BACKPACK) then
								--Check ornate
								if (isItemOrnate(p_bagId, p_slotIndex)) then
		                 		    itemData = PLAYER_INVENTORY.inventories[INVENTORY_BACKPACK].slots[p_slotIndex]
	                                if (itemData ~= nil) then
		                                markedItems[5][itemId] = true
										if (settings.showOrnateItemsInChat == true) then
											d(fco_loc["marked"] .. itemData.name .. fco_loc["ornate_item_found"])
										else
											debugMessage("[ScanInventory]: Ornate" .. fco_loc["marked"] .. itemData.name .. fco_loc["ornate_item_found"], false)
										end
										--Update the inventory
			                            updateInv = true
	                                end
								end
							end
	                    end
					end

					--Update researchable items
					if (otherAddons.researchAssistantActive == true and settings.autoMarkResearch == true) then
	                	if (markedItems[3][itemId] == false or markedItems[3][itemId] == nil) then
							--Check researchable
							if (p_bagId == BAG_BACKPACK) then
	                            if (itemData == nil) then
									itemData = PLAYER_INVENTORY.inventories[INVENTORY_BACKPACK].slots[p_slotIndex]
	                            end
	                            if (itemData ~= nil) then
		                        	if (itemData.researchAssistant ~= nil and itemData.researchAssistant == 'researchable') then
										markedItems[3][itemId] = true
										if (settings.showResearchItemsInChat == true) then
											d(fco_loc["marked"] .. itemData.name .. fco_loc["research_item_found"])
										else
											debugMessage("[ScanInventory]: Research" .. fco_loc["marked"] .. itemData.name .. fco_loc["research_item_found"], false)
										end
										--Update the inventory
			                            updateInv = true
									end
	                            end
							end
						end
			        end

                    --[[
			            --Debug messages
			            if (itemData ~= nil and itemData.researchAssistant ~= nil) then
				        	debugMessage("==============", false)
				            debugMessage("scanInventory: [" .. itemData.name .. "] BagId=" .. p_bagId .. ", SlotIndex: " .. p_slotIndex .. ", research: " .. itemData.researchAssistant, false)
				        else
				        	debugMessage("==============", false)
					   		debugMessage("scanInventory: BagId=" .. p_bagId .. ", SlotIndex: " .. p_slotIndex, false)
				        end
                    ]]

				end

			end --if (p_bagId ~= nil and p_slotIndex ~= nil) then

    	end --if (p_bagId == nil and p_slotIndex == nil) then

		--Update the inventories
        if updateInv == true then
        	FilterBasics(true)
        end

    end --if ( (settings.autoMarkOrnate == true and preventerVars.gScanningInv == false) or...

	debugMessage("[ScanInventory] End", false)

	--Inventory scan is latest finished here
	preventerVars.gScanningInv = false

end

--Enable the filters
local function EnableFilters(p_onoff)
	-- Enable the filters for the different panels (inventory, deconstruction, guild store, etc.)
    for filters = 1, locVars.gFCONumFilters, 1 do
		--function doFilter(onoff, p_button, filterId, beQuiet, doFilterBasicsPlayer, doUpdateButtonColors, onlyPlayerInvFilter, p_FilterPanelId)
    	doFilter(p_onoff, nil, filters, true, true, true, true)
    end
	--Reset the override flag for unregistering all filters if settings "Split filters" has changed
	locVars.gSplitFilterOverride = false
end

--Show a help inside the chat
local function help()
	d(fco_loc["chatcommands_info"])
	d("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
	d(fco_loc["chatcommands_help"])
	d(fco_loc["chatcommands_status"])
	d(fco_loc["chatcommands_split_filter"])
	if (settings.splitFilters == true) then
        d(fco_loc["chatcommands_filterpanels"])
        d(fco_loc["chatcommands_filterpanels2"])
        d(fco_loc["chatcommands_filtervalues"])
		d(fco_loc["chatcommands_filter1_new"])
		d(fco_loc["chatcommands_filter2_new"])
		d(fco_loc["chatcommands_filter3_new"])
		d(fco_loc["chatcommands_filter4_new"])
		d(fco_loc["chatcommands_filter_new"])
		d(fco_loc["chatcommands_filteron_new"])
		d(fco_loc["chatcommands_filteroff_new"])
		d(fco_loc["chatcommands_filtershow_new"])
    else
		d(fco_loc["chatcommands_filter1"])
		d(fco_loc["chatcommands_filter2"])
		d(fco_loc["chatcommands_filter3"])
		d(fco_loc["chatcommands_filter4"])
		d(fco_loc["chatcommands_filter"])
		d(fco_loc["chatcommands_filteron"])
		d(fco_loc["chatcommands_filteroff"])
		d(fco_loc["chatcommands_filtershow"])
    end
    d(fco_loc["chatcommands_debug"])
end

--Helper function for function filterStatus
local function filterStatusLoop(filterId, silent, givenArray, p_atLeastOneFilterActive)
    local returnArray = {}
    local atLeastOneFilterActive = p_atLeastOneFilterActive

	--Is an array already prepared?
	if #givenArray == 0 then
		--Create 2-dimensional arrays for the filters
	    for help_inv = 1, locVars.gFCONumFilterInventoryTypes, 1 do
		    returnArray[help_inv] = {false, false, false, false}
	    end
    else
	    --Array was already given, so use it
		returnArray = givenArray
    end

   	-- Check only one filter
	if (settings.splitFilters == true) then
		for j=1, locVars.gFCONumFilterInventoryTypes, 1 do
			if(getSettingsIsFilterOn(filterId, j)) then
				returnArray[j][filterId] = true
                atLeastOneFilterActive = true
				if     (j == INV_BAGS) then
					returnArray[j][filterId] = settings.allowInventoryFilter
					if (not silent) then
						d(fco_loc["filter_inventory"] .. fco_loc["chatcommands_status_filter" .. tostring(filterId)])
					end
				elseif     (j == INV_STORE) then
					returnArray[j][filterId] = settings.allowVendorFilter
					if (not silent) then
					d(fco_loc["filter_store"] .. fco_loc["chatcommands_status_filter" .. tostring(filterId)])
					end
				elseif (j == INV_GUILDBANK) then
					returnArray[j][filterId] = settings.allowGuildBankFilter
					if (not silent) then
						d(fco_loc["filter_guildbank"] .. fco_loc["chatcommands_status_filter" .. tostring(filterId)])
					end
				elseif (j == INV_GUILDSTORE) then
					returnArray[j][filterId] = settings.allowTradinghouseFilter
					if (not silent) then
						d(fco_loc["filter_guildstore"] .. fco_loc["chatcommands_status_filter" .. tostring(filterId)])
					end
				elseif (j == INV_BANK) then
					returnArray[j][filterId] = settings.allowBankFilter
					if (not silent) then
						d(fco_loc["filter_bank"] .. fco_loc["chatcommands_status_filter" .. tostring(filterId)])
					end
				elseif (j == INV_DECONSTRUCTION) then
					returnArray[j][filterId] = settings.allowDeconstructionFilter
					if (not silent) then
						d(fco_loc["filter_deconstruction"] .. fco_loc["chatcommands_status_filter" .. tostring(filterId)])
					end
				elseif (j == INV_IMPROVEMENT) then
					returnArray[j][filterId] = settings.allowImprovementFilter
					if (not silent) then
						d(fco_loc["filter_improvement"] .. fco_loc["chatcommands_status_filter" .. tostring(filterId)])
					end
				elseif (j == INV_MAIL) then
					returnArray[j][filterId] = settings.allowMailFilter
					if (not silent) then
						d(fco_loc["filter_mail"] .. fco_loc["chatcommands_status_filter" .. tostring(filterId)])
					end
				elseif (j == INV_TRADE) then
					returnArray[j][filterId] = settings.allowTradeFilter
					if (not silent) then
						d(fco_loc["filter_trade"] .. fco_loc["chatcommands_status_filter" .. tostring(filterId)])
					end
				elseif (j == INV_ENCHANTING_EXTRACTION) then
					returnArray[j][filterId] = settings.allowEnchantingFilter
					if (not silent) then
						d(fco_loc["filter_enchantingstation_extraction"] .. fco_loc["chatcommands_status_filter" .. tostring(filterId)])
					end
				elseif (j==INV_ENCHANTING_CREATION) then
					returnArray[j][filterId] = settings.allowEnchantingFilter
					if (not silent) then
						d(fco_loc["filter_enchantingstation_creation"] .. fco_loc["chatcommands_status_filter" .. tostring(filterId)])
					end
				end
			end -- if(getSettingsIsFilterOn(filterId, j)) then
		end -- for j=1, ...

	else --if (settings.splitFilters == true)

		if(getSettingsIsFilterOn(filterId)) then
               returnArray[1][filterId] = true
               atLeastOneFilterActive = true
			if (not silent) then
				d(fco_loc["chatcommands_status_filter" .. tostring(filterId)])
			end
		end

    end --if (settings.splitFilters == true)


    return returnArray, atLeastOneFilterActive
end

--Check a filter status
local function filterStatus(filterId, silent, doReturnFilterStatus)
	--Prepare the return array
    local retArray = {}
    local atLeastOneFilterActive = false

	if (filterId ~= -1) then
		retArray, atLeastOneFilterActive = filterStatusLoop(filterId, silent)
    else
    	-- Check all filters
		for i=1, locVars.gFCONumFilters, 1 do
        	retArray, atLeastOneFilterActive = filterStatusLoop(i, silent, retArray, atLeastOneFilterActive)
		end
    end

	--Was at least one active filter found and chat output is enabled?
    if atLeastOneFilterActive == false and silent == false then
    	d(fco_loc["chatcommands_status_nofilters"])
    end

	--Return the array with filter states?
    if (doReturnFilterStatus == true) then
	    return retArray
    end
end

--Show a status inside the chat
local function status()
	d(fco_loc["chatcommands_status_info"])

    --Check all filters, not silently (chat output: enabled)
    filterStatus(-1, false, false)

	if (settings.debug == true) then
		d(fco_loc["chatcommands_debug_on"])
	else
		d(fco_loc["chatcommands_debug_off"])
	end
end

--Check the commands ppl type to the chat
local function command_handler(args)
    --Parse the arguments string
	local options = {}
    local searchResult = { string.match(args, "^(%S*)%s*(.-)$") }
    for i,v in pairs(searchResult) do
        if (v ~= nil and v ~= "") then
            options[i] = string.lower(v)
        end
    end

	local opt3 = false
	if (options[3] ~= nil and (options[3] == false or options[3] == true or options[3] == "show")) then
		opt3 = true
		if 	   (options[3]=="show" or options[3]=="montre" or options[3]=="zeigen") then
			options[3] = -99
		elseif (options[3]==true or options[3]=="true" or options[3]=="vrai" or options[3]=="an") then
			options[3] = 1
		elseif (options[3]==false or options[3]=="false" or options[3]=="faux" or options[3]=="aus") then
			options[3] = 2
		end
	end

	if(options[1] == "help" or options[1] == "hilfe" or options[1] == "list") then
       	help()
	elseif(#options == 0 or options[1] == "status" or options[1] == "") then
       	status()
    elseif(options[1] == "filter1" or options[1] == "filtre1") then
        if (opt3) then
	    	doFilter(tonumber(options[3]), nil, 1, false, false, true, false, tonumber(options[2]))
        else
	        if (options[2] ~= nil) then
		    	doFilter(-1, nil, 1, false, false, true, false, tonumber(options[2]))
	        else
		    	doFilter(-1, nil, 1, false, false, true, false)
	        end
        end
	elseif(options[1] == "filter2" or options[1] == "filtre2") then
        if (opt3) then
	    	doFilter(tonumber(options[3]), nil, 2, false, false, true, false, tonumber(options[2]))
        else
	        if (options[2] ~= nil) then
		        doFilter(-1, nil, 2, false, false, true, false, tonumber(options[2]))
	        else
		        doFilter(-1, nil, 2, false, false, true, false)
	        end
        end
	elseif(options[1] == "filter3" or options[1] == "filtre3") then
        if (opt3) then
	    	doFilter(tonumber(options[3]), nil, 3, false, false, true, false, tonumber(options[2]))
        else
	        if (options[2] ~= nil) then
		    	doFilter(-1, nil, 3, false, false, true, false, tonumber(options[2]))
	        else
		    	doFilter(-1, nil, 3, false, false, true, false)
	        end
		end
	elseif(options[1] == "filter4" or options[1] == "filtre4") then
        if (opt3) then
	    	doFilter(tonumber(options[3]), nil, 4, false, false, true, false, tonumber(options[2]))
        else
	        if (options[2] ~= nil) then
		    	doFilter(-1, nil, 4, false, false, true, false, tonumber(options[2]))
	        else
		    	doFilter(-1, nil, 4, false, false, true, false)
	        end
		end
	elseif(options[1] == "filtern" or options[1] == "filter" or options[1] == "filtres") then
        if (opt3) then
	    	for i=1, locVars.gFCONumFilters, 1 do
		    	doFilter(tonumber(options[3]), nil, i, false, false, true, false, tonumber(options[2]))
            end
        else
	    	for i=1, locVars.gFCONumFilters, 1 do
		        if (options[2] ~= nil) then
			    	doFilter(-1, nil, i, false, false, true, false, tonumber(options[2]))
		        else
			    	doFilter(-1, nil, i, false, false, true, false)
		        end
	        end
		end
	elseif(options[1] == "allean" or options[1] == "allon" or options[1] == "touson") then
		if (settings.splitFilters == true) then
	    	for i=1, locVars.gFCONumFilters, 1 do
				for j=1, locVars.gFCONumFilterInventoryTypes, 1 do
					if (options[2] ~= nil) then
						doFilter(1, nil, i, false, false, true, false, j, tonumber(options[2]))
					else
						doFilter(1, nil, i, false, false, true, false, j)
					end
                end
	        end
        else
	    	for i=1, locVars.gFCONumFilters, 1 do
		    	doFilter(1, nil, i, false, false, true, false)
	        end
        end
	elseif(options[1] == "alleaus" or options[1] == "alloff" or options[1] == "tousoff") then
		if (settings.splitFilters == true) then
	    	for i=1, locVars.gFCONumFilters, 1 do
				for j=1, locVars.gFCONumFilterInventoryTypes, 1 do
					if (options[2] ~= nil) then
						doFilter(2, nil, i, false, false, true, false, j, tonumber(options[2]))
					else
						doFilter(2, nil, i, false, false, true, false, j)
					end
                end
	        end
        else
	    	for i=1, locVars.gFCONumFilters, 1 do
		    	doFilter(2, nil, i, false, false, true, false)
	        end
        end
	elseif(options[1] == "allezeigen" or options[1] == "allshow" or options[1] == "tousmontre") then
		if (settings.splitFilters == true) then
	    	for i=1, locVars.gFCONumFilters, 1 do
				for j=1, locVars.gFCONumFilterInventoryTypes, 1 do
					if (options[2] ~= nil) then
						doFilter(-99, nil, i, false, false, true, false, j, tonumber(options[2]))
					else
						doFilter(-99, nil, i, false, false, true, false, j)
					end
                end
	        end
        else
	    	for i=1, locVars.gFCONumFilters, 1 do
		    	doFilter(-99, nil, i, false, false, true, false)
	        end
        end
    elseif(options[1] == "split") then
    	settings.splitFilters = not settings.splitFilters
		--Set the global flag to override the function's unregisterFilters() unregister method
		--because it will check for settings.SplitFilters and is called from function dofilter() at
		--function EnableFilters(-100) below
		locVars.gSplitFilterOverride = true
		EnableFilters(-100)
    elseif(options[1] == "debug") then
    	settings.debug = not settings.debug
        if (settings.debug == true) then
        	d(fco_loc["chatcommands_debug_on"])
        else
        	settings.deepDebug = false
        	d(fco_loc["chatcommands_debug_off"])
        end
    elseif(options[1] == "deepdebug") then
        settings.deepDebug = not settings.deepDebug
        if (settings.deepDebug == true) then
			settings.debug = true
        	d(fco_loc["chatcommands_deepdebug_on"])
        else
			settings.debug = false
        	d(fco_loc["chatcommands_deepdebug_off"])
        end
    end
end

--Add the filter buttons to the inventory/bank/mail/trade/guild bank/guild store/crafting stations panels
local function AddOrChangeFilterButton(parentWindow, buttonId, pWidth, pHeight, pLeft, pTop, hide, p_FilterPanelId)
	--Get the current filter panel Id
    p_FilterPanelId = p_FilterPanelId or locVars.gFilterWhere

	local buttonName = parentWindow:GetName() .. "_FilterButton" .. tostring(buttonId)
    local tooltipText

    local checkIfButtonexists = WINDOW_MANAGER:GetControlByName(buttonName, "")
	--ADD new button (should only be executed after addon has loaded or a zone change)
    if (checkIfButtonexists == nil) then

	    --Check if the buttons should be not added, because the relating filter setting (e.g. settings.allowInventoryFilter) is disabled
	    if hide ~= true then

	    	debugMessage("AddOrChangeFilterButton: ADD ButtonName=" .. buttonName .. ", ButtonId=" .. buttonId .. ", PanelId=" .. locVars.gFilterWhere .. ", Width/Height: " .. pWidth .. "/" .. pHeight .. ", Left/Top: " .. pLeft .. "/" .. pTop .. ", Hide: " .. tostring(hide), true)
			local button = WINDOW_MANAGER:CreateControl(parentWindow:GetName() .. "_FilterButton" .. tostring(buttonId), parentWindow, CT_BUTTON)
		    button:SetDimensions(pWidth, pHeight)
	       --Move the button more to the right, if standard left values are used and GridView Addon is activated
	       if (otherAddons.inventoryGridViewActive == true
	            	and ( settings.filterButtonLeft[1] == locVars.gFilterButtonLeft[1] and settings.filterButtonLeft[2] == locVars.gFilterButtonLeft[2]
	                  and settings.filterButtonLeft[3] == locVars.gFilterButtonLeft[3] and settings.filterButtonLeft[4] == locVars.gFilterButtonLeft[4]) ) then
	        	pLeft = pLeft + locVars.gGridViewOffsetX
		    end
            --Place the buttons at the bottom of the inventory.
            --Special treatment for improvement panel here, because the "Booster container" is located at the bottom and the buttons
            --will be shown above him, not below (as he gives the BOTTOM anchor) of the inventory
            if (p_FilterPanelId ~= INV_IMPROVEMENT) then
	            button:SetAnchor(TOP, parentWindow, BOTTOM, pLeft, pTop)
            else
               	pTop = pTop + 7
            	button:SetAnchor(TOP, IMPROVEMENT_BOOSTER_CONTAINER, BOTTOM, pLeft, pTop)
			end
		    button:SetFont("ZoFontGameSmall")

			--Set a tooltip?
            if settings.showFilterButtonTooltip == true then
	            tooltipText = outputFilterState(false, settings.splitFilters, p_FilterPanelId, buttonId, settingsFilterStateToText[tostring(getSettingsIsFilterOn(buttonId, p_FilterPanelId))])
				if tooltipText ~= nil then
					button:SetHandler("OnMouseEnter", function(self)
			            if settings.showFilterButtonTooltip == true then
							ZO_Tooltips_ShowTextTooltip(button, BOTTOM, tooltipText)
                        else
							ZO_Tooltips_HideTextTooltip()
                        end
					end)
					button:SetHandler("OnMouseExit", function(self)
						ZO_Tooltips_HideTextTooltip()
					end)
		        end
			else
				ZO_Tooltips_HideTextTooltip()
            end

			--Overwrite the callback function of the button so it is channging the correct panel filter
	        --as mail, trade, bank, guild bank, and others all use the ZO_PlayerInventory buttons to filter!
		    button:SetHandler("OnClicked", function(butn)
	            debugMessage("=========>", true)
	        	debugMessage("[FilterButtonOnClick_ADD]: ButtonName: " .. butn:GetName() .. ", ButtonId: " .. buttonId .. ", PanelId: " .. locVars.gFilterWhere, true)
				--Hide the tooltip
				ZO_Tooltips_HideTextTooltip()
	            doFilter(-1, butn, buttonId, false, false, true, false, p_FilterPanelId)
	            if settings.showFilterButtonTooltip == true then
		            tooltipText = outputFilterState(false, settings.splitFilters, p_FilterPanelId, buttonId, settingsFilterStateToText[tostring(getSettingsIsFilterOn(buttonId, p_FilterPanelId))])
					ZO_Tooltips_ShowTextTooltip(button, BOTTOM, tooltipText)
                end
	        end)

	        --Create the texture for the button to hold the image
		    local texture = WINDOW_MANAGER:CreateControl(buttonName .. "Texture", button, CT_TEXTURE)
		    texture:SetAnchorFill()

			-- Fix to show icon 5 for filter 4
	        local offset = 0
	        if (buttonId == 4) then
	        	offset = 1
	        end
	        --Workaround to show at least a default texure, if none is found
	  		if (MARKER_TEXTURES[settings.icon[buttonId + offset].texture] ~= nil) then
		        --Set the texture now
		        texture:SetTexture(MARKER_TEXTURES[settings.icon[buttonId + offset].texture])
			else
		        --Set fallback texture now
		        texture:SetTexture(MARKER_TEXTURES[buttonId])
				--Set the fallback texture to the settings menu
	            settings.icon[buttonId + offset].texture = buttonId
	        end

			--Show the button and make it react on mouse input
        	button:SetHidden(false)
	    	button:SetMouseEnabled(true)

	        return button

        else -- if hide ~= true then
	    	debugMessage("AddOrChangeFilterButton: Filter setting is disabled! No button was added.  ButtonId=" .. buttonId .. ", PanelId=" .. locVars.gFilterWhere .. ", Hide: " .. tostring(hide), true)
        end -- if hide ~= true then

	--CHANGE an existing button (should be executed as another panel is active and someone clicks a filter button/uses the chat commands to filter)
    else
        -- Button exists already
    	debugMessage("AddOrChangeFilterButton: CHANGE ButtonName: " .. buttonName .. ", ButtonId=" .. buttonId .. ", PanelId=" .. locVars.gFilterWhere .. ", Width/Height: " .. pWidth .. "/" .. pHeight .. ", Left/Top: " .. pLeft .. "/" .. pTop .. ", Hide: " .. tostring(hide), true)

	    --Hide or show the button?
	    if (hide == true) then
        		debugMessage("AddOrChangeFilterButton: CHANGE Button will be hidden! ButtonName: " .. buttonName .. ", ButtonId=" .. buttonId .. ", PanelId=" .. locVars.gFilterWhere, true)
	        	checkIfButtonexists:SetHidden(true)
		    	checkIfButtonexists:SetMouseEnabled(false)
		else
				--Update left, top, width, height, and onClick handler
				if (otherAddons.inventoryGridViewActive == true
					and ( settings.filterButtonLeft[1] == locVars.gFilterButtonLeft[1] and settings.filterButtonLeft[2] == locVars.gFilterButtonLeft[2]
					and settings.filterButtonLeft[3] == locVars.gFilterButtonLeft[3] and settings.filterButtonLeft[4] == locVars.gFilterButtonLeft[4]) ) then
						pLeft = pLeft + locVars.gGridViewOffsetX
				end
	            --Place the buttons at the bottom of the inventory.
	            --Special treatment for improvement panel here, because the "Booster container" is located at the bottom and the buttons
	            --will be shown above him, not below (as he gives the BOTTOM anchor) of the inventory
	            if (p_FilterPanelId ~= INV_IMPROVEMENT) then
				    checkIfButtonexists:SetAnchor(TOP, parentWindow, BOTTOM, pLeft, pTop)
	            else
                	pTop = pTop + 7
				    checkIfButtonexists:SetAnchor(TOP, IMPROVEMENT_BOOSTER_CONTAINER, BOTTOM, pLeft, pTop)
				end
				checkIfButtonexists:SetFont("ZoFontGameSmall")

				--Set a tooltip?
            	if settings.showFilterButtonTooltip == true then
	                tooltipText = outputFilterState(false, settings.splitFilters, p_FilterPanelId, buttonId, settingsFilterStateToText[tostring(getSettingsIsFilterOn(buttonId, p_FilterPanelId))])
					if tooltipText ~= nil then
						checkIfButtonexists:SetHandler("OnMouseEnter", function(self)
			            	if settings.showFilterButtonTooltip == true then
								ZO_Tooltips_ShowTextTooltip(checkIfButtonexists, BOTTOM, tooltipText)
                            else
								ZO_Tooltips_HideTextTooltip()
                            end
						end)
						checkIfButtonexists:SetHandler("OnMouseExit", function(self)
							ZO_Tooltips_HideTextTooltip()
						end)
			        end
                else
					ZO_Tooltips_HideTextTooltip()
                end

				--Overwrite the callback function of the button so it is channging the correct panel filter
		        --as mail, trade, bank, guild bank, and others all use the ZO_PlayerInventory buttons to filter!
				checkIfButtonexists:SetHandler("OnClicked", function(btn)
		            debugMessage("=========>", true)
		        	debugMessage("[FilterButtonOnClick_CHANGE]: ButtonName: " .. btn:GetName() .. ", ButtonId:" .. buttonId .. ", PanelId: " .. locVars.gFilterWhere, true)
                    --Hide the tooltip
                    ZO_Tooltips_HideTextTooltip()
                    doFilter(-1, btn, buttonId, false, false, true, false, p_FilterPanelId)
		            if settings.showFilterButtonTooltip == true then
	                    tooltipText = outputFilterState(false, settings.splitFilters, p_FilterPanelId, buttonId, settingsFilterStateToText[tostring(getSettingsIsFilterOn(buttonId, p_FilterPanelId))])
	                    ZO_Tooltips_ShowTextTooltip(checkIfButtonexists, BOTTOM, tooltipText)
                    end
				end)

        		debugMessage("AddOrChangeFilterButton: CHANGE Button will be shown! ButtonName: " .. buttonName .. ", ButtonId=" .. buttonId .. ", PanelId=" .. locVars.gFilterWhere, true)

	        	checkIfButtonexists:SetHidden(false)
		    	checkIfButtonexists:SetMouseEnabled(true)
	    end

        return checkIfButtonexists
    end

	-- If we reach here hide is enabled and no button was added. Return nil then
    return nil
end

-- Puts additional points of data into control.dataEntry.data called "FCOItemSaverMarked1" to "FCOItemSaverMarked8"
-- The table is using the folowing indices: 1, 2, 3, 4, 5, 6, 7, 8 (each for one of the filter icons)
-- If e.g. "FCOItemSaverMarked[1] = true": The current item is marked with the vendor filter icon #1
-- If e.g. "FCOItemSaverMarked[8] = true": The current item is marked for the gear 5 filter icon #8
local function FCOItemSaver_AddInfoToData(control)
	if (control ~= nil) then
		if (control.dataEntry.data ~= nil) then
			local myItemInstanceId = MyGetItemInstanceId(control)
			if (myItemInstanceId ~= nil) then
				debugMessage("[FCOItemSaver_AddInfoToData] ItemIstanceId:" .. tostring(myItemInstanceId), true)
				--Check if the current item is marked with any of the marker icons
				--for markedIndex=1, locVars.gFCONumFilterIcons, 1 do
					control.dataEntry.data.FCOItemSaverMarked1 = markedItems[1][myItemInstanceId]
					control.dataEntry.data.FCOItemSaverMarked2 = markedItems[2][myItemInstanceId]
					control.dataEntry.data.FCOItemSaverMarked3 = markedItems[3][myItemInstanceId]
					control.dataEntry.data.FCOItemSaverMarked4 = markedItems[4][myItemInstanceId]
					control.dataEntry.data.FCOItemSaverMarked5 = markedItems[5][myItemInstanceId]
					control.dataEntry.data.FCOItemSaverMarked6 = markedItems[6][myItemInstanceId]
					control.dataEntry.data.FCOItemSaverMarked7 = markedItems[7][myItemInstanceId]
					control.dataEntry.data.FCOItemSaverMarked8 = markedItems[8][myItemInstanceId]
				--end
			end
        end
	end
end

--Create the textures inside inventories etc.
local function CreateTextures(whichTextures)

	if (whichTextures == 1 or whichTextures == 4 or whichTextures == 6 or whichTextures == -1) then
	    --Create textures in inventories
	    for _,v in pairs(PLAYER_INVENTORY.inventories) do
			local listView = v.listView
			if listView and listView.dataTypes and listView.dataTypes[1] then
				local hookedFunctions = listView.dataTypes[1].setupCallback

				listView.dataTypes[1].setupCallback =
					function(rowControl, slot)
						hookedFunctions(rowControl, slot)

	                    -- for all filters: Create/Update the icons
						for i=1, locVars.gFCONumFilterIcons, 1 do
							CreateMarkerControl(rowControl, i, settings.icon[i].size, settings.icon[i].size, MARKER_TEXTURES[settings.icon[i].texture])
	                    end
                        --Add additional FCO point to the dataEntry.data slot
                        --FCOItemSaver_AddInfoToData(rowControl)
					end
			end
		end
	end

	if (whichTextures == 3 or whichTextures == 4 or whichTextures == 8 or whichTextures == -1) then
	    --Create textures in repair window
		local listView = REPAIR_WINDOW.list
		if listView and listView.dataTypes and listView.dataTypes[1] then
			local hookedFunctions = listView.dataTypes[1].setupCallback

			listView.dataTypes[1].setupCallback =
				function(rowControl, slot)
					hookedFunctions(rowControl, slot)

					--Only create/update markers if the repair window player inventory is shown
					if (not REPAIR_WINDOW.list:IsHidden()) then
	                    -- for all filters: Create/Update the icons
						for i=1, locVars.gFCONumFilterIcons, 1 do
							CreateMarkerControl(rowControl, i, settings.icon[i].size, settings.icon[i].size, MARKER_TEXTURES[settings.icon[i].texture])
						end
	                    --Add additional FCO point to the dataEntry.data slot
	                    --FCOItemSaver_AddInfoToData(rowControl)
                    end
				end
		end
	end

	if (whichTextures == 5 or whichTextures == 6 or whichTextures == 8 or whichTextures == -1) then
	    -- Marker function for character equipment if character window is shown
        if (not CHARACTER:IsHidden() or addonVars.gAddonLoaded == false) then
	        RefreshEquipmentControl()
        end
	end
end

--Check if markers should be updated
local function checkMarker(markerId)
	markerId = markerId or -1
   	debugMessage("[CheckMarker] MarkerId: " .. tostring(markerId) .. ", CheckNow: " .. tostring(preventerVars.gUpdateMarkersNow) .. ", Gears changed: " .. tostring(preventerVars.gChangedGears), true)

	--Should we update the marker textures, size and color?
	if preventerVars.gUpdateMarkersNow == true or preventerVars.gChangedGears == true then
		--Update the textures now
		RefreshBackpack()
		RefreshBank()
		RefreshGuildBank()

		if not preventerVars.gChangedGears then
			zo_callLater(function()
				--d("character hidden: " .. tostring(CHARACTER:IsHidden()))
				if not CHARACTER:IsHidden() then
			        RefreshEquipmentControl()
	            end
	        end, 100)
		end

		--Set the global preventer variable back to false
		preventerVars.gUpdateMarkersNow = false
        preventerVars.gChangedGears		= false
	end
end

-- Change the gear right click menu texts to the settings value
local function changeEquipmentGearMenuTexts(gearId)
	if (gearId == -1) then
		--Overwrite the localized texts for all equipment gears, if changed in the settings
		if(settings.icon[2].name ~= nil and settings.icon[2].name ~= '' and settings.icon[2].name ~= fco_loc["rightclick_menu_add_gear1"]) then
			contextEntries.menu_add_gear1_text 		= fco_loc["rightclick_menu_add_start_gear"] .. settings.icon[2].name
            contextEntries.menu_remove_gear1_text 		= fco_loc["rightclick_menu_remove_start_gear"] .. settings.icon[2].name
			contextEntries.menu_add_all_gear1_text    	= fco_loc["rightclick_menu_add_all_start_gear"] .. settings.icon[2].name
            contextEntries.menu_remove_all_gear1_text 	= fco_loc["rightclick_menu_remove_all_start_gear"] .. settings.icon[2].name
        else
        	contextEntries.menu_add_gear1_text 		= fco_loc["rightclick_menu_add_gear1"]
            contextEntries.menu_remove_gear1_text 		= fco_loc["rightclick_menu_remove_gear1"]
			contextEntries.menu_add_all_gear1_text 	= fco_loc["rightclick_menu_add_all_gear1"]
            contextEntries.menu_remove_all_gear1_text  = fco_loc["rightclick_menu_remove_all_gear1"]
        end
		if(settings.icon[4].name ~= nil and settings.icon[4].name ~= '' and settings.icon[4].name ~= fco_loc["rightclick_menu_add_gear2"]) then
			contextEntries.menu_add_gear2_text 		= fco_loc["rightclick_menu_add_start_gear"] .. settings.icon[4].name
            contextEntries.menu_remove_gear2_text 		= fco_loc["rightclick_menu_remove_start_gear"] .. settings.icon[4].name
			contextEntries.menu_add_all_gear2_text    	= fco_loc["rightclick_menu_add_all_start_gear"] .. settings.icon[4].name
            contextEntries.menu_remove_all_gear2_text 	= fco_loc["rightclick_menu_remove_all_start_gear"] .. settings.icon[4].name
        else
        	contextEntries.menu_add_gear2_text 		= fco_loc["rightclick_menu_add_gear2"]
            contextEntries.menu_remove_gear2_text 		= fco_loc["rightclick_menu_remove_gear2"]
			contextEntries.menu_add_all_gear2_text 	= fco_loc["rightclick_menu_add_all_gear2"]
            contextEntries.menu_remove_all_gear2_text  = fco_loc["rightclick_menu_remove_all_gear2"]
	    end
		if(settings.icon[6].name ~= nil and settings.icon[6].name ~= '' and settings.icon[6].name ~= fco_loc["rightclick_menu_add_gear3"]) then
			contextEntries.menu_add_gear3_text 		= fco_loc["rightclick_menu_add_start_gear"] .. settings.icon[6].name
            contextEntries.menu_remove_gear3_text 		= fco_loc["rightclick_menu_remove_start_gear"] .. settings.icon[6].name
			contextEntries.menu_add_all_gear3_text    	= fco_loc["rightclick_menu_add_all_start_gear"] .. settings.icon[6].name
            contextEntries.menu_remove_all_gear3_text 	= fco_loc["rightclick_menu_remove_all_start_gear"] .. settings.icon[6].name
        else
        	contextEntries.menu_add_gear3_text 		= fco_loc["rightclick_menu_add_gear3"]
            contextEntries.menu_remove_gear3_text 		= fco_loc["rightclick_menu_remove_gear3"]
			contextEntries.menu_add_all_gear3_text 	= fco_loc["rightclick_menu_add_all_gear3"]
            contextEntries.menu_remove_all_gear3_text  = fco_loc["rightclick_menu_remove_all_gear3"]
	    end
		if(settings.icon[7].name ~= nil and settings.icon[7].name ~= '' and settings.icon[7].name ~= fco_loc["rightclick_menu_add_gear4"]) then
			contextEntries.menu_add_gear4_text 		= fco_loc["rightclick_menu_add_start_gear"] .. settings.icon[7].name
            contextEntries.menu_remove_gear4_text 		= fco_loc["rightclick_menu_remove_start_gear"] .. settings.icon[7].name
			contextEntries.menu_add_all_gear4_text    	= fco_loc["rightclick_menu_add_all_start_gear"] .. settings.icon[7].name
            contextEntries.menu_remove_all_gear4_text 	= fco_loc["rightclick_menu_remove_all_start_gear"] .. settings.icon[7].name
        else
        	contextEntries.menu_add_gear4_text 		= fco_loc["rightclick_menu_add_gear4"]
            contextEntries.menu_remove_gear4_text 		= fco_loc["rightclick_menu_remove_gear4"]
			contextEntries.menu_add_all_gear4_text 	= fco_loc["rightclick_menu_add_all_gear4"]
            contextEntries.menu_remove_all_gear4_text  = fco_loc["rightclick_menu_remove_all_gear4"]
	    end
		if(settings.icon[8].name ~= nil and settings.icon[8].name ~= '' and settings.icon[8].name ~= fco_loc["rightclick_menu_add_gear5"]) then
			contextEntries.menu_add_gear5_text 		= fco_loc["rightclick_menu_add_start_gear"] .. settings.icon[8].name
            contextEntries.menu_remove_gear5_text 		= fco_loc["rightclick_menu_remove_start_gear"] .. settings.icon[8].name
			contextEntries.menu_add_all_gear5_text    	= fco_loc["rightclick_menu_add_all_start_gear"] .. settings.icon[8].name
            contextEntries.menu_remove_all_gear5_text 	= fco_loc["rightclick_menu_remove_all_start_gear"] .. settings.icon[8].name
        else
        	contextEntries.menu_add_gear5_text 		= fco_loc["rightclick_menu_add_gear5"]
            contextEntries.menu_remove_gear5_text 		= fco_loc["rightclick_menu_remove_gear5"]
			contextEntries.menu_add_all_gear5_text 	= fco_loc["rightclick_menu_add_all_gear5"]
            contextEntries.menu_remove_all_gear5_text  = fco_loc["rightclick_menu_remove_all_gear5"]
	    end
	else
		--Overwrite the localized texts for the equipment gear geadId, if changed in the settings
		local iconId = 0
		if (gearId == 1) then
       		iconId = 2
			if(settings.icon[iconId].name ~= nil and settings.icon[iconId].name ~= '' and settings.icon[iconId].name ~= fco_loc["rightclick_menu_add_gear" .. tostring(gearId)]) then
			   contextEntries.menu_add_gear1_text	   	  = fco_loc["rightclick_menu_add_start_gear"] .. " " .. settings.icon[iconId].name
	           contextEntries.menu_remove_gear1_text     = fco_loc["rightclick_menu_remove_start_gear"] .. " " .. settings.icon[iconId].name
			   contextEntries.menu_add_all_gear1_text	  = fco_loc["rightclick_menu_add_all_start_gear"] .. " " .. settings.icon[iconId].name
	           contextEntries.menu_remove_all_gear1_text = fco_loc["rightclick_menu_remove_all_start_gear"] .. " " .. settings.icon[iconId].name
	        else
	           contextEntries.menu_add_gear1_text 		= fco_loc["rightclick_menu_add_gear" .. tostring(gearId)]
	           contextEntries.menu_remove_gear1_text 		= fco_loc["rightclick_menu_remove_gear" .. tostring(gearId)]
			   contextEntries.menu_add_all_gear1_text 	= fco_loc["rightclick_menu_add_all_gear" .. tostring(gearId)]
	           contextEntries.menu_remove_all_gear1_text  = fco_loc["rightclick_menu_remove_all_gear" .. tostring(gearId)]
	        end
        elseif (gearId == 2) then
        	iconId = 4
			if(settings.icon[iconId].name ~= nil and settings.icon[iconId].name ~= '' and settings.icon[iconId].name ~= fco_loc["rightclick_menu_add_gear" .. tostring(gearId)]) then
			   contextEntries.menu_add_gear2_text	   	  = fco_loc["rightclick_menu_add_start_gear"] .. " " .. settings.icon[iconId].name
	           contextEntries.menu_remove_gear2_text     = fco_loc["rightclick_menu_remove_start_gear"] .. " " .. settings.icon[iconId].name
			   contextEntries.menu_add_all_gear2_text	  = fco_loc["rightclick_menu_add_all_start_gear"] .. " " .. settings.icon[iconId].name
	           contextEntries.menu_remove_all_gear2_text = fco_loc["rightclick_menu_remove_all_start_gear"] .. " " .. settings.icon[iconId].name
	        else
	           contextEntries.menu_add_gear2_text 		= fco_loc["rightclick_menu_add_gear" .. tostring(gearId)]
	           contextEntries.menu_remove_gear2_text 		= fco_loc["rightclick_menu_remove_gear" .. tostring(gearId)]
			   contextEntries.menu_add_all_gear2_text 	= fco_loc["rightclick_menu_add_all_gear" .. tostring(gearId)]
	           contextEntries.menu_remove_all_gear2_text  = fco_loc["rightclick_menu_remove_all_gear" .. tostring(gearId)]
	        end
        elseif (gearId == 3) then
        	iconId = 6
			if(settings.icon[iconId].name ~= nil and settings.icon[iconId].name ~= '' and settings.icon[iconId].name ~= fco_loc["rightclick_menu_add_gear" .. tostring(gearId)]) then
			   contextEntries.menu_add_gear3_text	   	  = fco_loc["rightclick_menu_add_start_gear"] .. " " .. settings.icon[iconId].name
	           contextEntries.menu_remove_gear3_text     = fco_loc["rightclick_menu_remove_start_gear"] .. " " .. settings.icon[iconId].name
			   contextEntries.menu_add_all_gear3_text	  = fco_loc["rightclick_menu_add_all_start_gear"] .. " " .. settings.icon[iconId].name
	           contextEntries.menu_remove_all_gear3_text = fco_loc["rightclick_menu_remove_all_start_gear"] .. " " .. settings.icon[iconId].name
	        else
	           contextEntries.menu_add_gear3_text 		= fco_loc["rightclick_menu_add_gear" .. tostring(gearId)]
	           contextEntries.menu_remove_gear3_text 		= fco_loc["rightclick_menu_remove_gear" .. tostring(gearId)]
			   contextEntries.menu_add_all_gear3_text 	= fco_loc["rightclick_menu_add_all_gear" .. tostring(gearId)]
	           contextEntries.menu_remove_all_gear3_text  = fco_loc["rightclick_menu_remove_all_gear" .. tostring(gearId)]
            end
        elseif (gearId == 4) then
        	iconId = 7
			if(settings.icon[iconId].name ~= nil and settings.icon[iconId].name ~= '' and settings.icon[iconId].name ~= fco_loc["rightclick_menu_add_gear" .. tostring(gearId)]) then
			   contextEntries.menu_add_gear4_text	   	  = fco_loc["rightclick_menu_add_start_gear"] .. " " .. settings.icon[iconId].name
	           contextEntries.menu_remove_gear4_text     = fco_loc["rightclick_menu_remove_start_gear"] .. " " .. settings.icon[iconId].name
			   contextEntries.menu_add_all_gear4_text	  = fco_loc["rightclick_menu_add_all_start_gear"] .. " " .. settings.icon[iconId].name
	           contextEntries.menu_remove_all_gear4_text = fco_loc["rightclick_menu_remove_all_start_gear"] .. " " .. settings.icon[iconId].name
	        else
	           contextEntries.menu_add_gear4_text 		= fco_loc["rightclick_menu_add_gear" .. tostring(gearId)]
	           contextEntries.menu_remove_gear4_text 		= fco_loc["rightclick_menu_remove_gear" .. tostring(gearId)]
			   contextEntries.menu_add_all_gear4_text 	= fco_loc["rightclick_menu_add_all_gear" .. tostring(gearId)]
	           contextEntries.menu_remove_all_gear4_text  = fco_loc["rightclick_menu_remove_all_gear" .. tostring(gearId)]
            end
        elseif (gearId == 5) then
        	iconId = 8
			if(settings.icon[iconId].name ~= nil and settings.icon[iconId].name ~= '' and settings.icon[iconId].name ~= fco_loc["rightclick_menu_add_gear" .. tostring(gearId)]) then
			   contextEntries.menu_add_gear5_text	   	  = fco_loc["rightclick_menu_add_start_gear"] .. " " .. settings.icon[iconId].name
	           contextEntries.menu_remove_gear5_text     = fco_loc["rightclick_menu_remove_start_gear"] .. " " .. settings.icon[iconId].name
			   contextEntries.menu_add_all_gear5_text	  = fco_loc["rightclick_menu_add_all_start_gear"] .. " " .. settings.icon[iconId].name
	           contextEntries.menu_remove_all_gear5_text = fco_loc["rightclick_menu_remove_all_start_gear"] .. " " .. settings.icon[iconId].name
			else
			   contextEntries.menu_add_gear5_text 		= fco_loc["rightclick_menu_add_gear" .. tostring(gearId)]
			   contextEntries.menu_remove_gear5_text 		= fco_loc["rightclick_menu_remove_gear" .. tostring(gearId)]
			   contextEntries.menu_add_all_gear5_text 	= fco_loc["rightclick_menu_add_all_gear" .. tostring(gearId)]
			   contextEntries.menu_remove_all_gear5_text  = fco_loc["rightclick_menu_remove_all_gear" .. tostring(gearId)]
			end
        end
	end

	--Set texts for the right-click item menus
	lTextMark = {
		        contextEntries.menu_add_lock_text,
                contextEntries.menu_add_gear1_text,
                contextEntries.menu_add_research_text,
                contextEntries.menu_add_gear2_text,
                contextEntries.menu_add_sell_text,
                contextEntries.menu_add_gear3_text,
                contextEntries.menu_add_gear4_text,
                contextEntries.menu_add_gear5_text
	}
	lTextDemark = {
		        contextEntries.menu_remove_lock_text,
                contextEntries.menu_remove_gear1_text,
                contextEntries.menu_remove_research_text,
                contextEntries.menu_remove_gear2_text,
                contextEntries.menu_remove_sell_text,
                contextEntries.menu_remove_gear3_text,
                contextEntries.menu_remove_gear4_text,
                contextEntries.menu_remove_gear5_text
    }
	--Set texts for the right-click equipment item menus
	lTextEquipmentMark = {
		        contextEntries.menu_add_lock_text,
                contextEntries.menu_add_all_gear1_text,
                contextEntries.menu_add_research_text,
                contextEntries.menu_add_all_gear2_text,
                contextEntries.menu_add_sell_text,
                contextEntries.menu_add_all_gear3_text,
                contextEntries.menu_add_all_gear4_text,
                contextEntries.menu_add_all_gear5_text
	}
	lTextEquipmentDemark = {
		        contextEntries.menu_remove_lock_text,
                contextEntries.menu_remove_all_gear1_text,
                contextEntries.menu_remove_research_text,
                contextEntries.menu_remove_all_gear2_text,
                contextEntries.menu_remove_sell_text,
                contextEntries.menu_remove_all_gear3_text,
                contextEntries.menu_remove_all_gear4_text,
                contextEntries.menu_remove_all_gear5_text
    }
end

--Add/Change the filter buttons in inventory
local function updateFilterButtonsInInv(buttonId)

-- This function will only add the inventory buttons!
-- All other buttons will be added as the relating panel (bank, store, deconstruction, etc.) will be shown the first time.

    debugMessage("[UpdateFilterButtonsInInv] buttonId: " .. tostring(buttonId) .. ", numFilters: ".. tostring(locVars.gFCONumFilters) .. ", InvFiltering: " .. tostring(settings.allowInventoryFilter), true)

    -- Update the filter enable/disable buttons to the inventory, bank, crafting stations, enchantment station, guild store, guild bank, vendor, trade and mail panels
	if (buttonId == nil or buttonId == -1) then
		for buttonNr=1, locVars.gFCONumFilters, 1 do
            debugMessage("[UpdateFilterButtonsInInv] Next buttonId:" .. tostring(buttonNr), true)
			locVars.gFilterWhere = getFilterWhereBySettings(INV_BAGS)
			AddOrChangeFilterButton(INV, buttonNr, locVars.gFilterButtonWidth, locVars.gFilterButtonHeight, settings.filterButtonLeft[buttonNr], settings.filterButtonTop[buttonNr], not settings.allowInventoryFilter, locVars.gFilterWhere)
		end
    else
		locVars.gFilterWhere = getFilterWhereBySettings(INV_BAGS)
		AddOrChangeFilterButton(INV, buttonId, locVars.gFilterButtonWidth, locVars.gFilterButtonHeight, settings.filterButtonLeft[buttonId], settings.filterButtonTop[buttonId], not settings.allowInventoryFilter, locVars.gFilterWhere)
    end
end

--Show the FCO ItemSaver settings panel
local function ShowFCOItemSaverSettings()
	LAM:OpenToPanel(FCOSettingsPanel)
end

--this function is called as the context menu entry is clicked
local function contextMenuOnClicked(button)
	contextMenu[locVars.gFilterWhere]:SetHidden(true)
	local buttonName = button:GetName()
	local buttonNameLen = string.len(addonVars.gAddonName) + 1
	local sbegin
    local send
    sbegin, send = string.find(buttonName, "_", buttonNameLen)
	local buttonNr = string.sub(buttonName, buttonNameLen, sbegin - 1)
	--local buttonNameTotalLen = string.len(buttonName) - 1
    --local buttonEnd = string.sub(buttonName, buttonNameTotalLen)

	sbegin, sende = string.find(buttonNr, "UNDO")
	local isUNDOButton = sbegin ~= nil or false
	sbegin, sende = string.find(buttonNr, "REMOVEALLGEARS")
	local isREMOVEALLGEARSButton = sbegin ~= nil or false
	sbegin, sende = string.find(buttonNr, "REMOVEALL")
	local isREMOVEALLButton = sbegin ~= nil or false

    local atLeastOneMarkerChanged = false

    --Are we marking/unmarking items or are we undoing the last change at this current panel?
    if not isUNDOButton and not isREMOVEALLGEARSButton and not isREMOVEALLButton then

		--Mark/Unmark items
	    local iconId = buttonContextMenuToIconId[buttonNr].iconId
	    local doMark = buttonContextMenuToIconId[buttonNr].mark
	    debugMessage("Clicked context menu button " .. buttonNr .. ", IconId: " .. tostring(iconId) .. ", Mark: " .. tostring(doMark), true)

		if iconId ~= nil and doMark ~= nil then
		    --Get the current inventorytype
		    local inventoryType = InvToInventoryType[locVars.gFilterWhere]
		    --All non-filtered items will be in this list here:
			--PLAYER_INVENTORY.inventories[inventoryType].data[1-28].data   .bagId & ... .slotIndex
			local data
			local bagId
			local slotIndex
			local myItemInstanceId
	        local allowedToMark = false
            local undoTableCleared = false

	        --Loop over each not-filtered item data in the current inventory
	        for _,v in pairs(PLAYER_INVENTORY.inventories[inventoryType].listView.data) do
                --Get the data from current unfiltered inventory item
            	data = v.data
		        if v ~= nil and data ~= nil then
					--get the bag and slot from current unfiltered inventory item
		        	local bagId     = data.bagId
		            local slotIndex = data.slotIndex
					if bagId ~= nil and slotIndex ~= nil then
				        allowedToMark = false

	                    -- Check if equipment gear 1, 2, 3, 4, 5 or research is possible
						if (iconId == 2 or iconId == 3 or iconId == 4 or iconId == 6 or iconId == 7 or iconId == 8) then
							-- Check if item is researchable (as only researchable items can work as equipment too)
							allowedToMark = isItemResearchableNoControl(bagId, slotIndex)
	                    else
	                    	allowedToMark = true
						end

						--Is the item allowed to be marked with this iconId?
	                    if allowedToMark == true then
							myItemInstanceId = MyGetItemInstanceIdNoControl(bagId, slotIndex)
			                if myItemInstanceId ~= nil then
								--Clear the undo table once at the current panelId (keep all other panelIds !)
	    						if not undoTableCleared then
		    						undoMarkedItems[locVars.gFilterWhere] = {}
									undoMarkedItems[locVars.gFilterWhere][bagId] = {}
	                                undoTableCleared = true
	                            end

                                --Create the more dimensional arrays that are needed
			   	            	undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex] = {}
			   	            	undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex][iconId] = {}

                                --Mark: True
                                if doMark == true then
									--Check if the item is not marked already
				                    if not markedItems[iconId][myItemInstanceId] then

				                    	markedItems[iconId][myItemInstanceId] = true
		                                atLeastOneMarkerChanged = true

                                        --Old value: False
			   	            			undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex][iconId] = false
		                            end
                                --Mark: False
			                    else
									--Check if the item is marked already
				                    if markedItems[iconId][myItemInstanceId] then

				                    	markedItems[iconId][myItemInstanceId] = nil
		                                atLeastOneMarkerChanged = true

                                        --Old value: True
			   	            			undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex][iconId] = true
		                            end
			                    end
			                end

	                    end -- if allowedToMark == true ...
		            end
		        end
		    end --for _,v in pairs(PLAYER_INV...

		end

    else -- if not isUNDOButton then ...

		--UNDO
		if isUNDOButton then
	    	--Undo the last change at this panel Id
		    debugMessage("Clicked context menu button " .. buttonNr .. ". Will undo last change at panel " .. tostring(locVars.gFilterWhere) .. " now!", true)

	        atLeastOneMarkerChanged = false

			--Is there a backup set for the current panelId ?
	        if undoMarkedItems[locVars.gFilterWhere] then
	            local myItemInstanceId

	        	--Read the backup data for the current panel
		        for bagId, bagIdTab in pairs (undoMarkedItems[locVars.gFilterWhere]) do
                	for slotIndex, slotIndexTab in pairs(bagIdTab) do
                		for iconId, marked in pairs(slotIndexTab) do

							--d("[UNDO] slotIndex: " .. tostring(slotIndex) .. ", bag: " .. tostring(bagId) .. ", iconId: " .. tostring(iconId) .. ", marked: " .. tostring(marked))
							if bagId ~= nil and slotIndex ~= nil and iconId ~= nil and marked ~= nil then
								myItemInstanceId = MyGetItemInstanceIdNoControl(bagId, slotIndex)
				                if myItemInstanceId ~= nil then
									--Undo the last changes now
			                    	markedItems[iconId][myItemInstanceId] = marked
							        atLeastOneMarkerChanged = true

                                    --Update the undo table with the previous "mark" value so we are able to redo the undo again
                                    undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex][iconId] = not marked
			                    end
			               	end

                        end -- for _, iconId in pairs(slotIndex) do
                    end -- for _, slotIndex in pairs(bagId) do
	            end -- for _, bagId in pairs (undoMarkedItems[locVars.gFilterWhere]) do

	        end --if undoMarkedItems[locVars.gFilterWhere] and #undoMarkedItems[locVars.gFilterWhere] > 0 then

		--REMOVE ALL GEARS
        elseif isREMOVEALLGEARSButton then

		    debugMessage("Clicked context menu button " .. buttonNr .. ", Remove ALL GEARS", true)

		    --Get the current inventorytype
		    local inventoryType = InvToInventoryType[locVars.gFilterWhere]
		    --All non-filtered items will be in this list here:
			--PLAYER_INVENTORY.inventories[inventoryType].data[1-28].data   .bagId & ... .slotIndex
			local data
			local bagId
			local slotIndex
			local myItemInstanceId
	        local allowedToMark = false
            local undoTableCleared = false

	        --Loop over each not-filtered item data in the current inventory
	        for _,v in pairs(PLAYER_INVENTORY.inventories[inventoryType].listView.data) do
                --Initialize the undo item table
                allowedToMark = false

                --Get the data from current unfiltered inventory item
            	data = v.data
		        if v ~= nil and data ~= nil then
					--get the bag and slot from current unfiltered inventory item
		        	local bagId     = data.bagId
		            local slotIndex = data.slotIndex
					if bagId ~= nil and slotIndex ~= nil then
						-- Check if item is researchable (as only researchable items can work as equipment too)
						allowedToMark = isItemResearchableNoControl(bagId, slotIndex)

						--Is the item allowed to be marked with this iconId?
	                    if allowedToMark == true then

							myItemInstanceId = MyGetItemInstanceIdNoControl(bagId, slotIndex)
			                if myItemInstanceId ~= nil then

                                --Check all equipment gear icon IDs: 2, 4, 6, 7 and 8
								for i = 1, 5, 1 do
									--Map the iconIds of the 5 gear sets to the actual counter
                                	if     i==1 then
                                    	iconId = 2
                                    elseif i==2 then
                                    	iconId = 4
                                    elseif i==3 then
                                    	iconId = 6
                                    elseif i==4 then
                                    	iconId = 7
                                    elseif i==5 then
                                    	iconId = 8
                                    end

									--Check if the item is marked already
				                    if markedItems[iconId][myItemInstanceId] then

										--Clear the undo table once at the current panelId (keep all other panelIds !)
			    						if not undoTableCleared then
				    						undoMarkedItems[locVars.gFilterWhere] = {}
											undoMarkedItems[locVars.gFilterWhere][bagId] = {}
			                                undoTableCleared = true
			                            end

                                        --Only add the new array if not already entries in there
										if undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex] == nil then
						   	            	undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex] = {}
						   	            	undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex][iconId] = {}
                                        end

                       					--d("[REMOVE ALL GEARS] ADD slotIndex: " .. tostring(slotIndex) .. ", bag: " .. tostring(bagId) .. ", iconId: " .. tostring(iconId) .. ", marked: false")

                                        --Remove the marker for the current gear set item
				                    	markedItems[iconId][myItemInstanceId] = nil

										--Set the old marker value in the undo table
                                        undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex][iconId] = true

                                        atLeastOneMarkerChanged = true
		                            end
                                end -- for i = 1, 5, 1 do ...
			                end

	                    end -- if allowedToMark == true ...
		            end
		        end
		    end --for _,v in pairs(PLAYER_INV...


		--REMOVE ALL
        elseif isREMOVEALLButton then

		    debugMessage("Clicked context menu button " .. buttonNr .. ", Remove ALL", true)

		    --Get the current inventorytype
		    local inventoryType = InvToInventoryType[locVars.gFilterWhere]
		    --All non-filtered items will be in this list here:
			--PLAYER_INVENTORY.inventories[inventoryType].data[1-28].data   .bagId & ... .slotIndex
			local data
			local bagId
			local slotIndex
			local myItemInstanceId
            local undoTableCleared = false

	        --Loop over each not-filtered item data in the current inventory
	        for _,v in pairs(PLAYER_INVENTORY.inventories[inventoryType].listView.data) do
                --Initialize the undo item table
                allowedToMark = false

                --Get the data from current unfiltered inventory item
            	data = v.data
		        if v ~= nil and data ~= nil then
					--get the bag and slot from current unfiltered inventory item
		        	local bagId     = data.bagId
		            local slotIndex = data.slotIndex
					if bagId ~= nil and slotIndex ~= nil then

						myItemInstanceId = MyGetItemInstanceIdNoControl(bagId, slotIndex)
						if myItemInstanceId ~= nil then
							--Check all icon Ids
							for iconId = 1, locVars.gFCONumFilterIcons, 1 do
								--Check if the item is marked already
								if markedItems[iconId][myItemInstanceId] then
									--Clear the undo table once at the current panelId (keep all other panelIds !)
									if not undoTableCleared then
										undoMarkedItems[locVars.gFilterWhere] = {}
										undoMarkedItems[locVars.gFilterWhere][bagId] = {}
										undoTableCleared = true
									end

									--Only add the new array if not already entries in there
									if undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex] == nil then
										undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex] = {}
										undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex][iconId] = {}
									end

									--d("[REMOVE ALL] ADD slotIndex: " .. tostring(slotIndex) .. ", bag: " .. tostring(bagId) .. ", iconId: " .. tostring(iconId) .. ", marked: false")

									--Remove the marker for the current gear set item
									markedItems[iconId][myItemInstanceId] = nil

									--Set the old marker value in the undo table
									undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex][iconId] = true

									atLeastOneMarkerChanged = true
								end
							end -- for iconId = 1, locVars.gFCONumFilterIcons, 1 do
                        end

		            end
		        end
		    end --for _,v in pairs(PLAYER_INV...

        end

    end -- if buttonNr ~= "ButtonContextMenuUNDO" then


	--Did at least one marker change?
	if atLeastOneMarkerChanged == true then
		--Update the inventory & markers
		FilterBasics(false)
	end

end

--this function is called as the context menu entry is clicked at a crafting station
local function contextMenuOnClickedSmithing(button)
	contextMenu[locVars.gFilterWhere]:SetHidden(true)
	local buttonName = button:GetName()
	local buttonNameLen = string.len(addonVars.gAddonName) + 1
	local sbegin
    local send
    sbegin, send = string.find(buttonName, "_", buttonNameLen)
	local buttonNr = string.sub(buttonName, buttonNameLen, sbegin - 1)
	--local buttonNameTotalLen = string.len(buttonName) - 1
    --local buttonEnd = string.sub(buttonName, buttonNameTotalLen)

	sbegin, sende = string.find(buttonNr, "UNDO")
	local isUNDOButton = sbegin ~= nil or false
	sbegin, sende = string.find(buttonNr, "REMOVEALLGEARS")
	local isREMOVEALLGEARSButton = sbegin ~= nil or false
	sbegin, sende = string.find(buttonNr, "REMOVEALL")
	local isREMOVEALLButton = sbegin ~= nil or false

    local atLeastOneMarkerChanged = false

	local INVENTORY_TO_SEARCH
	--Deconstruction panel?
	if     (locVars.gFilterWhere == INV_DECONSTRUCTION) then
		INVENTORY_TO_SEARCH = DECONSTRUCTION
	elseif (locVars.gFilterWhere == INV_IMPROVEMENT) then
	--Improvement panel?
		INVENTORY_TO_SEARCH = IMPROVEMENT
	end

	--No inventory to search in given? Abort here!
	if INVENTORY_TO_SEARCH == nil then return end

    --Are we marking/unmarking items or are we undoing the last change at this current panel?
    if not isUNDOButton and not isREMOVEALLGEARSButton and not isREMOVEALLButton then

		--Mark/Unmark items
	    local iconId = buttonContextMenuToIconId[buttonNr].iconId
	    local doMark = buttonContextMenuToIconId[buttonNr].mark
	    debugMessage("Clicked SMITHING context menu button " .. buttonNr .. ", IconId: " .. tostring(iconId) .. ", Mark: " .. tostring(doMark), true)

		if iconId ~= nil and doMark ~= nil then
		    --Get the current inventorytype
		    local inventoryType = InvToInventoryType[locVars.gFilterWhere]
		    --All non-filtered items will be in this list here:
			--ZO_SmithingTopLevelDeconstructionPanelInventoryBackpack.data[]   .bagId & ... .slotIndex
			--ZO_SmithingTopLevelImprovementPanelInventoryBackpack.data[]   .bagId & ... .slotIndex
			local data
			local bagId
			local slotIndex
			local myItemInstanceId
	        local allowedToMark = false
            local undoTableCleared = false

	        --Loop over each not-filtered item data in the current inventory
	        for _,v in pairs(INVENTORY_TO_SEARCH.data) do
                --Get the data from current unfiltered inventory item
            	data = v.data
		        if v ~= nil and data ~= nil then
					--get the bag and slot from current unfiltered inventory item
		        	local bagId     = data.bagId
		            local slotIndex = data.slotIndex
					if bagId ~= nil and slotIndex ~= nil then
				        allowedToMark = false

	                    -- Check if equipment gear 1, 2, 3, 4, 5 or research is possible
						if (iconId == 2 or iconId == 3 or iconId == 4 or iconId == 6 or iconId == 7 or iconId == 8) then
							-- Check if item is researchable (as only researchable items can work as equipment too)
							allowedToMark = isItemResearchableNoControl(bagId, slotIndex)
	                    else
	                    	allowedToMark = true
						end

						--Is the item allowed to be marked with this iconId?
	                    if allowedToMark == true then
							myItemInstanceId = MyGetItemInstanceIdNoControl(bagId, slotIndex)
			                if myItemInstanceId ~= nil then
								--Clear the undo table once at the current panelId (keep all other panelIds !)
	    						if not undoTableCleared then
		    						undoMarkedItems[locVars.gFilterWhere] = {}
									undoMarkedItems[locVars.gFilterWhere][bagId] = {}
	                                undoTableCleared = true
	                            end

                                --Create the more dimensional arrays that are needed
			   	            	undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex] = {}
			   	            	undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex][iconId] = {}

                                --Mark: True
                                if doMark == true then
									--Check if the item is not marked already
				                    if not markedItems[iconId][myItemInstanceId] then

				                    	markedItems[iconId][myItemInstanceId] = true
		                                atLeastOneMarkerChanged = true

                                        --Old value: False
			   	            			undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex][iconId] = false
		                            end
                                --Mark: False
			                    else
									--Check if the item is marked already
				                    if markedItems[iconId][myItemInstanceId] then

				                    	markedItems[iconId][myItemInstanceId] = nil
		                                atLeastOneMarkerChanged = true

                                        --Old value: True
			   	            			undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex][iconId] = true
		                            end
			                    end
			                end

	                    end -- if allowedToMark == true ...
		            end
		        end
		    end --for _,v in pairs(PLAYER_INV...

		end

    else -- if not isUNDOButton then ...

		--UNDO
		if isUNDOButton then
	    	--Undo the last change at this panel Id
		    debugMessage("Clicked SMITHING context menu button " .. buttonNr .. ". Will undo last change at panel " .. tostring(locVars.gFilterWhere) .. " now!", true)

	        atLeastOneMarkerChanged = false

			--Is there a backup set for the current panelId ?
	        if undoMarkedItems[locVars.gFilterWhere] then
	            local myItemInstanceId

	        	--Read the backup data for the current panel
		        for bagId, bagIdTab in pairs (undoMarkedItems[locVars.gFilterWhere]) do
                	for slotIndex, slotIndexTab in pairs(bagIdTab) do
                		for iconId, marked in pairs(slotIndexTab) do

							--d("[UNDO] slotIndex: " .. tostring(slotIndex) .. ", bag: " .. tostring(bagId) .. ", iconId: " .. tostring(iconId) .. ", marked: " .. tostring(marked))
							if bagId ~= nil and slotIndex ~= nil and iconId ~= nil and marked ~= nil then
								myItemInstanceId = MyGetItemInstanceIdNoControl(bagId, slotIndex)
				                if myItemInstanceId ~= nil then
									--Undo the last changes now
			                    	markedItems[iconId][myItemInstanceId] = marked
							        atLeastOneMarkerChanged = true

                                    --Update the undo table with the previous "mark" value so we are able to redo the undo again
                                    undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex][iconId] = not marked
			                    end
			               	end

                        end -- for _, iconId in pairs(slotIndex) do
                    end -- for _, slotIndex in pairs(bagId) do
	            end -- for _, bagId in pairs (undoMarkedItems[locVars.gFilterWhere]) do

	        end --if undoMarkedItems[locVars.gFilterWhere] and #undoMarkedItems[locVars.gFilterWhere] > 0 then

		--REMOVE ALL GEARS
        elseif isREMOVEALLGEARSButton then

		    debugMessage("Clicked SMITHING context menu button " .. buttonNr .. ", Remove ALL GEARS", true)

		    --Get the current inventorytype
		    local inventoryType = InvToInventoryType[locVars.gFilterWhere]
		    --All non-filtered items will be in this list here:
			--PLAYER_INVENTORY.inventories[inventoryType].data[1-28].data   .bagId & ... .slotIndex
			local data
			local bagId
			local slotIndex
			local myItemInstanceId
	        local allowedToMark = false
            local undoTableCleared = false

	        --Loop over each not-filtered item data in the current inventory
			for _,v in pairs(INVENTORY_TO_SEARCH.data) do
                --Initialize the undo item table
                allowedToMark = false

                --Get the data from current unfiltered inventory item
            	data = v.data
		        if v ~= nil and data ~= nil then
					--get the bag and slot from current unfiltered inventory item
		        	local bagId     = data.bagId
		            local slotIndex = data.slotIndex
					if bagId ~= nil and slotIndex ~= nil then
						-- Check if item is researchable (as only researchable items can work as equipment too)
						allowedToMark = isItemResearchableNoControl(bagId, slotIndex)

						--Is the item allowed to be marked with this iconId?
	                    if allowedToMark == true then

							myItemInstanceId = MyGetItemInstanceIdNoControl(bagId, slotIndex)
			                if myItemInstanceId ~= nil then

                                --Check all equipment gear icon IDs: 2, 4, 6, 7 and 8
								for i = 1, 5, 1 do
									--Map the iconIds of the 5 gear sets to the actual counter
                                	if     i==1 then
                                    	iconId = 2
                                    elseif i==2 then
                                    	iconId = 4
                                    elseif i==3 then
                                    	iconId = 6
                                    elseif i==4 then
                                    	iconId = 7
                                    elseif i==5 then
                                    	iconId = 8
                                    end

									--Check if the item is marked already
				                    if markedItems[iconId][myItemInstanceId] then

										--Clear the undo table once at the current panelId (keep all other panelIds !)
			    						if not undoTableCleared then
				    						undoMarkedItems[locVars.gFilterWhere] = {}
											undoMarkedItems[locVars.gFilterWhere][bagId] = {}
			                                undoTableCleared = true
			                            end

                                        --Only add the new array if not already entries in there
										if undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex] == nil then
						   	            	undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex] = {}
						   	            	undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex][iconId] = {}
                                        end

                       					--d("[REMOVE ALL GEARS] ADD slotIndex: " .. tostring(slotIndex) .. ", bag: " .. tostring(bagId) .. ", iconId: " .. tostring(iconId) .. ", marked: false")

                                        --Remove the marker for the current gear set item
				                    	markedItems[iconId][myItemInstanceId] = nil

										--Set the old marker value in the undo table
                                        undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex][iconId] = true

                                        atLeastOneMarkerChanged = true
		                            end
                                end -- for i = 1, 5, 1 do ...
			                end

	                    end -- if allowedToMark == true ...
		            end
		        end
		    end --for _,v in pairs(PLAYER_INV...


		--REMOVE ALL
        elseif isREMOVEALLButton then

		    debugMessage("Clicked SMITHING context menu button " .. buttonNr .. ", Remove ALL", true)

		    --Get the current inventorytype
		    local inventoryType = InvToInventoryType[locVars.gFilterWhere]
		    --All non-filtered items will be in this list here:
			--PLAYER_INVENTORY.inventories[inventoryType].data[1-28].data   .bagId & ... .slotIndex
			local data
			local bagId
			local slotIndex
			local myItemInstanceId
            local undoTableCleared = false

	        --Loop over each not-filtered item data in the current inventory
			for _,v in pairs(INVENTORY_TO_SEARCH.data) do
                --Initialize the undo item table
                allowedToMark = false

                --Get the data from current unfiltered inventory item
            	data = v.data
		        if v ~= nil and data ~= nil then
					--get the bag and slot from current unfiltered inventory item
		        	local bagId     = data.bagId
		            local slotIndex = data.slotIndex
					if bagId ~= nil and slotIndex ~= nil then

						myItemInstanceId = MyGetItemInstanceIdNoControl(bagId, slotIndex)
						if myItemInstanceId ~= nil then
							--Check all icon Ids
							for iconId = 1, locVars.gFCONumFilterIcons, 1 do
								--Check if the item is marked already
								if markedItems[iconId][myItemInstanceId] then
									--Clear the undo table once at the current panelId (keep all other panelIds !)
									if not undoTableCleared then
										undoMarkedItems[locVars.gFilterWhere] = {}
										undoMarkedItems[locVars.gFilterWhere][bagId] = {}
										undoTableCleared = true
									end

									--Only add the new array if not already entries in there
									if undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex] == nil then
										undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex] = {}
										undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex][iconId] = {}
									end

									--d("[REMOVE ALL] ADD slotIndex: " .. tostring(slotIndex) .. ", bag: " .. tostring(bagId) .. ", iconId: " .. tostring(iconId) .. ", marked: false")

									--Remove the marker for the current gear set item
									markedItems[iconId][myItemInstanceId] = nil

									--Set the old marker value in the undo table
									undoMarkedItems[locVars.gFilterWhere][bagId][slotIndex][iconId] = true

									atLeastOneMarkerChanged = true
								end
							end -- for iconId = 1, locVars.gFCONumFilterIcons, 1 do
                        end

		            end
		        end
		    end --for _,v in pairs(PLAYER_INV...

        end

    end -- if buttonNr ~= "ButtonContextMenuUNDO" then


	--Did at least one marker change?
	if atLeastOneMarkerChanged == true then
		--Update the inventory & markers
		FilterBasics(false)
	end

end

--The nutton's context menu OnMouseExit event callback function
local function contextMenuOnMouseExit(button)
    if bdSelectedLine[locVars.gFilterWhere] ~= nil then
		bdSelectedLine[locVars.gFilterWhere]:SetHidden(true)
		local labelControl = button:GetLabelControl()
	    if labelControl ~= nil then
	    	labelControl:SetColor(ZO_DISABLED_TEXT:UnpackRGBA())
	    end
    end
end

--The nutton's context menu OnMouseEnter event callback function
local function contextMenuOnMouseEnter(button)
	--Create the background for the selected line
    if bdSelectedLine[locVars.gFilterWhere] == nil then
		bdSelectedLine[locVars.gFilterWhere] = WINDOW_MANAGER:CreateControl("bdBackDrop"..tostring(locVars.gFilterWhere), button, CT_BACKDROP)
		bdSelectedLine[locVars.gFilterWhere]:SetHidden(true)
		--bdSelectedLine[locVars.gFilterWhere]:SetEdgeColor(0.4,0.4,0.4)
		--bdSelectedLine[locVars.gFilterWhere]:SetCenterColor(0.1,0.1,0.1)
		bdSelectedLine[locVars.gFilterWhere]:SetDimensions(250,20)
		bdSelectedLine[locVars.gFilterWhere]:SetAlpha(1)
		bdSelectedLine[locVars.gFilterWhere]:SetDrawLayer(0)
        local menuInfo = menuInfo[MENU_TYPE_DEFAULT]
		bdSelectedLine[locVars.gFilterWhere]:SetCenterTexture(menuInfo.backdropCenter)
        bdSelectedLine[locVars.gFilterWhere]:SetEdgeTexture(menuInfo.backdropEdge, menuInfo.backdropEdgeWidth, menuInfo.backdropEdgeHeight)
        bdSelectedLine[locVars.gFilterWhere]:SetInsets(unpack(menuInfo.backdropInsets))
    end
	bdSelectedLine[locVars.gFilterWhere]:ClearAnchors()
	bdSelectedLine[locVars.gFilterWhere]:SetAnchor(TOPLEFT, button, TOPLEFT, -4, -2)
    bdSelectedLine[locVars.gFilterWhere]:SetAnchor(BOTTOMRIGHT, button, BOTTOMRIGHT, 0, 2)
	bdSelectedLine[locVars.gFilterWhere]:SetHidden(false)

	local labelControl = button:GetLabelControl()
    if labelControl ~= nil then
    	labelControl:SetColor(ZO_HIGHLIGHT_TEXT:UnpackRGBA())
    end
end

--Function that creates the context menu. Adding to it buttons and the handlers for them
local function createContextMenu(parentWindow, panelId)
	local button
	local labelControl
	--Abort if parent does not exist
    if parentWindow == nil then return end
	panelId = panelId or locVars.gFilterWhere
	local filterPanelId = panelId
    if filterPanelId < 10 then filterPanelId = "0" .. tostring(filterPanelId) end

    local parentName = parentWindow:GetName()
	local onClickedFunction = "normal"
    if 		string.find(parentName, "ZO_SmithingTopLevelDeconstruction") then
    	onClickedFunction = "deconstruction"
    elseif  string.find(parentName, "ZO_SmithingTopLevelImprovement") then
    	onClickedFunction = "improvement"
    elseif  string.find(parentName, "ZO_Enchanting") then
    	onClickedFunction = "enchanting"
	else
    	onClickedFunction = "normal"
    end
    --Array with callback functions OnClicked
    local onClickedFunctions = {
	    ["normal"] 			= contextMenuOnClicked,
	    ["deconstruction"]  = contextMenuOnClickedSmithing,
	    ["improvement"]  	= contextMenuOnClickedSmithing,
	    ["enchanting"]  	= contextMenuOnClickedEnchanting,
    }

    contextMenu[panelId] = WINDOW_MANAGER:CreateTopLevelWindow(addonVars.gAddonName .. "ContextMenu" .. "_" .. tostring(filterPanelId))
	local menuWindow = contextMenu[panelId]
	menuWindow:SetDimensions(250,381)
    menuWindow:SetAnchor(TOP,parentWindow,BOTTOM, 12, 6)
    menuWindow:SetMovable(false)
    menuWindow:SetMouseEnabled(true)

	menuWindow.bg = WINDOW_MANAGER:CreateControlFromVirtual(addonVars.gAddonName .. "ContextMenuBg" .. "_" .. tostring(filterPanelId), menuWindow, "ZO_DefaultBackdrop")

	menuWindow:SetDrawLevel(0)
	menuWindow:SetDrawLayer(1)
	menuWindow:SetDrawTier(1)

	button = WINDOW_MANAGER:CreateControl(addonVars.gAddonName .. "ButtonContextMenu1".. "_" .. tostring(filterPanelId) , menuWindow , CT_BUTTON)
	button:SetDimensions( 250,20 )
	button:SetFont("ZoFontGame")
	button:SetState( BSTATE_NORMAL )
	button:SetHandler( "OnClicked", function(...) onClickedFunctions[onClickedFunction](...) end)
	button:SetHandler( "OnMouseEnter", function (self) contextMenuOnMouseEnter(self) end)
	button:SetHandler( "OnMouseExit", function (self) contextMenuOnMouseExit(self) end)
	button:SetText(fco_loc["button_context_menu_lock_all"])
	button:SetHorizontalAlignment(TEXT_ALIGN_LEFT)
	button:SetAnchor(TOP,menuWindow,TOP,0,0)
	labelControl = button:GetLabelControl()
    if labelControl ~= nil then
    	labelControl:SetColor(ZO_NORMAL_TEXT:UnpackRGBA())
    end
    button = nil
    labelControl = nil

	button = WINDOW_MANAGER:CreateControl(addonVars.gAddonName .. "ButtonContextMenu2".. "_" .. tostring(filterPanelId) , menuWindow , CT_BUTTON)
	button:SetDimensions( 250,20 )
	button:SetFont("ZoFontGame")
	button:SetState( BSTATE_NORMAL )
	button:SetHandler( "OnClicked", function(...) onClickedFunctions[onClickedFunction](...) end)
	button:SetHandler( "OnMouseEnter", function (self) contextMenuOnMouseEnter(self) end)
	button:SetHandler( "OnMouseExit", function (self) contextMenuOnMouseExit(self) end)
	button:SetText(fco_loc["button_context_menu_unlock_all"])
	button:SetHorizontalAlignment(TEXT_ALIGN_LEFT)
	button:SetAnchor(TOP,menuWindow,TOP,0,21)
	labelControl = button:GetLabelControl()
    if labelControl ~= nil then
    	labelControl:SetColor(ZO_NORMAL_TEXT:UnpackRGBA())
    end
    button = nil
    labelControl = nil

	button = WINDOW_MANAGER:CreateControl(addonVars.gAddonName .. "ButtonContextMenu3".. "_" .. tostring(filterPanelId) , menuWindow , CT_BUTTON)
	button:SetDimensions( 250,20 )
	button:SetFont("ZoFontGame")
	button:SetState( BSTATE_NORMAL )
	button:SetHandler( "OnClicked", function(...) onClickedFunctions[onClickedFunction](...) end)
	button:SetHandler( "OnMouseEnter", function (self) contextMenuOnMouseEnter(self) end)
	button:SetHandler( "OnMouseExit", function (self) contextMenuOnMouseExit(self) end)
	button:SetText(contextEntries.menu_add_all_gear1_text)
	button:SetHorizontalAlignment(TEXT_ALIGN_LEFT)
	button:SetAnchor(TOP,menuWindow,TOP,0,41)
 	labelControl = button:GetLabelControl()
    if labelControl ~= nil then
    	labelControl:SetColor(ZO_NORMAL_TEXT:UnpackRGBA())
    end
    button = nil
    labelControl = nil

	button = WINDOW_MANAGER:CreateControl(addonVars.gAddonName .. "ButtonContextMenu4".. "_" .. tostring(filterPanelId) , menuWindow , CT_BUTTON)
	button:SetDimensions( 250,20 )
	button:SetFont("ZoFontGame")
	button:SetAnchor(TOP,menuWindow,TOP,0,0)
	button:SetState( BSTATE_NORMAL )
	button:SetHandler( "OnClicked", function(...) onClickedFunctions[onClickedFunction](...) end)
	button:SetHandler( "OnMouseEnter", function (self) contextMenuOnMouseEnter(self) end)
	button:SetHandler( "OnMouseExit", function (self) contextMenuOnMouseExit(self) end)
	button:SetText(contextEntries.menu_remove_all_gear1_text)
	button:SetHorizontalAlignment(TEXT_ALIGN_LEFT)
	button:SetAnchor(TOP,menuWindow,TOP,0,61)
 	labelControl = button:GetLabelControl()
    if labelControl ~= nil then
    	labelControl:SetColor(ZO_NORMAL_TEXT:UnpackRGBA())
    end
    button = nil
    labelControl = nil

	button = WINDOW_MANAGER:CreateControl(addonVars.gAddonName .. "ButtonContextMenu5".. "_" .. tostring(filterPanelId) , menuWindow , CT_BUTTON)
	button:SetDimensions( 250,20 )
	button:SetFont("ZoFontGame")
	button:SetState( BSTATE_NORMAL )
	button:SetHandler( "OnClicked", function(...) onClickedFunctions[onClickedFunction](...) end)
	button:SetHandler( "OnMouseEnter", function (self) contextMenuOnMouseEnter(self) end)
	button:SetHandler( "OnMouseExit", function (self) contextMenuOnMouseExit(self) end)
	button:SetText(contextEntries.menu_add_all_gear2_text)
	button:SetHorizontalAlignment(TEXT_ALIGN_LEFT)
	button:SetAnchor(TOP,menuWindow,TOP,0,81)
 	labelControl = button:GetLabelControl()
    if labelControl ~= nil then
    	labelControl:SetColor(ZO_NORMAL_TEXT:UnpackRGBA())
    end
    button = nil
    labelControl = nil

	button = WINDOW_MANAGER:CreateControl(addonVars.gAddonName .. "ButtonContextMenu6".. "_" .. tostring(filterPanelId) , menuWindow , CT_BUTTON)
	button:SetDimensions( 250,20 )
	button:SetFont("ZoFontGame")
	button:SetState( BSTATE_NORMAL )
	button:SetHandler( "OnClicked", function(...) onClickedFunctions[onClickedFunction](...) end)
	button:SetHandler( "OnMouseEnter", function (self) contextMenuOnMouseEnter(self) end)
	button:SetHandler( "OnMouseExit", function (self) contextMenuOnMouseExit(self) end)
	button:SetText(contextEntries.menu_remove_all_gear2_text)
	button:SetHorizontalAlignment(TEXT_ALIGN_LEFT)
	button:SetAnchor(TOP,menuWindow,TOP,0,101)
 	labelControl = button:GetLabelControl()
    if labelControl ~= nil then
    	labelControl:SetColor(ZO_NORMAL_TEXT:UnpackRGBA())
    end
    button = nil
    labelControl = nil

	button = WINDOW_MANAGER:CreateControl(addonVars.gAddonName .. "ButtonContextMenu7".. "_" .. tostring(filterPanelId) , menuWindow , CT_BUTTON)
	button:SetDimensions( 250,20 )
	button:SetFont("ZoFontGame")
	button:SetState( BSTATE_NORMAL )
	button:SetHandler( "OnClicked", function(...) onClickedFunctions[onClickedFunction](...) end)
	button:SetHandler( "OnMouseEnter", function (self) contextMenuOnMouseEnter(self) end)
	button:SetHandler( "OnMouseExit", function (self) contextMenuOnMouseExit(self) end)
	button:SetText(contextEntries.menu_add_all_gear3_text)
	button:SetHorizontalAlignment(TEXT_ALIGN_LEFT)
	button:SetAnchor(TOP,menuWindow,TOP,0,121)
	labelControl = button:GetLabelControl()
    if labelControl ~= nil then
    	labelControl:SetColor(ZO_NORMAL_TEXT:UnpackRGBA())
    end
    button = nil
    labelControl = nil

	button = WINDOW_MANAGER:CreateControl(addonVars.gAddonName .. "ButtonContextMenu8".. "_" .. tostring(filterPanelId) , menuWindow , CT_BUTTON)
	button:SetDimensions( 250,20 )
	button:SetFont("ZoFontGame")
	button:SetState( BSTATE_NORMAL )
	button:SetHandler( "OnClicked", function(...) onClickedFunctions[onClickedFunction](...) end)
	button:SetHandler( "OnMouseEnter", function (self) contextMenuOnMouseEnter(self) end)
	button:SetHandler( "OnMouseExit", function (self) contextMenuOnMouseExit(self) end)
	button:SetText(contextEntries.menu_remove_all_gear3_text)
	button:SetHorizontalAlignment(TEXT_ALIGN_LEFT)
	button:SetAnchor(TOP,menuWindow,TOP,0,141)
	labelControl = button:GetLabelControl()
    if labelControl ~= nil then
    	labelControl:SetColor(ZO_NORMAL_TEXT:UnpackRGBA())
    end
    button = nil
    labelControl = nil

	button = WINDOW_MANAGER:CreateControl(addonVars.gAddonName .. "ButtonContextMenu9".. "_" .. tostring(filterPanelId) , menuWindow , CT_BUTTON)
	button:SetDimensions( 250,20 )
	button:SetFont("ZoFontGame")
	button:SetState( BSTATE_NORMAL )
	button:SetHandler( "OnClicked", function(...) onClickedFunctions[onClickedFunction](...) end)
	button:SetHandler( "OnMouseEnter", function (self) contextMenuOnMouseEnter(self) end)
	button:SetHandler( "OnMouseExit", function (self) contextMenuOnMouseExit(self) end)
	button:SetText(contextEntries.menu_add_all_gear4_text)
	button:SetHorizontalAlignment(TEXT_ALIGN_LEFT)
	button:SetAnchor(TOP,menuWindow,TOP,0,161)
	labelControl = button:GetLabelControl()
    if labelControl ~= nil then
    	labelControl:SetColor(ZO_NORMAL_TEXT:UnpackRGBA())
    end
    button = nil
    labelControl = nil

	button = WINDOW_MANAGER:CreateControl(addonVars.gAddonName .. "ButtonContextMenu10".. "_" .. tostring(filterPanelId) , menuWindow , CT_BUTTON)
	button:SetDimensions( 250,20 )
	button:SetFont("ZoFontGame")
	button:SetState( BSTATE_NORMAL )
	button:SetHandler( "OnClicked", function(...) onClickedFunctions[onClickedFunction](...) end)
	button:SetHandler( "OnMouseEnter", function (self) contextMenuOnMouseEnter(self) end)
	button:SetHandler( "OnMouseExit", function (self) contextMenuOnMouseExit(self) end)
	button:SetText(contextEntries.menu_remove_all_gear4_text)
	button:SetHorizontalAlignment(TEXT_ALIGN_LEFT)
	button:SetAnchor(TOP,menuWindow,TOP,0,181)
	labelControl = button:GetLabelControl()
    if labelControl ~= nil then
    	labelControl:SetColor(ZO_NORMAL_TEXT:UnpackRGBA())
    end
    button = nil
    labelControl = nil

	button = WINDOW_MANAGER:CreateControl(addonVars.gAddonName .. "ButtonContextMenu11".. "_" .. tostring(filterPanelId) , menuWindow , CT_BUTTON)
	button:SetDimensions( 250,20 )
	button:SetFont("ZoFontGame")
	button:SetState( BSTATE_NORMAL )
	button:SetHandler( "OnClicked", function(...) onClickedFunctions[onClickedFunction](...) end)
	button:SetHandler( "OnMouseEnter", function (self) contextMenuOnMouseEnter(self) end)
	button:SetHandler( "OnMouseExit", function (self) contextMenuOnMouseExit(self) end)
	button:SetText(contextEntries.menu_add_all_gear5_text)
	button:SetHorizontalAlignment(TEXT_ALIGN_LEFT)
	button:SetAnchor(TOP,menuWindow,TOP,0,201)
	labelControl = button:GetLabelControl()
    if labelControl ~= nil then
    	labelControl:SetColor(ZO_NORMAL_TEXT:UnpackRGBA())
    end
    button = nil
    labelControl = nil

	button = WINDOW_MANAGER:CreateControl(addonVars.gAddonName .. "ButtonContextMenu12".. "_" .. tostring(filterPanelId) , menuWindow , CT_BUTTON)
	button:SetDimensions( 250,20 )
	button:SetFont("ZoFontGame")
	button:SetState( BSTATE_NORMAL )
	button:SetHandler( "OnClicked", function(...) onClickedFunctions[onClickedFunction](...) end)
	button:SetHandler( "OnMouseEnter", function (self) contextMenuOnMouseEnter(self) end)
	button:SetHandler( "OnMouseExit", function (self) contextMenuOnMouseExit(self) end)
	button:SetText(contextEntries.menu_remove_all_gear5_text)
	button:SetHorizontalAlignment(TEXT_ALIGN_LEFT)
	button:SetAnchor(TOP,menuWindow,TOP,0,221)
	labelControl = button:GetLabelControl()
    if labelControl ~= nil then
    	labelControl:SetColor(ZO_NORMAL_TEXT:UnpackRGBA())
    end
    button = nil
    labelControl = nil

	button = WINDOW_MANAGER:CreateControl(addonVars.gAddonName .. "ButtonContextMenu13".. "_" .. tostring(filterPanelId) , menuWindow , CT_BUTTON)
	button:SetDimensions( 250,20 )
	button:SetFont("ZoFontGame")
	button:SetState( BSTATE_NORMAL )
	button:SetHandler( "OnClicked", function(...) onClickedFunctions[onClickedFunction](...) end)
	button:SetHandler( "OnMouseEnter", function (self) contextMenuOnMouseEnter(self) end)
	button:SetHandler( "OnMouseExit", function (self) contextMenuOnMouseExit(self) end)
	button:SetText(fco_loc["button_context_menu_research_all"])
	button:SetHorizontalAlignment(TEXT_ALIGN_LEFT)
	button:SetAnchor(TOP,menuWindow,TOP,0,241)
	labelControl = button:GetLabelControl()
    if labelControl ~= nil then
    	labelControl:SetColor(ZO_NORMAL_TEXT:UnpackRGBA())
    end
    button = nil
    labelControl = nil

	button = WINDOW_MANAGER:CreateControl(addonVars.gAddonName .. "ButtonContextMenu14".. "_" .. tostring(filterPanelId) , menuWindow , CT_BUTTON)
	button:SetDimensions( 250,20 )
	button:SetFont("ZoFontGame")
	button:SetState( BSTATE_NORMAL )
	button:SetHandler( "OnClicked", function(...) onClickedFunctions[onClickedFunction](...) end)
	button:SetHandler( "OnMouseEnter", function (self) contextMenuOnMouseEnter(self) end)
	button:SetHandler( "OnMouseExit", function (self) contextMenuOnMouseExit(self) end)
	button:SetText(fco_loc["button_context_menu_dont_research_all"])
	button:SetHorizontalAlignment(TEXT_ALIGN_LEFT)
	button:SetAnchor(TOP,menuWindow,TOP,0,261)
	labelControl = button:GetLabelControl()
    if labelControl ~= nil then
    	labelControl:SetColor(ZO_NORMAL_TEXT:UnpackRGBA())
    end
    button = nil
    labelControl = nil

	button = WINDOW_MANAGER:CreateControl(addonVars.gAddonName .. "ButtonContextMenu15".. "_" .. tostring(filterPanelId) , menuWindow , CT_BUTTON)
	button:SetDimensions( 250,20 )
	button:SetFont("ZoFontGame")
	button:SetState( BSTATE_NORMAL )
	button:SetHandler( "OnClicked", function(...) onClickedFunctions[onClickedFunction](...) end)
	button:SetHandler( "OnMouseEnter", function (self) contextMenuOnMouseEnter(self) end)
	button:SetHandler( "OnMouseExit", function (self) contextMenuOnMouseExit(self) end)
	button:SetText(fco_loc["button_context_menu_sell_all"])
	button:SetHorizontalAlignment(TEXT_ALIGN_LEFT)
	button:SetAnchor(TOP,menuWindow,TOP,0,281)
	labelControl = button:GetLabelControl()
    if labelControl ~= nil then
    	labelControl:SetColor(ZO_NORMAL_TEXT:UnpackRGBA())
    end
    button = nil
    labelControl = nil

	button = WINDOW_MANAGER:CreateControl(addonVars.gAddonName .. "ButtonContextMenu16".. "_" .. tostring(filterPanelId) , menuWindow , CT_BUTTON)
	button:SetDimensions( 250,20 )
	button:SetFont("ZoFontGame")
	button:SetState( BSTATE_NORMAL )
	button:SetHandler( "OnClicked", function(...) onClickedFunctions[onClickedFunction](...) end)
	button:SetHandler( "OnMouseEnter", function (self) contextMenuOnMouseEnter(self) end)
	button:SetHandler( "OnMouseExit", function (self) contextMenuOnMouseExit(self) end)
	button:SetText(fco_loc["button_context_menu_dont_sell_all"])
	button:SetHorizontalAlignment(TEXT_ALIGN_LEFT)
	button:SetAnchor(TOP,menuWindow,TOP,0,301)
	labelControl = button:GetLabelControl()
    if labelControl ~= nil then
    	labelControl:SetColor(ZO_NORMAL_TEXT:UnpackRGBA())
    end
    button = nil
    labelControl = nil

	button = WINDOW_MANAGER:CreateControl(addonVars.gAddonName .. "ButtonContextMenuREMOVEALLGEARS".. "_" .. tostring(filterPanelId) , menuWindow , CT_BUTTON)
	button:SetDimensions( 250,20 )
	button:SetFont("ZoFontGame")
	button:SetState( BSTATE_NORMAL )
	button:SetHandler( "OnClicked", function(...) onClickedFunctions[onClickedFunction](...) end)
	button:SetHandler( "OnMouseEnter", function (self) contextMenuOnMouseEnter(self) end)
	button:SetHandler( "OnMouseExit", function (self) contextMenuOnMouseExit(self) end)
	button:SetText(fco_loc["button_context_menu_remove_all_gears"])
	button:SetHorizontalAlignment(TEXT_ALIGN_LEFT)
	button:SetAnchor(TOP,menuWindow,TOP,0,321)
	labelControl = button:GetLabelControl()
    if labelControl ~= nil then
    	labelControl:SetColor(ZO_NORMAL_TEXT:UnpackRGBA())
    end
    button = nil
    labelControl = nil

	button = WINDOW_MANAGER:CreateControl(addonVars.gAddonName .. "ButtonContextMenuREMOVEALL".. "_" .. tostring(filterPanelId) , menuWindow , CT_BUTTON)
	button:SetDimensions( 250,20 )
	button:SetFont("ZoFontGame")
	button:SetState( BSTATE_NORMAL )
	button:SetHandler( "OnClicked", function(...) onClickedFunctions[onClickedFunction](...) end)
	button:SetHandler( "OnMouseEnter", function (self) contextMenuOnMouseEnter(self) end)
	button:SetHandler( "OnMouseExit", function (self) contextMenuOnMouseExit(self) end)
	button:SetText(fco_loc["button_context_menu_demark_all"])
	button:SetHorizontalAlignment(TEXT_ALIGN_LEFT)
	button:SetAnchor(TOP,menuWindow,TOP,0,341)
	labelControl = button:GetLabelControl()
    if labelControl ~= nil then
    	labelControl:SetColor(ZO_NORMAL_TEXT:UnpackRGBA())
    end
    button = nil
    labelControl = nil

	button = WINDOW_MANAGER:CreateControl(addonVars.gAddonName .. "ButtonContextMenuUNDO" .. "_" .. tostring(filterPanelId), menuWindow , CT_BUTTON)
    button:SetDimensions( 250,20 )
	button:SetFont("ZoFontGame")
	button:SetState( BSTATE_NORMAL )
	button:SetHandler( "OnClicked", function(...) onClickedFunctions[onClickedFunction](...) end)
	button:SetHandler( "OnMouseEnter", function (self) contextMenuOnMouseEnter(self) end)
	button:SetHandler( "OnMouseExit", function (self) contextMenuOnMouseExit(self) end)
	button:SetText(fco_loc["button_context_menu_undo"])
	button:SetHorizontalAlignment(TEXT_ALIGN_LEFT)
	button:SetAnchor(TOP,menuWindow,TOP,0,361)
    button:SetMouseEnabled(false)
	labelControl = button:GetLabelControl()
    if labelControl ~= nil then
    	labelControl:SetColor(ZO_DISABLED_TEXT:UnpackRGBA())
    end
    button = nil
    labelControl = nil

	--Hide the menu now
 	menuWindow:SetHidden(true)

	return menuWindow
end

--Function that display the context menu after the player click on the button on the inventory
local function showContextMenu(parentWindow)
	if preventerVars.gContextCreated[locVars.gFilterWhere] == false then
		contextMenu[locVars.gFilterWhere] = createContextMenu(parentWindow, locVars.gFilterWhere)
	end
	if contextMenu[locVars.gFilterWhere] ~= nil then
		preventerVars.gContextCreated[locVars.gFilterWhere] = true
		if contextMenu[locVars.gFilterWhere]:IsHidden() then
	        contextMenu[locVars.gFilterWhere]:ClearAnchors()
			contextMenu[locVars.gFilterWhere]:SetAnchor(TOP,parentWindow,BOTTOM, -145, 4)

            local labelControl
            --Update the UNDO button: If nothing is there to undo it should be black text color and no click handler active (mouse disabled)
			local filterPanelId = locVars.gFilterWhere
		    if filterPanelId < 10 then filterPanelId = "0" .. tostring(filterPanelId) end
			local undoButton = WINDOW_MANAGER:GetControlByName(addonVars.gAddonName .. "ButtonContextMenuUNDO" .. "_" .. tostring(filterPanelId), "")
            if undoButton ~= nil then
                --Is anything there to undo?
            	if undoMarkedItems[locVars.gFilterWhere] then
					undoButton:SetMouseEnabled(true)
					labelControl = undoButton:GetLabelControl()
				    if labelControl ~= nil then
				    	labelControl:SetColor(ZO_NORMAL_TEXT:UnpackRGBA())
				    end
                else
					undoButton:SetMouseEnabled(false)
					labelControl = undoButton:GetLabelControl()
				    if labelControl ~= nil then
				    	labelControl:SetColor(ZO_DISABLED_TEXT:UnpackRGBA())
				    end
                end
            end
   			contextMenu[locVars.gFilterWhere]:SetHidden(false)
		else
			contextMenu[locVars.gFilterWhere]:SetHidden(true)
	    end
    end
end

--Add a button to an existing parent control
local function AddButton(parent, name, callbackFunction, text, font, tooltipText, tooltipAlign, textureNormal, textureMouseOver, textureClicked, width, height, left, top, alignMain, alignBackup, alignControl, hideButton)
	--Abort needed?
    if  (parent == nil or name == nil or callbackFunction == nil
     or width <= 0 or height <= 0 or alignMain == nil or alignBackup == nil)
     and (textureNormal == nil or text == nil) then
    	return nil
    end

    local button
    --Does the button already exist?
    button = WINDOW_MANAGER:GetControlByName(name, "")
    if button == nil then
    	--Button does not exist yet and it should be hidden? Abort here!
        if hideButton == true then return nil end
		--Create the button control at the parent
		button = WINDOW_MANAGER:CreateControl(name, parent, CT_BUTTON)
    end
	--Button was created?
    if button ~= nil then
		--Button should be hidden?
		if hideButton == false then
	        --Set the button's size
		    button:SetDimensions(width, height)

			--Align the button
	        if alignControl == nil then
	        	alignControl = parent
	        end

			--SetAnchor(point, relativeTo, relativePoint, offsetX, offsetY)
			button:SetAnchor(alignMain, alignControl, alignBackup, left, top)

	        --Texture or text?
			if (text ~= nil) then
				--Text
				--Set the button's font
			    if font == nil then
				    button:SetFont("ZoFontGameSmall")
			    else
				    button:SetFont(font)
			    end

	            --Set the button's text
	            button:SetText(text)

	        else
	        	--Texture
				local texture

                --Check if texture exists
                texture = WINDOW_MANAGER:GetControlByName(name .. "Texture", "")
				if texture == nil then
				    --Create the texture for the button to hold the image
				    texture = WINDOW_MANAGER:CreateControl(name .. "Texture", button, CT_TEXTURE)
                end
			    texture:SetAnchorFill()

		        --Set the texture for normale state now
		        texture:SetTexture(textureNormal)

				--Do we have seperate textures for the button states?
	 			button.upTexture 	  = textureNormal
				button.downTexture 	  = textureMouseOver or textureNormal
				button.clickedTexture = textureClicked or textureNormal
			end

            if tooltipAlign == nil then tooltipAlign = TOP end

			--Set a tooltip?
			if tooltipText ~= nil then
            	tooltipText = locVars.preChatTextGreen .. tooltipText
				button:SetHandler("OnMouseEnter", function(self)
	                self:GetChild(1):SetTexture(self.downTexture)
					ZO_Tooltips_ShowTextTooltip(button, tooltipAlign, tooltipText)
				end)
				button:SetHandler("OnMouseExit", function(self)
	                self:GetChild(1):SetTexture(self.upTexture)
					ZO_Tooltips_HideTextTooltip()
				end)
	        else
				button:SetHandler("OnMouseEnter", function(self)
	                self:GetChild(1):SetTexture(self.downTexture)
				end)
				button:SetHandler("OnMouseExit", function(self)
	                self:GetChild(1):SetTexture(self.upTexture)
				end)
	        end
			--Set the callback function of the button
		    button:SetHandler("OnClicked", function(...)
	        	callbackFunction(...)
	        end)
			button:SetHandler("OnMouseDown", function(butn, ctrl, alt, shift, command)
				butn:GetChild(1):SetTexture(butn.clickedTexture)
			end)

			--Show the button and make it react on mouse input
		    button:SetHidden(false)
		   	button:SetMouseEnabled(true)

	        --Return the button control
			return button
    	else
			--Hide the button and make it not reacting on mouse input
		    button:SetHidden(true)
		   	button:SetMouseEnabled(false)
        end
	else
    	return nil
    end
end

--Add additonal buttons, controlled by the settings
local function AddAdditionalButtons(buttonName)
	--Add all additional buttons
	if (buttonName == -1) then
		AddAdditionalButtons("FCOSettings")
		AddAdditionalButtons("FCOInventoriesContextMenuButtons")
    else
	--Add only a specific button
    	if buttonName == "FCOSettings" then
			--Add or hide a button to the main menu category bar, right to the help button, if enabled in settings
			AddButton(ZO_MainMenuCategoryBar, "ZO_MainMenuCategoryBarButtonFCOSettings", ShowFCOItemSaverSettings, nil, nil, fco_loc["button_FCOIS_settings_tooltip"], RIGHT, "/esoui/art/charactercreate/rotate_right_up.dds", "/esoui/art/charactercreate/rotate_right_over.dds", "/esoui/art/charactercreate/rotate_right_down.dds", 32, 32, ZO_MainMenuCategoryBarButton12:GetWidth() + 30, 35, BOTTOM, TOP, ZO_MainMenuCategoryBarButton12, not settings.showFCOISMenuBarButton)
		elseif buttonName == "FCOInventoriesContextMenuButtons" then
			AddAdditionalButtons("FCOPlayerInventoryContextMenu")
			AddAdditionalButtons("FCOPlayerBankWithdrawContextMenu")
			AddAdditionalButtons("FCOGuildBankWithdrawContextMenu")
			AddAdditionalButtons("FCODeconstructionContextMenu")
			AddAdditionalButtons("FCOImprovementContextMenu")
			AddAdditionalButtons("FCOEnchantingContextMenu")
    	elseif buttonName == "FCOPlayerInventoryContextMenu" then
			--Add or hide a button to the player bank withdraw sort bar
			AddButton(ZO_PlayerInventorySortBy, "ZO_PlayerInventoryButtonAdditionalOptions", showContextMenu, nil, nil, fco_loc["button_context_menu_tooltip"], RIGHT, "/esoui/art/ava/tabicon_bg_score_inactive.dds", "/esoui/art/ava/tabicon_bg_score_disabled.dds", "/esoui/art/ava/tabicon_bg_score_disabled.dds", 32, 32, -16, 32, BOTTOM, TOP, ZO_PlayerInventorySortByNew, not settings.showFCOISAdditionalInventoriesButton)
    	elseif buttonName == "FCOPlayerBankWithdrawContextMenu" then
			--Add or hide a button to the player bank withdraw sort bar
			AddButton(ZO_PlayerBankSortBy, "ZO_PlayerBankFCOWithdrawButtonAdditionalOptions", showContextMenu, nil, nil, fco_loc["button_context_menu_tooltip"], RIGHT, "/esoui/art/ava/tabicon_bg_score_inactive.dds", "/esoui/art/ava/tabicon_bg_score_disabled.dds", "/esoui/art/ava/tabicon_bg_score_disabled.dds", 32, 32, -213, 26, BOTTOM, TOP, ZO_PlayerBankSortByName, not settings.showFCOISAdditionalInventoriesButton)
    	elseif buttonName == "FCOGuildBankWithdrawContextMenu" then
			--Add or hide a button to the player bank withdraw sort bar
			AddButton(ZO_GuildBankSortBy, "ZO_GuildBankFCOWithdrawButtonAdditionalOptions", showContextMenu, nil, nil, fco_loc["button_context_menu_tooltip"], RIGHT, "/esoui/art/ava/tabicon_bg_score_inactive.dds", "/esoui/art/ava/tabicon_bg_score_disabled.dds", "/esoui/art/ava/tabicon_bg_score_disabled.dds", 32, 32, -213, 26, BOTTOM, TOP, ZO_GuildBankSortByName, not settings.showFCOISAdditionalInventoriesButton)
    	elseif buttonName == "FCODeconstructionContextMenu" then
			--Add or hide a button to the crafting station deconstruction sort bar
			AddButton(ZO_SmithingTopLevelDeconstructionPanelInventorySortBy, "ZO_SmithingTopLevelDeconstructionPanelInventoryButtonAdditionalOptions", showContextMenu, nil, nil, fco_loc["button_context_menu_tooltip"], RIGHT, "/esoui/art/ava/tabicon_bg_score_inactive.dds", "/esoui/art/ava/tabicon_bg_score_disabled.dds", "/esoui/art/ava/tabicon_bg_score_disabled.dds", 32, 32, -208, 26, BOTTOM, TOP, ZO_SmithingTopLevelDeconstructionPanelInventorySortByName, not settings.showFCOISAdditionalInventoriesButton)
    	elseif buttonName == "FCOImprovementContextMenu" then
			--Add or hide a button to the crafting station improvement sort bar
			AddButton(ZO_SmithingTopLevelImprovementPanelInventorySortBy, "ZO_SmithingTopLevelImprovementPanelInventoryButtonAdditionalOptions", showContextMenu, nil, nil, fco_loc["button_context_menu_tooltip"], RIGHT, "/esoui/art/ava/tabicon_bg_score_inactive.dds", "/esoui/art/ava/tabicon_bg_score_disabled.dds", "/esoui/art/ava/tabicon_bg_score_disabled.dds", 32, 32, -208, 26, BOTTOM, TOP, ZO_SmithingTopLevelImprovementPanelInventorySortByName, not settings.showFCOISAdditionalInventoriesButton)
		end
    end
end

-- Build the options menu
local function BuildAddonMenu()
	local panelData = {
		type = 'panel',
		name = addonVars.addonNameMenu,
		author = 'Baertram',
		version = addonVars.addonVersionOptions,
		registerForRefresh = true,
		registerForDefaults = true,
	}

	-- Build options menu parts
	local texturesList = {
		fco_loc["options_lock"],
		fco_loc["options_gear"],
		fco_loc["options_research"],
		fco_loc["options_coins"],
        fco_loc["options_star"],
        fco_loc["options_flag"],
        fco_loc["options_boxstar"],
        fco_loc["options_medic"],
        fco_loc["options_timer"],
        fco_loc["options_flask"],
		fco_loc["options_cancel_up"],
	    fco_loc["options_info_up"],
	    fco_loc["options_pinned_normal"],
	    fco_loc["options_cadwell_indexicon_gold_up"],
	    fco_loc["options_cadwell_indexicon_silver_up"],
	    fco_loc["options_campaignbonus_keepicon"],
	    fco_loc["options_campaignbonus_scrollicon"],
	    fco_loc["options_campaignbrowser_columnheader_ad"],
	    fco_loc["options_campaignbrowser_columnheader_dc"],
	    fco_loc["options_campaignbrowser_columnheader_ep"],
	    fco_loc["options_campaignbrowser_guild"],
	    fco_loc["options_campaignbrowser_indexicon_normal_up"],
	    fco_loc["options_overview_indexicon_scoring_up"],
	    fco_loc["options_charactercreate_bodyicon_up"],
	    fco_loc["options_gearslot_offhand"],
	    fco_loc["options_gearslot_mainhand"],
	    fco_loc["options_gearslot_costume"],
	    fco_loc["options_chat_mail_up"],
	    fco_loc["options_chat_notification_up"],
	    fco_loc["options_alchemy_tabicon_reagent_up"],
	    fco_loc["options_smithing_tabicon_refine_up"],
	    fco_loc["options_deathrecap_killingblow_icon"],
		fco_loc["options_bait_emptyslot"],
		fco_loc["options_guildhistory_indexicon_guildbank_up"],
		fco_loc["options_guild_indexicon_member_up"],
		fco_loc["options_tabicon_roster_up"],
		fco_loc["options_poi_dungeon_complete"],
		fco_loc["options_poi_groupinstance_complete"],
		fco_loc["options_servicepin_magesguild"],
		fco_loc["options_servicepin_fightersguild"],
		fco_loc["options_lfg_dps_up"],
		fco_loc["options_lfg_leader_icon"],
		fco_loc["options_lfg_tank_up"],
		fco_loc["options_lfg_veterandungeon_up"],
		fco_loc["options_lfg_normaldungeon_up"],
		fco_loc["options_icon_dualwield"],
		fco_loc["options_icon_firestaff"],
		fco_loc["options_icon_bows"],
		fco_loc["options_icon_2handed"],
		fco_loc["options_icon_1handed"],
		fco_loc["options_progression_tabicon_backup_inactive"],
		fco_loc["options_inventory_tabicon_repair_disabled"],
		fco_loc["options_selectedquesthighlight"],
		fco_loc["options_writ"],
	}

	local languageOptions = {
	    [1] = fco_loc["options_language_dropdown_selection1"],
	    [2] = fco_loc["options_language_dropdown_selection2"],
	    [3] = fco_loc["options_language_dropdown_selection3"]
	}

    local savedVariablesOptions = {
    	[1] = fco_loc["options_savedVariables_dropdown_selection1"],
        [2] = fco_loc["options_savedVariables_dropdown_selection2"]
    }

	FCOSettingsPanel = LAM:RegisterAddonPanel(addonVars.gAddonName, panelData)

	--Filter icon selection
	local FilterIcons, filter1Preview, filter2Preview, filter3Preview, filter4Preview, filter5Preview
	FilterIcons = function(panel)
		if panel == FCOSettingsPanel then
			filter1Preview = WINDOW_MANAGER:CreateControl(nil, panel.controlsToRefresh[8], CT_TEXTURE)
			filter1Preview:SetAnchor(RIGHT, panel.controlsToRefresh[7].dropdown:GetControl(), LEFT, -10, 0)
			filter1Preview:SetTexture(MARKER_TEXTURES[settings.icon[1].texture])
			filter1Preview:SetDimensions(settings.icon[1].size, settings.icon[1].size)
            filter1Preview:SetColor(settings.icon[1].color.r, settings.icon[1].color.g, settings.icon[1].color.b, settings.icon[1].color.a)

			filter2Preview = WINDOW_MANAGER:CreateControl(nil, panel.controlsToRefresh[12], CT_TEXTURE)
			filter2Preview:SetAnchor(RIGHT, panel.controlsToRefresh[11].dropdown:GetControl(), RIGHT, 45, 0)
			filter2Preview:SetTexture(MARKER_TEXTURES[settings.icon[2].texture])
			filter2Preview:SetDimensions(settings.icon[2].size, settings.icon[2].size)
            filter2Preview:SetColor(settings.icon[2].color.r, settings.icon[2].color.g, settings.icon[2].color.b, settings.icon[2].color.a)

			filter4Preview = WINDOW_MANAGER:CreateControl(nil, panel.controlsToRefresh[17], CT_TEXTURE)
			filter4Preview:SetAnchor(RIGHT, panel.controlsToRefresh[16].dropdown:GetControl(), RIGHT, 45, 0)
			filter4Preview:SetTexture(MARKER_TEXTURES[settings.icon[4].texture])
			filter4Preview:SetDimensions(settings.icon[4].size, settings.icon[4].size)
            filter4Preview:SetColor(settings.icon[4].color.r, settings.icon[4].color.g, settings.icon[4].color.b, settings.icon[4].color.a)

			filter6Preview = WINDOW_MANAGER:CreateControl(nil, panel.controlsToRefresh[22], CT_TEXTURE)
			filter6Preview:SetAnchor(RIGHT, panel.controlsToRefresh[21].dropdown:GetControl(), RIGHT, 45, 0)
			filter6Preview:SetTexture(MARKER_TEXTURES[settings.icon[6].texture])
			filter6Preview:SetDimensions(settings.icon[6].size, settings.icon[6].size)
            filter6Preview:SetColor(settings.icon[6].color.r, settings.icon[6].color.g, settings.icon[6].color.b, settings.icon[6].color.a)

			filter7Preview = WINDOW_MANAGER:CreateControl(nil, panel.controlsToRefresh[27], CT_TEXTURE)
			filter7Preview:SetAnchor(RIGHT, panel.controlsToRefresh[26].dropdown:GetControl(), RIGHT, 45, 0)
			filter7Preview:SetTexture(MARKER_TEXTURES[settings.icon[7].texture])
			filter7Preview:SetDimensions(settings.icon[7].size, settings.icon[7].size)
            filter7Preview:SetColor(settings.icon[7].color.r, settings.icon[7].color.g, settings.icon[7].color.b, settings.icon[7].color.a)

			filter8Preview = WINDOW_MANAGER:CreateControl(nil, panel.controlsToRefresh[32], CT_TEXTURE)
			filter8Preview:SetAnchor(RIGHT, panel.controlsToRefresh[31].dropdown:GetControl(), RIGHT, 45, 0)
			filter8Preview:SetTexture(MARKER_TEXTURES[settings.icon[8].texture])
			filter8Preview:SetDimensions(settings.icon[8].size, settings.icon[8].size)
            filter8Preview:SetColor(settings.icon[8].color.r, settings.icon[8].color.g, settings.icon[8].color.b, settings.icon[8].color.a)

			filter3Preview = WINDOW_MANAGER:CreateControl(nil, panel.controlsToRefresh[36], CT_TEXTURE)
			filter3Preview:SetAnchor(RIGHT, panel.controlsToRefresh[35].dropdown:GetControl(), LEFT, -10, 0)
			filter3Preview:SetTexture(MARKER_TEXTURES[settings.icon[3].texture])
			filter3Preview:SetDimensions(settings.icon[3].size, settings.icon[3].size)
            filter3Preview:SetColor(settings.icon[3].color.r, settings.icon[3].color.g, settings.icon[3].color.b, settings.icon[3].color.a)

			filter5Preview = WINDOW_MANAGER:CreateControl(nil, panel.controlsToRefresh[39], CT_TEXTURE)
			filter5Preview:SetAnchor(RIGHT, panel.controlsToRefresh[38].dropdown:GetControl(), LEFT, -10, 0)
			filter5Preview:SetTexture(MARKER_TEXTURES[settings.icon[5].texture])
			filter5Preview:SetDimensions(settings.icon[5].size, settings.icon[5].size)
            filter5Preview:SetColor(settings.icon[5].color.r, settings.icon[5].color.g, settings.icon[5].color.b, settings.icon[5].color.a)

			CALLBACK_MANAGER:UnregisterCallback("LAM-RefreshPanel", FilterIcons)
		end
	end
	CALLBACK_MANAGER:RegisterCallback("LAM-PanelControlsCreated", FilterIcons)

	local optionsTable =
    {	-- BEGIN OF OPTIONS TABLE

		[1] =
		{
			type = 'description',
			text = fco_loc["options_description"],
		},
--==============================================================================
        [2] =
		{
        	type = 'header',
        	name = fco_loc["options_header1"],
        },
		[3] =
		{
			type = 'dropdown',
			name = fco_loc["options_language"],
			tooltip = fco_loc["options_language_tooltip"],
			choices = languageOptions,
            getFunc = function() return languageOptions[defaultSettings.language] end,
            setFunc = function(value)
                for i,v in pairs(languageOptions) do
                    if v == value then
                        debugMessage("[Settings language] v: " .. tostring(v) .. ", i: " .. tostring(i), false)
                    	defaultSettings.language = i
                        --Tell the settings that you have manually chosen the language and want to keep it
                        --Read in function Localization() after ReloadUI()
                        settings.languageChoosen = true
						--fco_loc			  	 = fco_loc[i]
                        ReloadUI()
                    end
                end
            end,
            warning = fco_loc["options_language_description1"],
		},
		[4] =
		{
			type = 'dropdown',
			name = fco_loc["options_savedvariables"],
			tooltip = fco_loc["options_savedvariables_tooltip"],
			choices = savedVariablesOptions,
            getFunc = function() return savedVariablesOptions[defaultSettings.saveMode] end,
            setFunc = function(value)
                for i,v in pairs(savedVariablesOptions) do
                    if v == value then
                        debugMessage("[Settings save mode] v: " .. tostring(v) .. ", i: " .. tostring(i), false)
                        defaultSettings.saveMode = i
                        ReloadUI()
                    end
                end
            end,
            warning = fco_loc["options_language_description1"],
		},
--==============================================================================
		[5] = {
			type = "header",
			name = fco_loc["options_header_color"]
		},

		[6] = {
			type = "colorpicker",
			name = fco_loc["options_icon1_color"],
			tooltip = fco_loc["options_icon1_color_tooltip"],
			getFunc = function() return settings.icon[1].color.r, settings.icon[1].color.g, settings.icon[1].color.b, settings.icon[1].color.a end,
            setFunc = function(r,g,b,a)
            	settings.icon[1].color = {["r"] = r, ["g"] = g, ["b"] = b, ["a"] = a}
				--CreateTextures(-1) --Update the icon colors
                filter1Preview:SetColor(r, g, b, a)
				--Set global variable to update the marker colors and textures
                preventerVars.gUpdateMarkersNow = true
			end,
            width="half",
            default = settings.icon[1].color,
		},
		[7] = {
			type = "dropdown",
			name = fco_loc["options_icon1_texture"],
			tooltip = fco_loc["options_icon1_texture_tooltip"],
			choices = texturesList,
			getFunc = function()
            	return texturesList[settings.icon[1].texture]
            end,
			setFunc = function(selected)
					for index, name in ipairs(texturesList) do
						if name == selected then
							settings.icon[1].texture = index
							filter1Preview:SetTexture(MARKER_TEXTURES[index])
							--CreateTextures(-1) --Update the icon colors
                            local p_button = WINDOW_MANAGER:GetControlByName("ZO_PlayerInventory_FilterButton1", "")
                            UpdateButtonColors(1, p_button, -999)
							--Set global variable to update the marker colors and textures
			                preventerVars.gUpdateMarkersNow = true
							break
						end
					end
				end,
            width="half",
			default = texturesList[settings.icon[1].texture],
		},
		[8] = {
			type = "slider",
			name = fco_loc["options_icon1_size"],
			tooltip = fco_loc["options_icon1_size_tooltip"],
			min = 12,
			max = 64,
			getFunc = function() return settings.icon[1].size end,
			setFunc = function(size)
					settings.icon[1].size = size
					filter1Preview:SetDimensions(size, size)
					--CreateTextures(-1) --Update the icon colors
 				end,
            width="full",
			default = settings.icon[1].size,
		},
--==============================================================================
   		[9] = {
			type = "editbox",
			name = fco_loc["options_icon2_name"],
			tooltip = fco_loc["options_icon2_name_tooltip"],
			getFunc = function() return settings.icon[2].name end,
			setFunc = function(newValue)
            	settings.icon[2].name = newValue
				changeEquipmentGearMenuTexts(1)
            end,
			width = "half",
			default = fco_loc["options_icon2_name"],
            disabled = function() return not settings.isGearEnabled[1] end,
		},
		[10] = {
			type = "colorpicker",
			name = fco_loc["options_icon2_color"],
			tooltip = fco_loc["options_icon2_color_tooltip"],
			getFunc = function() return settings.icon[2].color.r, settings.icon[2].color.g, settings.icon[2].color.b, settings.icon[2].color.a end,
			setFunc = function(r,g,b,a)
				settings.icon[2].color = {["r"] = r, ["g"] = g, ["b"] = b, ["a"] = a}
				--CreateTextures(-1) --Update the icon colors
                filter2Preview:SetColor(r, g, b, a)
				--Set global variable to update the marker colors and textures
                preventerVars.gUpdateMarkersNow = true
			end,
            width = "half",
			default = settings.icon[2].color,
            disabled = function() return not settings.isGearEnabled[1] end,
		},
		[11] = {
			type = "dropdown",
			name = fco_loc["options_icon2_texture"],
			tooltip = fco_loc["options_icon2_texture_tooltip"],
			choices = texturesList,
			getFunc = function()
            	return texturesList[settings.icon[2].texture]
            end,
			setFunc = function(selected)
					for index, name in ipairs(texturesList) do
						if name == selected then
							settings.icon[2].texture = index
							filter2Preview:SetTexture(MARKER_TEXTURES[index])
							--CreateTextures(-1) --Update the icon colors
                            local p_button = WINDOW_MANAGER:GetControlByName("ZO_PlayerInventory_FilterButton2", "")
                            UpdateButtonColors(2, p_button, -999)
							--Set global variable to update the marker colors and textures
			                preventerVars.gUpdateMarkersNow = true
							break
						end
					end
				end,
        	width = "half",
			default = texturesList[settings.icon[2].texture],
            disabled = function() return not settings.isGearEnabled[1] end,
		},
		[12] = {
			type = "slider",
			name = fco_loc["options_icon2_size"],
			tooltip = fco_loc["options_icon2_size_tooltip"],
			min = 12,
			max = 64,
			getFunc = function() return settings.icon[2].size end,
			setFunc = function(size)
					settings.icon[2].size = size
					filter2Preview:SetDimensions(size, size)
					--CreateTextures(-1) --Update the icon colors
					--Set global variable to update the marker colors and textures
	                preventerVars.gUpdateMarkersNow = true
				end,
            width = "half",
			default = settings.icon[2].size,
            disabled = function() return not settings.isGearEnabled[1] end,
		},

		[13] = {
			type = "checkbox",
			name = fco_loc["options_icon2_enabled"],
			tooltip = fco_loc["options_icon2_enabled_tooltip"],
			getFunc = function() return settings.isGearEnabled[1] end,
			setFunc = function(value) settings.isGearEnabled[1] = not settings.isGearEnabled[1]
				--Hide the textures for gear 1
            	if not settings.isGearEnabled[1] then
                    --Character equipment
                	RefreshEquipmentControl(nil, false, 2)
                else
                    --Character equipment, create if not yet created
                	RefreshEquipmentControl(nil, true, 2)
                end
                preventerVars.gChangedGears 	= true
                preventerVars.gUpdateMarkersNow = true
            end,
		},

--==============================================================================
   		[14] = {
			type = "editbox",
			name = fco_loc["options_icon4_name"],
			tooltip = fco_loc["options_icon4_name_tooltip"],
			getFunc = function() return settings.icon[4].name end,
			setFunc = function(newValue)
            	settings.icon[4].name = newValue
				changeEquipmentGearMenuTexts(2)
			end,
			width = "half",
			default = fco_loc["options_icon4_name"],
            disabled = function() return not settings.isGearEnabled[2] end,
		},
		[15] = {
			type = "colorpicker",
			name = fco_loc["options_icon4_color"],
			tooltip = fco_loc["options_icon4_color_tooltip"],
			getFunc = function() return settings.icon[4].color.r, settings.icon[4].color.g, settings.icon[4].color.b, settings.icon[4].color.a end,
			setFunc = function(r,g,b,a)
				settings.icon[4].color = {["r"] = r, ["g"] = g, ["b"] = b, ["a"] = a}
				--CreateTextures(-1) --Update the icon colors
                filter4Preview:SetColor(r, g, b, a)
				--Set global variable to update the marker colors and textures
                preventerVars.gUpdateMarkersNow = true
			end,
            width = "half",
			default = settings.icon[4].color,
            disabled = function() return not settings.isGearEnabled[2] end,
		},
		[16] = {
			type = "dropdown",
			name = fco_loc["options_icon4_texture"],
			tooltip = fco_loc["options_icon4_texture_tooltip"],
			choices = texturesList,
			getFunc = function()
            	return texturesList[settings.icon[4].texture]
            end,
			setFunc = function(selected)
					for index, name in ipairs(texturesList) do
						if name == selected then
							settings.icon[4].texture = index
							filter4Preview:SetTexture(MARKER_TEXTURES[index])
							--CreateTextures(-1) --Update the icon colors
							--Set global variable to update the marker colors and textures
			                preventerVars.gUpdateMarkersNow = true
							break
						end
					end
				end,
            width = "half",
			default = texturesList[settings.icon[4].texture],
            disabled = function() return not settings.isGearEnabled[2] end,
		},
		[17] = {
			type = "slider",
			name = fco_loc["options_icon4_size"],
			tooltip = fco_loc["options_icon4_size_tooltip"],
			min = 12,
			max = 64,
			getFunc = function() return settings.icon[4].size end,
			setFunc = function(size)
					settings.icon[4].size = size
					filter4Preview:SetDimensions(size, size)
					--CreateTextures(-1) --Update the icon colors
					--Set global variable to update the marker colors and textures
	                preventerVars.gUpdateMarkersNow = true
				end,
            width = "half",
			default = settings.icon[4].size,
            disabled = function() return not settings.isGearEnabled[2] end,
		},

		[18] = {
			type = "checkbox",
			name = fco_loc["options_icon4_enabled"],
			tooltip = fco_loc["options_icon4_enabled_tooltip"],
			getFunc = function() return settings.isGearEnabled[2] end,
			setFunc = function(value) settings.isGearEnabled[2] = not settings.isGearEnabled[2]
				--Hide the textures for gear 2
            	if not settings.isGearEnabled[2] then
                    --Character equipment
                	RefreshEquipmentControl(nil, false, 4)
                else
                    --Character equipment, create if not yet created
                	RefreshEquipmentControl(nil, true, 4)
                end
                preventerVars.gChangedGears 	= true
                preventerVars.gUpdateMarkersNow = true
            end,
		},

--==============================================================================
   		[19] = {
			type = "editbox",
			name = fco_loc["options_icon6_name"],
			tooltip = fco_loc["options_icon6_name_tooltip"],
			getFunc = function() return settings.icon[6].name end,
			setFunc = function(newValue)
            	settings.icon[6].name = newValue
				changeEquipmentGearMenuTexts(3)
			end,
			width = "half",
			default = fco_loc["options_icon6_name"],
            disabled = function() return not settings.isGearEnabled[3] end,
		},
		[20] = {
			type = "colorpicker",
			name = fco_loc["options_icon6_color"],
			tooltip = fco_loc["options_icon6_color_tooltip"],
			getFunc = function() return settings.icon[6].color.r, settings.icon[6].color.g, settings.icon[6].color.b, settings.icon[6].color.a end,
			setFunc = function(r,g,b,a)
				settings.icon[6].color = {["r"] = r, ["g"] = g, ["b"] = b, ["a"] = a}
				--CreateTextures(-1) --Update the icon colors
                filter6Preview:SetColor(r, g, b, a)
				--Set global variable to update the marker colors and textures
                preventerVars.gUpdateMarkersNow = true
			end,
            width = "half",
			default = settings.icon[6].color,
            disabled = function() return not settings.isGearEnabled[3] end,
		},
		[21] = {
			type = "dropdown",
			name = fco_loc["options_icon6_texture"],
			tooltip = fco_loc["options_icon6_texture_tooltip"],
			choices = texturesList,
			getFunc = function()
            	return texturesList[settings.icon[6].texture]
            end,
			setFunc = function(selected)
					for index, name in ipairs(texturesList) do
						if name == selected then
							settings.icon[6].texture = index
							filter6Preview:SetTexture(MARKER_TEXTURES[index])
							--CreateTextures(-1) --Update the icon colors
							--Set global variable to update the marker colors and textures
			                preventerVars.gUpdateMarkersNow = true
							break
						end
					end
				end,
            width = "half",
			default = texturesList[settings.icon[6].texture],
            disabled = function() return not settings.isGearEnabled[3] end,
		},
		[22] = {
			type = "slider",
			name = fco_loc["options_icon6_size"],
			tooltip = fco_loc["options_icon6_size_tooltip"],
			min = 12,
			max = 64,
			getFunc = function() return settings.icon[6].size end,
			setFunc = function(size)
					settings.icon[6].size = size
					filter6Preview:SetDimensions(size, size)
					--CreateTextures(-1) --Update the icon colors
					--Set global variable to update the marker colors and textures
	                preventerVars.gUpdateMarkersNow = true
				end,
            width = "half",
			default = settings.icon[6].size,
            disabled = function() return not settings.isGearEnabled[3] end,
		},

		[23] = {
			type = "checkbox",
			name = fco_loc["options_icon6_enabled"],
			tooltip = fco_loc["options_icon6_enabled_tooltip"],
			getFunc = function() return settings.isGearEnabled[3] end,
			setFunc = function(value) settings.isGearEnabled[3] = not settings.isGearEnabled[3]
				--Hide the textures for gear 3
            	if not settings.isGearEnabled[3] then
                    --Character equipment
                	RefreshEquipmentControl(nil, false, 6)
                else
                    --Character equipment, create if not yet created
                	RefreshEquipmentControl(nil, true, 6)
                end
                preventerVars.gChangedGears 	= true
                preventerVars.gUpdateMarkersNow = true
            end,
		},

--==============================================================================
   		[24] = {
			type = "editbox",
			name = fco_loc["options_icon7_name"],
			tooltip = fco_loc["options_icon7_name_tooltip"],
			getFunc = function() return settings.icon[7].name end,
			setFunc = function(newValue)
            	settings.icon[7].name = newValue
				changeEquipmentGearMenuTexts(4)
			end,
			width = "half",
			default = fco_loc["options_icon7_name"],
            disabled = function() return not settings.isGearEnabled[4] end,
		},
		[25] = {
			type = "colorpicker",
			name = fco_loc["options_icon7_color"],
			tooltip = fco_loc["options_icon7_color_tooltip"],
			getFunc = function() return settings.icon[7].color.r, settings.icon[7].color.g, settings.icon[7].color.b, settings.icon[7].color.a end,
			setFunc = function(r,g,b,a)
				settings.icon[7].color = {["r"] = r, ["g"] = g, ["b"] = b, ["a"] = a}
				--CreateTextures(-1) --Update the icon colors
                filter7Preview:SetColor(r, g, b, a)
				--Set global variable to update the marker colors and textures
                preventerVars.gUpdateMarkersNow = true
			end,
            width = "half",
			default = settings.icon[7].color,
            disabled = function() return not settings.isGearEnabled[4] end,
		},
		[26] = {
			type = "dropdown",
			name = fco_loc["options_icon7_texture"],
			tooltip = fco_loc["options_icon7_texture_tooltip"],
			choices = texturesList,
			getFunc = function()
            	return texturesList[settings.icon[7].texture]
            end,
			setFunc = function(selected)
					for index, name in ipairs(texturesList) do
						if name == selected then
							settings.icon[7].texture = index
							filter7Preview:SetTexture(MARKER_TEXTURES[index])
							--CreateTextures(-1) --Update the icon colors
							--Set global variable to update the marker colors and textures
			                preventerVars.gUpdateMarkersNow = true
							break
						end
					end
				end,
            width = "half",
			default = texturesList[settings.icon[7].texture],
            disabled = function() return not settings.isGearEnabled[4] end,
		},
		[27] = {
			type = "slider",
			name = fco_loc["options_icon7_size"],
			tooltip = fco_loc["options_icon7_size_tooltip"],
			min = 12,
			max = 64,
			getFunc = function() return settings.icon[7].size end,
			setFunc = function(size)
					settings.icon[7].size = size
					filter7Preview:SetDimensions(size, size)
					--CreateTextures(-1) --Update the icon colors
					--Set global variable to update the marker colors and textures
	                preventerVars.gUpdateMarkersNow = true
				end,
            width = "half",
			default = settings.icon[7].size,
            disabled = function() return not settings.isGearEnabled[4] end,
		},

		[28] = {
			type = "checkbox",
			name = fco_loc["options_icon7_enabled"],
			tooltip = fco_loc["options_icon7_enabled_tooltip"],
			getFunc = function() return settings.isGearEnabled[4] end,
			setFunc = function(value) settings.isGearEnabled[4] = not settings.isGearEnabled[4]
				--Hide the textures for gear 4
            	if not settings.isGearEnabled[4] then
                    --Character equipment
                	RefreshEquipmentControl(nil, false, 7)
                else
                    --Character equipment, create if not yet created
                	RefreshEquipmentControl(nil, true, 7)
                end
                preventerVars.gChangedGears 	= true
                preventerVars.gUpdateMarkersNow = true
            end,
		},

--==============================================================================
   		[29] = {
			type = "editbox",
			name = fco_loc["options_icon8_name"],
			tooltip = fco_loc["options_icon8_name_tooltip"],
			getFunc = function() return settings.icon[8].name end,
			setFunc = function(newValue)
            	settings.icon[8].name = newValue
				changeEquipmentGearMenuTexts(5)
			end,
			width = "half",
			default = fco_loc["options_icon8_name"],
            disabled = function() return not settings.isGearEnabled[5] end,
		},
		[30] = {
			type = "colorpicker",
			name = fco_loc["options_icon8_color"],
			tooltip = fco_loc["options_icon8_color_tooltip"],
			getFunc = function() return settings.icon[8].color.r, settings.icon[8].color.g, settings.icon[8].color.b, settings.icon[8].color.a end,
			setFunc = function(r,g,b,a)
				settings.icon[8].color = {["r"] = r, ["g"] = g, ["b"] = b, ["a"] = a}
				--CreateTextures(-1) --Update the icon colors
                filter8Preview:SetColor(r, g, b, a)
				--Set global variable to update the marker colors and textures
                preventerVars.gUpdateMarkersNow = true
			end,
            width = "half",
			default = settings.icon[8].color,
            disabled = function() return not settings.isGearEnabled[5] end,
		},
		[31] = {
			type = "dropdown",
			name = fco_loc["options_icon8_texture"],
			tooltip = fco_loc["options_icon8_texture_tooltip"],
			choices = texturesList,
			getFunc = function()
            	return texturesList[settings.icon[8].texture]
            end,
			setFunc = function(selected)
					for index, name in ipairs(texturesList) do
						if name == selected then
							settings.icon[8].texture = index
							filter8Preview:SetTexture(MARKER_TEXTURES[index])
							--CreateTextures(-1) --Update the icon colors
							--Set global variable to update the marker colors and textures
			                preventerVars.gUpdateMarkersNow = true
							break
						end
					end
				end,
            width = "half",
			default = texturesList[settings.icon[8].texture],
            disabled = function() return not settings.isGearEnabled[5] end,
		},
		[32] = {
			type = "slider",
			name = fco_loc["options_icon8_size"],
			tooltip = fco_loc["options_icon8_size_tooltip"],
			min = 12,
			max = 64,
			getFunc = function() return settings.icon[8].size end,
			setFunc = function(size)
					settings.icon[8].size = size
					filter8Preview:SetDimensions(size, size)
					--CreateTextures(-1) --Update the icon colors
					--Set global variable to update the marker colors and textures
	                preventerVars.gUpdateMarkersNow = true
				end,
            width = "half",
			default = settings.icon[8].size,
            disabled = function() return not settings.isGearEnabled[5] end,
		},

		[33] = {
			type = "checkbox",
			name = fco_loc["options_icon8_enabled"],
			tooltip = fco_loc["options_icon8_enabled_tooltip"],
			getFunc = function() return settings.isGearEnabled[5] end,
			setFunc = function(value) settings.isGearEnabled[5] = not settings.isGearEnabled[5]
				--Hide the textures for gear 5
            	if not settings.isGearEnabled[5] then
                    --Character equipment
                	RefreshEquipmentControl(nil, false, 8)
                else
                    --Character equipment, create if not yet created
                	RefreshEquipmentControl(nil, true, 8)
                end
                preventerVars.gChangedGears 	= true
                preventerVars.gUpdateMarkersNow = true
            end,
		},

--==============================================================================
		[34] = {
			type = "colorpicker",
			name = fco_loc["options_icon3_color"],
			tooltip = fco_loc["options_icon3_color_tooltip"],
			getFunc = function() return settings.icon[3].color.r, settings.icon[3].color.g, settings.icon[3].color.b, settings.icon[3].color.a end,
			setFunc = function(r,g,b,a)
				settings.icon[3].color = {["r"] = r, ["g"] = g, ["b"] = b, ["a"] = a}
				--CreateTextures(-1) --Update the icon colors
                filter3Preview:SetColor(r, g, b, a)
				--Set global variable to update the marker colors and textures
                preventerVars.gUpdateMarkersNow = true
			end,
            width = "half",
			default = settings.icon[3].color,
		},
		[35] = {
			type = "dropdown",
			name = fco_loc["options_icon3_texture"],
			tooltip = fco_loc["options_icon3_texture_tooltip"],
			choices = texturesList,
			getFunc = function()
            	return texturesList[settings.icon[3].texture]
            end,
			setFunc = function(selected)
					for index, name in ipairs(texturesList) do
						if name == selected then
							settings.icon[3].texture = index
							filter3Preview:SetTexture(MARKER_TEXTURES[index])
							--CreateTextures(-1) --Update the icon colors
                            local p_button = WINDOW_MANAGER:GetControlByName("ZO_PlayerInventory_FilterButton3", "")
                            UpdateButtonColors(3, p_button, -999)
							--Set global variable to update the marker colors and textures
			                preventerVars.gUpdateMarkersNow = true
							break
						end
					end
				end,
            width = "half",
			default = texturesList[settings.icon[3].texture],
		},
		[36] = {
			type = "slider",
			name = fco_loc["options_icon3_size"],
			tooltip = fco_loc["options_icon3_size_tooltip"],
			min = 12,
			max = 64,
			getFunc = function() return settings.icon[3].size end,
			setFunc = function(size)
					settings.icon[3].size = size
					filter3Preview:SetDimensions(size, size)
					--CreateTextures(-1) --Update the icon colors
					--Set global variable to update the marker colors and textures
	                preventerVars.gUpdateMarkersNow = true
				end,
            width = "full",
			default = settings.icon[3].size,
		},
--==============================================================================
		[37] = {
			type = "colorpicker",
			name = fco_loc["options_icon5_color"],
			tooltip = fco_loc["options_icon5_color_tooltip"],
			getFunc = function() return settings.icon[5].color.r, settings.icon[5].color.g, settings.icon[5].color.b, settings.icon[5].color.a end,
			setFunc = function(r,g,b,a)
				settings.icon[5].color = {["r"] = r, ["g"] = g, ["b"] = b, ["a"] = a}
				--CreateTextures(-1) --Update the icon colors
                filter5Preview:SetColor(r, g, b, a)
				--Set global variable to update the marker colors and textures
                preventerVars.gUpdateMarkersNow = true
			end,
            width = "half",
			default = settings.icon[5].color,
		},
		[38] = {
			type = "dropdown",
			name = fco_loc["options_icon5_texture"],
			tooltip = fco_loc["options_icon5_texture_tooltip"],
			choices = texturesList,
			getFunc = function()
            	return texturesList[settings.icon[5].texture]
            end,
			setFunc = function(selected)
					for index, name in ipairs(texturesList) do
						if name == selected then
							settings.icon[5].texture = index
							filter5Preview:SetTexture(MARKER_TEXTURES[index])
							--CreateTextures(-1) --Update the icon colors
                            local p_button = WINDOW_MANAGER:GetControlByName("ZO_PlayerInventory_FilterButton4", "")
                            UpdateButtonColors(4, p_button, -999)
							--Set global variable to update the marker colors and textures
			                preventerVars.gUpdateMarkersNow = true
							break
						end
					end
				end,
            width = "half",
			default = texturesList[settings.icon[5].texture],
		},
		[39] = {
			type = "slider",
			name = fco_loc["options_icon5_size"],
			tooltip = fco_loc["options_icon5_size_tooltip"],
			min = 12,
			max = 64,
			getFunc = function() return settings.icon[5].size end,
			setFunc = function(size)
					settings.icon[5].size = size
					filter5Preview:SetDimensions(size, size)
					--CreateTextures(-1) --Update the icon colors
					--Set global variable to update the marker colors and textures
	                preventerVars.gUpdateMarkersNow = true
				end,
            width = "full",
			default = settings.icon[5].size,
		},
--==============================================================================
		[40] = {
			type = "header",
			name = fco_loc["options_header_filters"]
		},

		[41] = {
			type = "checkbox",
			name = fco_loc["options_split_filters"],
			tooltip = fco_loc["options_split_filters_tooltip"],
			getFunc = function() return settings.splitFilters end,
			setFunc = function(value)
            	settings.splitFilters = not settings.splitFilters
                --Set the global flag to override the function's unregisterFilters() unregister method
                --because it will check for settings.SplitFilters and is called from function doFilter() at
                --function EnableFilters(-100) below
                locVars.gSplitFilterOverride = true
		        --Hide the filter buttons at the filter panel Id
				updateFilterButtonsInInv(-1)
                --Unregister and reregister the inventory filter INV_BAGS
				EnableFilters(-100)
            end,
		},

		[42] = {
			type = "checkbox",
			name = fco_loc["options_enable_filter_in_inventory"],
			tooltip = fco_loc["options_enable_filter_in_inventory_tooltip"],
			getFunc = function() return settings.allowInventoryFilter end,
			setFunc = function(value) settings.allowInventoryFilter = not settings.allowInventoryFilter
		        --Hide the filter buttons at the filter panel Id
				updateFilterButtonsInInv(-1)
                --Unregister and reregister the inventory filter INV_BAGS
            	EnableFilters(-100)
            end,
		},
		[43] = {
			type = "checkbox",
			name = fco_loc["options_enable_filter_in_bank"],
			tooltip = fco_loc["options_enable_filter_in_bank_tooltip"],
			getFunc = function() return settings.allowBankFilter end,
			setFunc = function(value) settings.allowBankFilter = not settings.allowBankFilter
            end,
		},
		[44] = {
			type = "checkbox",
			name = fco_loc["options_enable_filter_in_guildbank"],
			tooltip = fco_loc["options_enable_filter_in_guildbank_tooltip"],
			getFunc = function() return settings.allowGuildBankFilter end,
			setFunc = function(value) settings.allowGuildBankFilter = not settings.allowGuildBankFilter
            end,
		},
		[45] = {
			type = "checkbox",
			name = fco_loc["options_enable_filter_in_tradinghouse"],
			tooltip = fco_loc["options_enable_filter_in_tradinghouse_tooltip"],
			getFunc = function() return settings.allowTradinghouseFilter end,
			setFunc = function(value) settings.allowTradinghouseFilter = not settings.allowTradinghouseFilter
            end,
		},
		[46] = {
			type = "checkbox",
			name = fco_loc["options_enable_filter_in_trade"],
			tooltip = fco_loc["options_enable_filter_in_trade_tooltip"],
			getFunc = function() return settings.allowTradeFilter end,
			setFunc = function(value) settings.allowTradeFilter = not settings.allowTradeFilter
            end,
		},
		[47] = {
			type = "checkbox",
			name = fco_loc["options_enable_filter_in_mail"],
			tooltip = fco_loc["options_enable_filter_in_mail_tooltip"],
			getFunc = function() return settings.allowMailFilter end,
			setFunc = function(value) settings.allowMailFilter = not settings.allowMailFilter
            end,
		},
		[48] = {
			type = "checkbox",
			name = fco_loc["options_enable_filter_in_store"],
			tooltip = fco_loc["options_enable_filter_in_store_tooltip"],
			getFunc = function() return settings.allowVendorFilter end,
			setFunc = function(value) settings.allowVendorFilter = not settings.allowVendorFilter
            end,
		},
		[49] = {
			type = "checkbox",
			name = fco_loc["options_enable_filter_in_deconstruction"],
			tooltip = fco_loc["options_enable_filter_in_deconstruction_tooltip"],
			getFunc = function() return settings.allowDeconstructionFilter end,
			setFunc = function(value) settings.allowDeconstructionFilter = not settings.allowDeconstructionFilter
            end,
		},
		[50] = {
			type = "checkbox",
			name = fco_loc["options_enable_filter_in_improvement"],
			tooltip = fco_loc["options_enable_filter_in_improvement_tooltip"],
			getFunc = function() return settings.allowImprovementFilter end,
			setFunc = function(value) settings.allowImprovementFilter = not settings.allowImprovementFilter
            end,
		},
		[51] = {
			type = "checkbox",
			name = fco_loc["options_enable_filter_in_enchanting"],
			tooltip = fco_loc["options_enable_filter_in_enchanting_tooltip"],
			getFunc = function() return settings.allowEnchantingFilter end,
			setFunc = function(value) settings.allowEnchantingFilter = not settings.allowEnchantingFilter
            end,
		},
		[52] = {
			type = "checkbox",
			name = fco_loc["options_research_filter"],
			tooltip = fco_loc["options_research_filter_tooltip"],
			getFunc = function() return settings.allowResearch end,
			setFunc = function(value) settings.allowResearch = not settings.allowResearch end,
		},

--==============================================================================
		[53] = {
			type = "header",
			name = fco_loc["options_header_items"]
		},

		[54] = {
			type = "checkbox",
			name = fco_loc["options_enable_auto_mark_ornate_items"],
			tooltip = fco_loc["options_enable_auto_mark_ornate_items_tooltip"],
			getFunc = function() return settings.autoMarkOrnate end,
			setFunc = function(value)
            	settings.autoMarkOrnate = not settings.autoMarkOrnate
                if (settings.autoMarkOrnate == true) then
                	scanInventoryForOrnate()
                end
            end,
            width = "half",
		},
		[55] = {
			type = "checkbox",
			name = fco_loc["options_enable_auto_mark_ornate_items_in_chat"],
			tooltip = fco_loc["options_enable_auto_mark_ornate_items_tooltip_in_chat"],
			getFunc = function() return settings.showOrnateItemsInChat end,
			setFunc = function(value)
            	settings.showOrnateItemsInChat = not settings.showOrnateItemsInChat
            end,
            disabled = function() return not settings.autoMarkOrnate end,
            width = "half",
		},

		[56] = {
			type = "checkbox",
			name = fco_loc["options_enable_auto_mark_research_items"],
			tooltip = fco_loc["options_enable_auto_mark_research_items_tooltip"],
			getFunc = function() return settings.autoMarkResearch end,
			setFunc = function(value)
            	settings.autoMarkResearch = not settings.autoMarkResearch
                if (settings.autoMarkResearch == true and otherAddons.researchAssistantActive == true) then
	            	scanInventoryForResearch()
                end
            end,
            disabled = function() return not otherAddons.researchAssistantActive end,
            warning = fco_loc["options_enable_auto_mark_research_items_hint"],
            width = "half",
		},
		[57] = {
			type = "checkbox",
			name = fco_loc["options_enable_auto_mark_research_items_in_chat"],
			tooltip = fco_loc["options_enable_auto_mark_research_items_tooltip_in_chat"],
			getFunc = function() return settings.showResearchItemsInChat end,
			setFunc = function(value)
            	settings.showResearchItemsInChat = not settings.showResearchItemsInChat
            end,
            disabled = function()
	            	if (otherAddons.researchAssistantActive == false) then
	            		return true
	            	else
	            		return not settings.autoMarkResearch
	                end
            	end,
            warning = fco_loc["options_enable_auto_mark_research_items_hint"],
            width = "half",
		},

--==============================================================================
		[58] = {
			type = "header",
			name = fco_loc["options_header_equipment"]
		},
		[59] = {
			type = "checkbox",
			name = fco_loc["options_equipment_markall_gear"],
			tooltip = fco_loc["options_equipment_markall_gear_tooltip"],
			getFunc = function() return settings.autoMarkAllEquipment end,
			setFunc = function(value) settings.autoMarkAllEquipment = not settings.autoMarkAllEquipment
            	--Update the right-click menu
            end,
		},
		[60] = {
			type = "checkbox",
			name = fco_loc["options_equipment_markall_gear_add_weapons"],
			tooltip = fco_loc["options_equipment_markall_gear_add_weapons_tooltip"],
			getFunc = function() return settings.autoMarkAllWeapon end,
			setFunc = function(value) settings.autoMarkAllWeapon = not settings.autoMarkAllWeapon end,
            disabled = function() return not settings.autoMarkAllEquipment end,
		},
		[61] = {
			type = "checkbox",
			name = fco_loc["options_equipment_markall_gear_add_jewelry"],
			tooltip = fco_loc["options_equipment_markall_gear_add_jewelry_tooltip"],
			getFunc = function() return settings.autoMarkAllJewelry end,
			setFunc = function(value) settings.autoMarkAllJewelry = not settings.autoMarkAllJewelry end,
            disabled = function() return not settings.autoMarkAllEquipment end,
		},

--==============================================================================
		[62] = {	type = "header",
				name = fco_loc["options_header_filter_buttons"],
	               },

		[63] = {
			type = "editbox",
			name = fco_loc["options_filter_button1_left"],
			tooltip = fco_loc["options_filter_button1_left_tooltip"],
			getFunc = function() return settings.filterButtonLeft[1] end,
			setFunc = function(newValue)
	           		settings.filterButtonLeft[1] = newValue
	                        updateFilterButtonsInInv(1)
			end,
			width = "half",
			default = settings.filterButtonLeft[1],
		},

		[64] = {
			type = "editbox",
			name = fco_loc["options_filter_button1_top"],
			tooltip = fco_loc["options_filter_button1_top_tooltip"],
			getFunc = function() return settings.filterButtonTop[1] end,
			setFunc = function(newValue)
	           		settings.filterButtonTop[1] = newValue
	                        updateFilterButtonsInInv(1)
			end,
			width = "half",
			default = settings.filterButtonTop[1],
		},

		[65] = {
			type = "editbox",
			name = fco_loc["options_filter_button2_left"],
			tooltip = fco_loc["options_filter_button2_left_tooltip"],
			getFunc = function() return settings.filterButtonLeft[2] end,
			setFunc = function(newValue)
	           		settings.filterButtonLeft[2] = newValue
	                        updateFilterButtonsInInv(2)
			end,
			width = "half",
			default = settings.filterButtonLeft[2],
		},

		[66] = {
			type = "editbox",
			name = fco_loc["options_filter_button2_top"],
			tooltip = fco_loc["options_filter_button2_top_tooltip"],
			getFunc = function() return settings.filterButtonTop[2] end,
			setFunc = function(newValue)
	           		settings.filterButtonTop[2] = newValue
	                        updateFilterButtonsInInv(2)
			end,
			width = "half",
			default = settings.filterButtonTop[2],
		},

		[67] = {
			type = "editbox",
			name = fco_loc["options_filter_button3_left"],
			tooltip = fco_loc["options_filter_button3_left_tooltip"],
			getFunc = function() return settings.filterButtonLeft[3] end,
			setFunc = function(newValue)
	           		settings.filterButtonLeft[3] = newValue
	                        updateFilterButtonsInInv(3)
			end,
			width = "half",
			default = settings.filterButtonLeft[3],
		},

		[68] = {
			type = "editbox",
			name = fco_loc["options_filter_button3_top"],
			tooltip = fco_loc["options_filter_button3_top_tooltip"],
			getFunc = function() return settings.filterButtonTop[3] end,
			setFunc = function(newValue)
	           		settings.filterButtonTop[3] = newValue
	                        updateFilterButtonsInInv(3)
			end,
			width = "half",
			default = settings.filterButtonTop[3],
		},

		[69] = {
			type = "editbox",
			name = fco_loc["options_filter_button4_left"],
			tooltip = fco_loc["options_filter_button4_left_tooltip"],
			getFunc = function() return settings.filterButtonLeft[4] end,
			setFunc = function(newValue)
	           		settings.filterButtonLeft[4] = newValue
	                        updateFilterButtonsInInv(4)
			end,
			width = "half",
			default = settings.filterButtonLeft[4],
		},

		[70] = {
			type = "editbox",
			name = fco_loc["options_filter_button4_top"],
			tooltip = fco_loc["options_filter_button4_top_tooltip"],
			getFunc = function() return settings.filterButtonTop[4] end,
			setFunc = function(newValue)
	           		settings.filterButtonTop[4] = newValue
	                        updateFilterButtonsInInv(4)
			end,
			width = "half",
			default = settings.filterButtonTop[4],
		},

		[71] = {
			type = "checkbox",
			name = fco_loc["options_filter_buttons_show_tooltip"],
			tooltip = fco_loc["options_filter_buttons_show_tooltip_tooltip"],
			getFunc = function() return settings.showFilterButtonTooltip end,
			setFunc = function(value) settings.showFilterButtonTooltip = not settings.showFilterButtonTooltip end,
		},

--==============================================================================
		[72] = {	type = "header",
				name = fco_loc["options_header_filter_chat"],
	               },
		[73] = {
			type = "checkbox",
			name = fco_loc["options_chat_filter_status"],
			tooltip = fco_loc["options_chat_filter_status_tooltip"],
			getFunc = function() return settings.showFilterStatusInChat end,
			setFunc = function(value) settings.showFilterStatusInChat = not settings.showFilterStatusInChat end,
		},

--==============================================================================
		[74] = {
			type = "header",
			name = fco_loc["options_header_anti_destroy"]
		},
		[75] = {
			type = "checkbox",
			name = fco_loc["options_enable_block_destroying"],
			tooltip = fco_loc["options_enable_block_destroying_tooltip"],
			getFunc = function() return settings.blockDestroying end,
			setFunc = function(value) settings.blockDestroying = not settings.blockDestroying
	           end,
		},
		[76] = {
			type = "checkbox",
			name = fco_loc["options_enable_block_deconstruction"],
			tooltip = fco_loc["options_enable_block_deconstruction_tooltip"],
			getFunc = function() return settings.blockDeconstruction end,
			setFunc = function(value) settings.blockDeconstruction = not settings.blockDeconstruction
	           end,
		},
		[77] = {
			type = "checkbox",
			name = fco_loc["options_enable_block_improvement"],
			tooltip = fco_loc["options_enable_block_improvement_tooltip"],
			getFunc = function() return settings.blockImprovement end,
			setFunc = function(value) settings.blockImprovement = not settings.blockImprovement
	           end,
		},
		[78] = {
			type = "checkbox",
			name = fco_loc["options_enable_block_selling"],
			tooltip = fco_loc["options_enable_block_selling_tooltip"],
			getFunc = function() return settings.blockSelling end,
			setFunc = function(value) settings.blockSelling = not settings.blockSelling
	           end,
		},
		[79] = {
			type = "checkbox",
			name = fco_loc["options_block_selling_exception_ornate"],
			tooltip = fco_loc["options_block_selling_exception_ornate_tooltip"],
			getFunc = function() return settings.allowSellingForBlockedOrnate end,
			setFunc = function(value) settings.allowSellingForBlockedOrnate = not settings.allowSellingForBlockedOrnate
	           end,
			disabled = function() return not settings.blockSelling end,
		},
		[80] = {
			type = "checkbox",
			name = fco_loc["options_enable_block_trading"],
			tooltip = fco_loc["options_enable_block_trading_tooltip"],
			getFunc = function() return settings.blockTrading end,
			setFunc = function(value) settings.blockTrading = not settings.blockTrading
	           end,
		},
		[81] = {
			type = "checkbox",
			name = fco_loc["options_enable_block_sending_mail"],
			tooltip = fco_loc["options_enable_block_sending_mail_tooltip"],
			getFunc = function() return settings.blockSendingByMail end,
			setFunc = function(value) settings.blockSendingByMail = not settings.blockSendingByMail
	           end,
		},

--==============================================================================
		[82] = {	type = "header",
				name = fco_loc["options_header_additional_buttons"],
	               },
		[83] = {
			type = "checkbox",
			name = fco_loc["options_additional_buttons_FCOIS_settings"],
			tooltip = fco_loc["options_additional_buttons_FCOIS_settings_tooltip"],
			getFunc = function() return settings.showFCOISMenuBarButton end,
			setFunc = function(value) settings.showFCOISMenuBarButton = not settings.showFCOISMenuBarButton
	            AddAdditionalButtons("FCOSettings")
            end,
		},

		[84] = {
			type = "checkbox",
			name = fco_loc["options_additional_buttons_FCOIS_additional_options"],
			tooltip = fco_loc["options_additional_buttons_FCOIS_additional_options_tooltip"],
			getFunc = function() return settings.showFCOISAdditionalInventoriesButton end,
			setFunc = function(value) settings.showFCOISAdditionalInventoriesButton = not settings.showFCOISAdditionalInventoriesButton
	            AddAdditionalButtons("FCOInventoriesContextMenuButtons")
            end,
		},

--==============================================================================

	} -- END OF OPTIONS TABLE

	LAM:RegisterOptionControls(addonVars.gAddonName, optionsTable)
end

--==============================================================================
--============================== END SETTINGS ==================================
--==============================================================================

--Check if the 4 filter buttons exist at the currently shown panel and create them if they are missing.
--Update the color of the buttons too
local function CheckFilterButtonsAtCurrentPanel(doUpdateLists, comingFrom)
	debugMessage("[CheckFilterButtonsAtCurrentPanel] Start - Coming from panel ID: " ..tostring(comingFrom), false)

	--Should we update the marker textures, size and color?
    checkMarker(-1)

	--get the currently shown panel
	local buttonsParentCtrl = checkActivePanel(comingFrom)

	--Is an inventory found?
    if buttonsParentCtrl ~= nil then
		debugMessage("[CheckFilterButtonsAtCurrentPanel] " .. buttonsParentCtrl:GetName() .. ", FilterPanelId: " .. tostring(locVars.gFilterWhere) .. ", UseFilters: " .. tostring(settings.atPanelEnabled[locVars.gFilterWhere]["filters"]), true)

		--Is the setting enabled to use the filters at this panel?
	    if settings.atPanelEnabled[locVars.gFilterWhere]["filters"] == true then

			local filterBtn
			local isFilterActivated

			--Change the filter buttons & callback functions
			for buttonNr=1, locVars.gFCONumFilters, 1 do
				filterBtn = AddOrChangeFilterButton(buttonsParentCtrl, buttonNr, locVars.gFilterButtonWidth, locVars.gFilterButtonHeight, settings.filterButtonLeft[buttonNr], settings.filterButtonTop[buttonNr], false, locVars.gFilterWhere)
	            if (filterBtn ~= nil) then
					--Get the filter's state
  					isFilterActivated = getSettingsIsFilterOn(buttonNr, locVars.gFilterWhere)
				    --Update the button's color
				    UpdateButtonColors(buttonNr, filterBtn, isFilterActivated, locVars.gFilterWhere)
					--(Re)register the filter (for the given panel)?
                    if isFilterActivated == false then
                        unregisterFilters(buttonNr, false, locVars.gFilterWhere)
                    else
	                    registerFilters(buttonNr, false, locVars.gFilterWhere)
                    end
	            end
			end

		else -- if settings.atPanelEnabled...

			--Change the filter buttons & callback functions
			for buttonNr=1, locVars.gFCONumFilters, 1 do
				--Hide the filter buttons
				AddOrChangeFilterButton(buttonsParentCtrl, buttonNr, locVars.gFilterButtonWidth, locVars.gFilterButtonHeight, settings.filterButtonLeft[buttonNr], settings.filterButtonTop[buttonNr], true, locVars.gFilterWhere)
			end
        	--Unregister the filter as the settings don't want a button and filter here
            unregisterFilters(-1, false, locVars.gFilterWhere)

        end -- if settings.atPanelEnabled...

		--Should the curent panel's iventory list be updated?
		if (doUpdateLists == true and settings.atPanelEnabled[locVars.gFilterWhere]["filters"] == true) then
	    	FilterBasics(false)
	    end
	end
end

--Check for other addons and react on them
local function CheckIfOtherAddonsActive()
    if (otherAddons.inventoryGridViewActive == false) then
		--Check if Inventory Gridview is active
	    local gridViewControlName = WINDOW_MANAGER:GetControlByName(otherAddons.GRIDVIEWBUTTON, "")
	    if gridViewControlName ~=  nil then
	    	debugMessage("Addon Inventory Gridview is active", false)
	    	otherAddons.inventoryGridViewActive = true
	    end
    end

	--Check if Chat Merchant is active
	if ( otherAddons.chatMerchantActive == false) then
	    otherAddons.chatMerchantActive = false
	    local chatMerchantControlName = WINDOW_MANAGER:GetControlByName(otherAddons.CHATMERCHANTBUTTON, "")
	    if chatMerchantControlName ~=  nil then
	    	debugMessage("Addon ChatMerchant is active", false)
	    	otherAddons.chatMerchantActive = true
        end
    end
	--Was ChatMerchant addon's control found now?
	if ( otherAddons.chatMerchantActive == true) then
		if chatMerchantControlName ~=  nil then
			chatMerchantControlName:ClearAnchors()
	        if (otherAddons.inventoryGridViewActive == true) then
		        -- With Inventory Grid View activated
		        chatMerchantControlName:SetAnchor(TOP, ZO_PlayerInventory, BOTTOM, -18, 6)
	        else
		        -- Without Inventory Grid View activated
		        chatMerchantControlName:SetAnchor(TOP, ZO_PlayerInventory, BOTTOM, -10, 6)
	        end
	    end
    end
end

--==============================================================================
--==================== START EVENT CALLBACK FUNCTIONS===========================
--==============================================================================

--Event upon opening of a crafting station
local function FCOItemSaver_Crafting_Interact(craftSkill, sameStation)
    preventerVars.gActiveFilterPanel = true
	--EVENT_MANAGER:RegisterForEvent(addonVars.gAddonName, EVENT_END_CRAFTING_STATION_INTERACT, FCOItemSaver_End_Crafting_Interact)

    --Abort if crafting station type is invalid
	if craftSkill == 0 then return end
	debugMessage("[EVENT] Crafting Interact: Craft skill: ".. tostring(craftSkill), true)

	--Check the filter buttons and create them if they are not there. Update the inventory afterwards too
	--CheckFilterButtonsAtCurrentPanel(true) -- done in PreHook of ZO_Smithing:SetMode()
end

--Event upon closing of a crafting station
local function FCOItemSaver_End_Crafting_Interact(...)
	debugMessage("[EVENT] End crafting", true)
	if preventerVars.gNoCloseEvent == false then
	    --Update the inventory filter buttons
	    updateFilterButtonsInInv(-1)
		--Update the 4 inventory button's color
		UpdateButtonColors(-1, nil, -1, INV_BAGS)
    end
	preventerVars.gNoCloseEvent      = false
    preventerVars.gActiveFilterPanel = false
    hideContextMenu()
end

--Event upon opening of a vendor store
local function FCOItemSaver_Open_Store()
    preventerVars.gActiveFilterPanel = true
	debugMessage("[EVENT] Open store", true)
	--Check the filter buttons and create them if they are not there. Update the inventory afterwards too
	CheckFilterButtonsAtCurrentPanel(true, INV_STORE)
end

--Event upon closing of a vendor store
local function FCOItemSaver_Close_Store()
	debugMessage("[EVENT] Close store", true)
	if preventerVars.gNoCloseEvent == false then
	    --Update the inventory filter buttons
	    updateFilterButtonsInInv(-1)
		--Update the 4 inventory button's color
		UpdateButtonColors(-1, nil, -1, INV_BAGS)
	end
	preventerVars.gNoCloseEvent 	 = false
    preventerVars.gActiveFilterPanel = false
    hideContextMenu()
end

--Event upon opening of a guild store
local function FCOItemSaver_Open_Trading_House()
    preventerVars.gActiveFilterPanel = true
	debugMessage("[EVENT] Open trading house", true)
	--Check the filter buttons and create them if they are not there. Update the inventory afterwards too
	CheckFilterButtonsAtCurrentPanel(true, INV_GUILDSTORE)
end

--Event upon closing of a guild store
local function FCOItemSaver_Close_Trading_House()
	debugMessage("[EVENT] Close trading house", true)
	if preventerVars.gNoCloseEvent == false then
	    --Update the inventory filter buttons
	    updateFilterButtonsInInv(-1)
		--Update the 4 inventory button's color
		UpdateButtonColors(-1, nil, -1, INV_BAGS)
    end
	preventerVars.gNoCloseEvent 	 = false
    preventerVars.gActiveFilterPanel = false
    hideContextMenu()
end

--Event upon opening of a guild bank
local function FCOItemSaver_Open_Guild_Bank()
    preventerVars.gActiveFilterPanel = true
	debugMessage("[EVENT] Open guild bank", true)
	--Check the filter buttons and create them if they are not there. Update the inventory afterwards too
	CheckFilterButtonsAtCurrentPanel(true, INV_GUILDBANK)
end

--Event upon closing of a guild bank
local function FCOItemSaver_Close_Guild_Bank()
	debugMessage("[EVENT] Close guild bank", true)
	if preventerVars.gNoCloseEvent == false then
	    --Update the inventory filter buttons
	    updateFilterButtonsInInv(-1)
		--Update the 4 inventory button's color
		UpdateButtonColors(-1, nil, -1, INV_BAGS)
    end
	preventerVars.gNoCloseEvent 	 = false
    preventerVars.gActiveFilterPanel = false
    hideContextMenu()
end

--Event upon opening of a player bank
local function FCOItemSaver_Open_Player_Bank()
    preventerVars.gActiveFilterPanel = true
	debugMessage("[EVENT] Open bank", true)
	--Check the filter buttons and create them if they are not there. Update the inventory afterwards too
	CheckFilterButtonsAtCurrentPanel(true, INV_BANK)
end

--Event upon closing of a player bank
local function FCOItemSaver_Close_Player_Bank()
	debugMessage("[EVENT] Close bank", true)
	if preventerVars.gNoCloseEvent == false then
	    --Update the inventory filter buttons
	    updateFilterButtonsInInv(-1)
		--Update the 4 inventory button's color
		UpdateButtonColors(-1, nil, -1, INV_BAGS)
    end
	preventerVars.gNoCloseEvent 	 = false
    preventerVars.gActiveFilterPanel = false
    hideContextMenu()
end

--Event upon opening of the trade panel
local function FCOItemSaver_Open_Trade_Panel()
   	preventerVars.gActiveFilterPanel = true
	debugMessage("[EVENT] Start trading", true)
	--Check the filter buttons and create them if they are not there. Update the inventory afterwards too
	CheckFilterButtonsAtCurrentPanel(true, INV_TRADE)
    hideContextMenu()
end

--Event upon closing of the trade panel
local function FCOItemSaver_Close_Trade_Panel()
	debugMessage("[EVENT] End trading", true)
	if preventerVars.gNoCloseEvent == false then
	    --Update the inventory filter buttons
	    updateFilterButtonsInInv(-1)
		--Update the 4 inventory button's color
		UpdateButtonColors(-1, nil, -1, INV_BAGS)
    end
	preventerVars.gNoCloseEvent 	 = false
   	preventerVars.gActiveFilterPanel = false
    hideContextMenu()
end

--PreHook function for the buttons at the top menus of banks, crafting stations, mail panel, tradin, etc.
local function FCOItemSaver_PreHookButtonHandler(comingFrom)
   	preventerVars.gActiveFilterPanel = true
	debugMessage("~~~~~~~~~~~~~~~~~~~~~~~~", true)
	debugMessage("[PreHookButtonHandler] Coming from panel ID: " ..tostring(comingFrom), true)

	--Check the filter buttons and create them if they are not there
	CheckFilterButtonsAtCurrentPanel(true, comingFrom)

    hideContextMenu()

	--Return false to call the normal callback handler of the button afterwards
    return false
end

-- Fires each time after addons were loaded and player is ready to move (after each zone change too)
local function FCOItemSaver_Player_Activated(...)
	--Prevent this event to be fired again and again upon each zone change
	EVENT_MANAGER:UnregisterForEvent(addonVars.gAddonName, EVENT_PLAYER_ACTIVATED)

    debugMessage("[EVENT] Player activated", true)

	--Check if other Addons active
    CheckIfOtherAddonsActive()

	--Add/update the filter buttons, but only if not done already in addon initialization
    if addonVars.gAddonLoaded == false then
	    updateFilterButtonsInInv(-1)
    end
    addonVars.gAddonLoaded = false

	--Output some text color values
	--local r, g, b, a = DEFAULT_TEXT_COLOR_NORMAL:UnpackRGBA()
	--debugMessage("Normal: " .. r .. ", " .. g .. ", " .. b .. ", " .. a, false)
	--r, g, b, a = DEFAULT_TEXT_COLOR_SELECTED:UnpackRGBA()
	--debugMessage("Selected: " .. r .. ", " .. g .. ", " .. b .. ", " .. a, false)
	--r, g, b, a = DEFAULT_TEXT_COLOR_HIGHLIGHT:UnpackRGBA()
	--debugMessage("Highlight: " .. r .. ", " .. g .. ", " .. b .. ", " .. a, false)
end

--event handler for ACTION LAYER POPPED
local function FCOItemsaver_OnActionLayerPopped(layerIndex, activeLayerIndex)
	debugMessage("[Action layer popped]", false)
    --Check if the mark all/demark all context menu should be closed
	hideContextMenu()
end

--==============================================================================
--===================== END EVENT CALLBACK FUNCTIONS============================
--==============================================================================


--==============================================================================
--			Anti - Destruction & Destroy & Send by Mail & Trade Methods
--==============================================================================

--Modify the destroy dialog
local function overrideDialogYesButton(dialog)
	if dialog.info == DestroyItemDialog then
		--Get the "yes" button control of the destroy popup window
		local button1 = dialog:GetNamedChild("Button1")

		--Use standard behaviour of "YES" button in destroy dialog
		button1:SetText(GetString(SI_YES))
		button1:SetClickSound(SOUNDS.DIALOG_ACCEPT)
		button1.m_callback = DestroyItemDialog.buttons[1].callback
		button1:SetMouseEnabled(true)
		button1:SetHidden(false)
		button1:SetKeybindEnabled(true)

        if not preventerVars.gAllowDestroyItem then
        	--Use own behaviour of "YES" button in destroy dialog:
            --Button's text will change to "NO"
            --Button will be set invisible and callback method changes so that
            --using a keybind will press ESC key instead
            button1:SetText(GetString(SI_NO))
            button1:SetClickSound(SOUNDS.DIALOG_DECLINE)
            button1.m_callback = DestroyItemDialog.noChoiceCallback
           	button1:SetMouseEnabled(false)
			button1:SetHidden(true)
			button1:SetKeybindEnabled(false)
        end
    end
end

-- fired when user selects an item to destroy
-- warns user if the item is marked with any of the filter icons
local function DestroySelectionHandler(bag, slot, echo, parentControl)
    debugMessage("[DestroySelectionHandler] Bag: " .. tostring(bag) .. ", Slot: " .. tostring(slot), true)
	--Are we coming from the character window?
	if bag == BAG_WORN and parentControl ~= nil then
    	preventerVars.gCheckEquipmentSlots = true
    end

	-- if not active then abort checking
	if( not settings.blockDestroying ) then return nil end

	-- get instance id of the item, this value is persistant across all game
    local id = MyGetItemInstanceIdNoControl(bag, slot)

	-- if item is in any protection list, warn user
	for i=1, locVars.gFCONumFilterIcons, 1 do
		if( markedItems[i][id] ) then
            if (echo == true) then
            	d(fco_loc["destroying_not_allowed"] .. " [" .. zo_strformat(SI_TOOLTIP_ITEM_NAME, GetItemName(bag, slot)) .. "]")
            end
			return true
		end
	end
	return false
end

-- fired when user selects an item by drag&drop, doubleclick etc.
-- warns user if the item is marked with any of the filter icons
local function ItemSelectionHandler(bag, slot, echo)
    debugMessage("[ItemSelectionHandler] Bag: " .. tostring(bag) .. ", Slot: " .. tostring(slot), true)
	local whereAreWe = 1
	--Inside mail panel?
  	if (not MAIL_SEND:IsHidden() or locVars.gFilterWhere == INV_MAIL) then
		-- if not active then abort checking
        whereAreWe = 2
		if( not settings.blockSendingByMail) then return false end
	--Inside trading player 2 player panel?
    elseif (not PLAYER_TRADE:IsHidden() or locVars.gFilterWhere == INV_TRADE) then
	    whereAreWe = 3
		-- if not active then abort checking
		if( not settings.blockTrading ) then return false end
	--Inside vendor sell?
    elseif (not VENDOR_SELL:IsHidden() or locVars.gFilterWhere == INV_STORE) then
	    whereAreWe = 4
		-- if not active then abort checking
		if( not settings.blockSelling ) then return false end
	--Inside crafting station improvement
    elseif (not IMPROVEMENT:IsHidden() or locVars.gFilterWhere == INV_IMPROVEMENT) then
	    whereAreWe = 5
		-- if not active then abort checking
		if( not settings.blockImprovement ) then return false end
	--Inside crafting station deconstruction
    elseif (not DECONSTRUCTION:IsHidden() or locVars.gFilterWhere == INV_DECONSTRUCTION) then
	    whereAreWe = 6
		-- if not active then abort checking
		if( not settings.blockDeconstruction ) then return false end
	--Inside enchanting station
    elseif (not ENCHANTING_STATION:IsHidden()) then
		--Enchanting Extraction panel?
    	if ENCHANTING.enchantingMode == 2 or locVars.gFilterWhere == INV_ENCHANTING_EXTRACTION then
		    whereAreWe = 7
			-- if not active then abort checking
			if( not settings.blockDeconstruction ) then return false end
		--Enchanting Creation panel?
    	elseif ENCHANTING.enchantingMode == 1 or locVars.gFilterWhere == INV_ENCHANTING_CREATION then
		    whereAreWe = 8
			-- if not active then abort checking
			if( not settings.blockDeconstruction ) then return false end
		end
	--Inside guild store selling?
    elseif (not GUILD_STORE:IsHidden() or locVars.gFilterWhere == INV_GUILDSTORE) then
	    whereAreWe = 9
		-- if not active then abort checking
		if( not settings.blockSelling ) then return false end
	--Are we at a bank/inventory and trying to withdraw/deposit some items by double clicking it?
    elseif (not BANK:IsHidden() or locVars.gFilterWhere == INV_BANK) then
       	return false
	--Are we at the inventory and trying to use/equip an item?
    elseif (not BACKPACK:IsHidden() or locVars.gFilterWhere == INV_BAGS) then
       	return false
	else
		whereAreWe = 1
		-- if not active then abort checking
		if( not settings.blockDestroying ) then return false end
	end

	-- get instance id of the item, this value is persistant across all game
    local id = MyGetItemInstanceIdNoControl(bag, slot)

	-- if item is in any protection list, warn user
	for i=1, locVars.gFCONumFilterIcons, 1 do
		if( markedItems[i][id] ) then

			--If current checked panel = sell or guild store sell and filter Id equals "Ornate" and the item is marked,
			--and the settings to allow selling of marked ornate items is enabled -> Abort here
			if ( (whereAreWe == 4 or whereAreWe == 9) and i==5
            	 and settings.atPanelEnabled[locVars.gFilterWhere]["anti-sell-exception"] == true ) then
				return false
			end

            if (echo == true) then
                if (whereAreWe == 2) then
	            	d(fco_loc["sendbymail_not_allowed"] .. " [" .. zo_strformat(SI_TOOLTIP_ITEM_NAME, GetItemName(bag, slot)) .. "]")
                elseif (whereAreWe == 3) then
	            	d(fco_loc["trading_not_allowed"] .. " [" .. zo_strformat(SI_TOOLTIP_ITEM_NAME, GetItemName(bag, slot)) .. "]")
                elseif (whereAreWe == 4) then
                	d(fco_loc["selling_not_allowed"] .. " [" .. zo_strformat(SI_TOOLTIP_ITEM_NAME, GetItemName(bag, slot)) .. "]")
                elseif (whereAreWe == 5) then
	            	d(fco_loc["improvement_not_allowed"] .. " [" .. zo_strformat(SI_TOOLTIP_ITEM_NAME, GetItemName(bag, slot)) .. "]")
                elseif (whereAreWe == 6) then
	            	d(fco_loc["deconstruction_not_allowed"] .. " [" .. zo_strformat(SI_TOOLTIP_ITEM_NAME, GetItemName(bag, slot)) .. "]")
                elseif (whereAreWe == 7) then
	            	d(fco_loc["enchanting_extraction_not_allowed"] .. " [" .. zo_strformat(SI_TOOLTIP_ITEM_NAME, GetItemName(bag, slot)) .. "]")
                elseif (whereAreWe == 8) then
	            	d(fco_loc["enchanting_creation_not_allowed"] .. " [" .. zo_strformat(SI_TOOLTIP_ITEM_NAME, GetItemName(bag, slot)) .. "]")
                elseif (whereAreWe == 9) then
	            	d(fco_loc["guild_store_sell_not_allowed"] .. " [" .. zo_strformat(SI_TOOLTIP_ITEM_NAME, GetItemName(bag, slot)) .. "]")
                else
                	d(fco_loc["destroying_not_allowed"] .. " [" .. zo_strformat(SI_TOOLTIP_ITEM_NAME, GetItemName(bag, slot)) .. "]")
                end
            end

			return true
		end
	end
	return false
end

-- fired when user selects an item to deconstruct
-- warns user if the item is marked with any of the filter icons
local function DeconstructionSelectionHandler(bag, slot, echo)
    debugMessage("[DeconstructionSelectionHandler] Bag: " .. tostring(bag) .. ", Slot: " .. tostring(slot), true)

	-- make sure user is on deconstruction panel
	if( not DECONSTRUCTION_BAG or DECONSTRUCTION_BAG:IsHidden() ) then
        if (not IMPROVEMENT:IsHidden() or locVars.gFilterWhere == INV_IMPROVEMENT) then
        	return ItemSelectionHandler(bag, slot, echo)
        elseif (not ENCHANTING_STATION:IsHidden() or locVars.gFilterWhere == INV_ENCHANTING_CREATION or locVars.gFilterWhere == INV_ENCHANTING_EXTRACTION) then
        	return ItemSelectionHandler(bag, slot, echo)
        else
    		return false
        end
    end

	-- if not active then abort checking
	if( not settings.blockDeconstruction ) then return nil end

	-- get instance id of the item, this value is persistant across all game
    local id = MyGetItemInstanceIdNoControl(bag, slot)

	-- if item is in any protection list, warn user
	for i=1, locVars.gFCONumFilterIcons, 1 do
		if( markedItems[i][id] ) then
            if (echo == true) then
               	d(fco_loc["deconstruction_not_allowed"] .. " [" .. zo_strformat(SI_TOOLTIP_ITEM_NAME, GetItemName(bag, slot)) .. "]")
            end

			return true
		end
	end
	return false
end


--==============================================================================
--			Right-click /context menu functions
--==============================================================================

--Pre Hook function for adding right click/context menu entries. Will be executed on mouse enter on a slot in inventories
--and if the slot entry is clicked
local function preAddSlotAction(self, actionStringId, ...)
    --If current panel is not activated for anti-destroy, anti-deconstruction, anti-mail, anti-trade, anti-sell:
    --local activePanel = checkActivePanel(locVars.gFilterWhere)
    --[[
	    settings.atPanelEnabled[activePanel]["anti-destroy"]
	    settings.atPanelEnabled[activePanel]["anti-deconstruct"]
	    settings.atPanelEnabled[activePanel]["anti-improve"]
	    settings.atPanelEnabled[activePanel]["anti-sell"]
	    settings.atPanelEnabled[activePanel]["anti-sell-exception"]
	    settings.atPanelEnabled[activePanel]["anti-trade"]
	    settings.atPanelEnabled[activePanel]["anti-mail"]
	]]

	local parentControl
	--Chracter equipment, or normal slot?
    if (not CHARACTER:IsHidden()) then
		parentControl = self.m_inventorySlot
    else
		parentControl = self.m_inventorySlot:GetParent()
    end
	--No parent found? Abort
	if parentControl == nil then return false end

    --debugMessage("[preAddSlotAction] Parent: " .. parentControl:GetName(), true)
    local bag
    local slotIndex
    bag, slotIndex = MyGetItemDetails(parentControl)

	--Hide the button contextMenu if shown and if we right clicked another item
	if self.m_contextMenuMode == true then
		--Hide the button contextMenu if shown
		hideContextMenu()
    end

	--Destroy item, trade item, attach item to mail or sell item
    if    actionStringId == SI_ITEM_ACTION_DESTROY then

	    --Only execute if context menu is visible?
	    if self.m_contextMenuMode == false then
			return
        end

		--Abort if parent control cannot be found
        --Is item marked with any of the FCOItemSaver icons? Then don't show the actionStringId in the contextmenu
    	return DestroySelectionHandler(bag, slotIndex, false, parentControl)

	--Add item to crafting station, improvement, enchanting table
	elseif actionStringId == SI_ITEM_ACTION_ADD_TO_CRAFT or actionStringId == SI_ITEM_ACTION_RESEARCH then
        --Is item marked with any of the FCOItemSaver icons? Then don't show the actionStringId in the contextmenu
    	return DeconstructionSelectionHandler(bag, slotIndex, false)

	--Trade an item, mark item as junk, attach item to mail, sell item
    elseif actionStringId == SI_ITEM_ACTION_TRADE_ADD or actionStringId == SI_ITEM_ACTION_MARK_AS_JUNK
       or actionStringId == SI_ITEM_ACTION_MAIL_ATTACH or actionStringId == SI_ITEM_ACTION_SELL
	   --Anbieten / sell in guild shop
       or actionStringId == 3763 then
        --Is item marked with any of the FCOItemSaver icons? Then don't show the actionStringId in the contextmenu
        return ItemSelectionHandler(bag, slotIndex, false)

	--"Ausrüsten"=1927
	--elseif actionStringId == SI_ITEM_ACTION_EQUIP then
		--Only for character items
        --d("Ausrüsten!")
		--return false

	--"Ablegen"=1928
	elseif actionStringId == SI_ITEM_ACTION_UNEQUIP then
	    --Only execute if context menu is visible?
	    if self.m_contextMenuMode == false then
			return
        end

		--Only for character items
		if( bag ~= nil and slotIndex ~= nil and bag == BAG_WORN) then
	        --d("Ablegen!")
			--Check if the markers for the character slots are still correct or can be hidden
			zo_callLater(function() checkForEmptyOrTwoHandedWeaponMarker(slotIndex, false) end, 200)
		end
		return

    --else
    	--d("actionStringId: " .. actionStringId .. " = " .. GetString(actionStringId))
    end
end


--==============================================================================
--			ON EVENT Implementation Methods
--==============================================================================

-- handler function for EVENT_INVENTORY_SLOT_UNLOCKED global event
-- will be fired (after EVENT_INVENTORY_SLOT_LOCKED) if you have pickuped an item (e.g. by drag&drop) and drop it again
local function FCOItemSaver_OnInventorySlotUnLocked(self, bag, slot)
	debugMessage("[Event] OnInventorySlotUnLocked: bag: " .. tostring(bag) .. ", slot: " .. tostring(slot), true)

	if preventerVars.gCheckEquipmentSlots == true and preventerVars.gItemSlotIsLocked == true then
		zo_callLater(function() checkForEmptyOrTwoHandedWeaponMarker(slot, false) end, 100)
        preventerVars.gCheckEquipmentSlots = false
    end

    preventerVars.gItemSlotIsLocked = false
end

-- handler function for EVENT_INVENTORY_SLOT_LOCKED global event
-- will be fired (before EVENT_CURSOR_PICKUP) if you pickup an item (e.g. by drag&drop)
local function FCOItemSaver_OnInventorySlotLocked(self, bag, slot)
	debugMessage("[Event] OnInventorySlotLocked", true)

    preventerVars.gItemSlotIsLocked = true

	if(not DECONSTRUCTION_BAG:IsHidden() ) then
		-- check if deconstruction is forbidden
		-- if so, clear item hold by cursor
		if( DeconstructionSelectionHandler(bag, slot, true) ) then
			--Remove the picked item from drag&drop cursor
	        --HideKeybind()
	    	ClearCursor()
            return false
	    end
    else
        --Enable drag&drop during crafting improvement -> Destroying is not possible during crafting
        --except for deconstruction, which is handled with different methods (see above)
       	if(not IMPROVEMENT_BAG:IsHidden() ) then
        	if (not settings.blockImprovement) then
        		return false
            end

    	--Enable drag&drop during mail sending if enabled in the settings
    	elseif (not GUILD_STORE:IsHidden()) then
        	if (not settings.blockSelling) then
        		return false
            end

    	--Enable drag&drop during mail sending if enabled in the settings
    	elseif (not MAIL_SEND:IsHidden()) then
        	if (not settings.blockSendingByMail) then
        		return false
            end

    	--Enable drag&drop during player trading if enabled in the settings
    	elseif (not PLAYER_TRADE:IsHidden()) then
        	if (not settings.blockTrading) then
        		return false
            end

        --Enable drag&drop during selling at a vendor if enabled in the settings
    	elseif (not VENDOR_SELL:IsHidden()) then
        	if (not settings.blockSelling) then
        		return false
            end

        --Enable drag&drop during enchanting (creation & extraction) if enabled in the settings
    	elseif (not ENCHANTING_STATION:IsHidden()) then
        	if (not settings.blockDeconstruction) then
        		return false
            end

        --Enable drag&drop in the inventory, bank and guild banks to stack, or disposite e.g.
        elseif (not BACKPACK_BAG:IsHidden() or not BANK_BAG:IsHidden() or not GUILD_BANK_BAG:IsHidden()) then
        	return false
        end

		-- check if destroying, sending or trading is forbidden
		-- if so, clear item hold by cursor
		if( ItemSelectionHandler(bag, slot, true) ) then
			--Remove the picked item from drag&drop cursor
	        --HideKeybind()
	    	ClearCursor()
            return false
	    end
    end
end

-- handler function for character window item controls' OnMouseDoubleClick event
local function FCOItemSaver_CharacterItem_OnMouseDoubleClick(self, ...)
    local bagId, slotId

	-- make sure control contains an item
	if( self.dataEntry and self.dataEntry.data ) then
		bagId = self.dataEntry.data.bagId
		slotId = self.dataEntry.data.slotIndex
    else
		if( self.slotIndex and self.bagId ) then
			bagId  = self.bagId
			slotId = self.slotIndex
		end
    end

	--Only for character items
	if( bagId ~= nil and slotId ~= nil and bagId == BAG_WORN and self.stackCount > 0) then
		--Check if the markers for the character slots are still correct or can be hidden
		zo_callLater(function() checkForEmptyOrTwoHandedWeaponMarker(slotId, false) end, 200)
	end

	-- call the original handler function
	local func = GetEventHandler("OnMouseDoubleClick", self:GetName())
	if ( not func ) then return false end

	return func(self, ...)
end


-- handler function for inventory item controls' OnMouseDoubleClick event
local function FCOItemSaver_InventoryItem_OnMouseDoubleClick(self, ...)
    local bagId, slotId
	-- make sure control contains an item
	if( self.dataEntry and self.dataEntry.data ) then
		bagId = self.dataEntry.data.bagId
		slotId = self.dataEntry.data.slotIndex
    else
		if( self.slotIndex and self.bagId ) then
			bagId  = self.bagId
			slotId = self.slotIndex
		end
    end

	if( bagId ~= nil and slotId ~= nil ) then

	    -- Inside deconstruction?
		if(not DECONSTRUCTION_BAG:IsHidden() ) then
			-- check if deconstruction is forbidden
			-- if so, return false to prevent selection of the item
			if( DeconstructionSelectionHandler(bagId, slotId, true) ) then
                --Enable the OnMouseEnter and OnMouseExit events
            	--preventerVars.gNoOnMouseEvents = false
            	--HideKeybind()
                ClearCursor()
            	return false
            else
				--ShowKeybind()
            end
        --Others
        else
            if( ItemSelectionHandler(bagId, slotId, true) ) then
            	--HideKeybind()
                ClearCursor()
				-- item is not allowed to work with return false to prevent selection of the item
            	return false
            else
            	--ShowKeybind()
            end
		end

	end

	-- call the original handler function
	local func = GetEventHandler("OnMouseDoubleClick", self:GetName())
	if ( not func ) then return false end

	return func(self, ...)
end

--==============================================================================
--			ON EVENT Methods
--==============================================================================

-- handler function for decontruction bag's OnEffectivelyShown event
local function FCOItemSaver_CharacterOnEffectivelyShown(self, ...)
    hideContextMenu()

	if ( not self ) then return false end

    local equipmentSlotName
	for i = 1, self:GetNumChildren() do
		-- override OnMouseDoubleClick event of character window item controls, for each row (children)
    	equipmentSlotName = CHARACTER:GetChild(i):GetName()
        if equipmentSlotName ~= nil then
			if(string.find(equipmentSlotName, "ZO_CharacterEquipmentSlots")) then
				if( self:GetChild(i):GetHandler("OnMouseDoubleClick") ~= FCOItemSaver_CharacterItem_OnMouseDoubleClick ) then
					PreHookHandler( "OnMouseDoubleClick", self:GetChild(i), FCOItemSaver_CharacterItem_OnMouseDoubleClick )
				end
	        end
        end
	end
	-- call the original handler function
	local func = GetEventHandler("OnEffectivelyShown", self:GetName())
	if ( not func ) then return false end
	return func(self, ...)
end

-- handler function for decontruction bag's OnEffectivelyShown event
local function FCOItemSaver_OnEffectivelyShown(self, ...)
    hideContextMenu()

	--Should we update the marker textures, size and color?
    checkMarker(-1)

	if ( not self ) then return false end

	for i = 1, self:GetNumChildren() do
		-- override OnMouseDoubleClick event of inventory item controls, for each row (children)
		if( self:GetChild(i):GetHandler("OnMouseDoubleClick") ~= FCOItemSaver_InventoryItem_OnMouseDoubleClick ) then
			PreHookHandler( "OnMouseDoubleClick", self:GetChild(i), FCOItemSaver_InventoryItem_OnMouseDoubleClick )
		end
	end
	-- call the original handler function
	local func = GetEventHandler("OnEffectivelyShown", self:GetName())
	if ( not func ) then return false end
	return func(self, ...)
end

--Executed if item should be destroyed manually
local function FCOItemSaver_OnMouseRequestDestroyItem(eventCode, bagId, slotIndex, itemCount, name, needsConfirm)
    hideContextMenu()

    if not needsConfirm then
		--Only react if anti destroy setting is enabled
	    if (not settings.blockDestroying) then
	        preventerVars.gAllowDestroyItem = true
        	return nil
        end

		preventerVars.gAllowDestroyItem = false

		if( bagId and slotIndex ) then
	    	-- check if deconstruction/destroying is forbidden because item is marked
			if( DestroySelectionHandler(bagId, slotIndex, true) == true ) then
		        preventerVars.gAllowDestroyItem = false
	        else
		        preventerVars.gAllowDestroyItem = true
	        end
	        --Change the "YES" button of the destroy dialog and disable keybind
	        overrideDialogYesButton(ZO_Dialog1)
		end
	end
end

--==============================================================================
--==============================================================================

--Inventory slot gets updated function
function FCOItemSaver_Inv_Update( eventCode, bagId, slotId, isNewItem, itemSoundCategory, updateReason )
    --Check if item in slot is still there
	if GetItemType(bagId, slotId) == ITEMTYPE_NONE then return end

    --Abort if not new item is added to inventory
	if (not isNewItem) then return end

    --Only check for normal player inventory
    if (bagId == BAG_BACKPACK) then
		debugMessage("==============", true)
		debugMessage("Event 'Single inventory slot updated' raised! NewItem=" .. tostring(isNewItem) .. ", bagId=" .. bagId .. ", slotIndex=" .. slotId .. ", updateReason=" .. tostring(updateReason), true)
		--if(preventerVars.canUpdateInv == true) then
			--preventerVars.canUpdateInv = false
			zo_callLater(function()
   				debugMessage("Event 'Single inventory slot updated' executed now! bagId=" .. bagId .. ", slotIndex=" .. slotId, true)
                scanInventory(bagId, slotId)
            end, 250)
			--preventerVars.canUpdateInv = true
		--end
    end
end

--==============================================================================
--===== HOOKS BEGIN ============================================================
--==============================================================================
--Create the hooks & pre-hooks
local function CreateHooks()
	local parentControl
   	-- Hook functions for the inventory/store contextmenus
	ZO_PreHook("ZO_InventorySlot_ShowContextMenu", function(rowControl)
        local noGear
        local gearId

		--check each iconId
		for i = 1, locVars.gFCONumFilterIcons, 1 do
            --Get the gear Id
        	gearId = iconToGear[i]

			--Was a gearId found?
            if gearId == nil then
            	noGear = true
            else
            	noGear = false
            end

			--Only add context menu entry if no gear Id or gear Id is enabled in settings
			if noGear == true or settings.isGearEnabled[gearId] then
	    		parentControl = rowControl:GetParent()
	        	if(parentControl ~= CHARACTER) then
					zo_callLater(function() AddMark(parentControl, i, false) end, 50)
				else
					zo_callLater(function() AddMark(rowControl, i, true) end, 50)
				end
			end
        end
	end)

	-- Hook functions for the item equiped event in character window
	ZO_PreHook("PlayOnEquippedAnimation", function (parent)
		local parentName = parent:GetName()
		debugMessage("[Hook] Play on equipped animation: Parent="..parentName, true)

        --Only execute if character window is shown
        if (not CHARACTER:IsHidden()) then
	        --Add or refresh the equipped items, create marker control if not already there
	        RefreshEquipmentControl(parent, true)

			--Are we chaning equipped weapons?
			--MainHand and OffHand
			if   parentName == characterEquipmentSlotNameByIndex[4]  or parentName == characterEquipmentSlotNameByIndex[5]
			--BackupHand and BackupOffHand
			  or parentName == characterEquipmentSlotNameByIndex[20] or parentName == characterEquipmentSlotNameByIndex[21] then
				--Get the slot of the current character equipment control
	            local slot = parent.slotIndex
				--Check if the offhand weapons are 2hd weapons and remove the markers then
				zo_callLater(function() checkForEmptyOrTwoHandedWeaponMarker(slot, true) end, 100)
			end
        end
	end)

	--Crafting station improvement inventory hook
	local hookedFunctions = IMPROVEMENT.dataTypes[1].setupCallback
	IMPROVEMENT.dataTypes[1].setupCallback = function(rowControl, slot)
			hookedFunctions(rowControl, slot)
            -- for all filters: Create/Update the icons
			for i = 1, locVars.gFCONumFilterIcons, 1 do
				CreateMarkerControl(rowControl, i, locVars.gIconWidth, locVars.gIconHeight, MARKER_TEXTURES[settings.icon[i].texture])
            end
			--Add additional FCO point to the dataEntry.data slot
			--FCOItemSaver_AddInfoToData(rowControl)
		end

	--deconstruction hook
	local hookedFunctions = DECONSTRUCTION.dataTypes[1].setupCallback
	DECONSTRUCTION.dataTypes[1].setupCallback = function(rowControl, slot)
			hookedFunctions(rowControl, slot)
            -- for all filters: Create/Update the icons
			for i = 1, locVars.gFCONumFilterIcons, 1 do
				CreateMarkerControl(rowControl, i, locVars.gIconWidth, locVars.gIconHeight, MARKER_TEXTURES[settings.icon[i].texture])
            end
			--Add additional FCO point to the dataEntry.data slot
			--FCOItemSaver_AddInfoToData(rowControl)
		end

	--enchanting hook
	local hookedFunctions = ENCHANTING_STATION.dataTypes[1].setupCallback
	ENCHANTING_STATION.dataTypes[1].setupCallback = function(rowControl, slot)
			hookedFunctions(rowControl, slot)
            -- for all filters: Create/Update the icons
			for i = 1, locVars.gFCONumFilterIcons, 1 do
				CreateMarkerControl(rowControl, i, locVars.gIconWidth, locVars.gIconHeight, MARKER_TEXTURES[settings.icon[i].texture])
            end
			--Add additional FCO point to the dataEntry.data slot
			--FCOItemSaver_AddInfoToData(rowControl)
		end

	--research list hook
	local hookedFunctions = LIST_DIALOG.dataTypes[1].setupCallback
	LIST_DIALOG.dataTypes[1].setupCallback = function(rowControl, slot)
		hookedFunctions(rowControl, slot)

   		local data = rowControl.dataEntry.data

		--Check if rowControl is a soulgem
		--GLOBALOL = data
		local isSoulGem = false
		if(data and GetSoulGemItemInfo(data.bag, data.index) > 0) then
			isSoulGem = true
		end

        local myItemInstanceIdOfControl = MyGetItemInstanceId(rowControl)
		--Is researching allowed in the settings?
        local isFilterActivated = not settings.allowResearch
        local disableResearch

		-- Check the rowControl if the item is marked and update the OnMouseUp functions and the color of the item row then
		for j = 1, locVars.gFCONumFilters, 1 do
        	--isFilterActivated = getSettingsIsFilterOn(j)
			--Just to test: Set filter activated to always true and only check if the item is marked,
            -- without checking if the relating filter is activated at the deconstruction panel

            if (j == 1 ) then
	        	if(isFilterActivated == true and not isSoulGem and markedItems[j][myItemInstanceIdOfControl]) then
                    disableResearch = true
                    break -- leave for ... do loop
                else
                	disableResearch = false
	            end
            --Special behaviour for filter 2, as it is controling icons 2, 4, 6, 7 and 8 (Gear sets) too
            elseif (j == 2) then
				if(isFilterActivated == true and not isSoulGem and
                   (   markedItems[j][myItemInstanceIdOfControl] or markedItems[4][myItemInstanceIdOfControl]
                    or markedItems[6][myItemInstanceIdOfControl] or markedItems[7][myItemInstanceIdOfControl]
                    or markedItems[8][myItemInstanceIdOfControl])
                ) then
                    disableResearch = true
                    break -- leave for ... do loop
                else
                    disableResearch = false
                end
            --Special behaviour for filter 3, Research filter
            elseif (j == 3) then
				if(isFilterActivated == true and not isSoulGem and markedItems[j][myItemInstanceIdOfControl]) then
                    disableResearch = true
                    break -- leave for ... do loop
                else
					disableResearch = false
                end
			--Special behaviour for filter 4, as it controls the marked items 5
			elseif (j == 4) then
				if(isFilterActivated == true and not isSoulGem and markedItems[j+1][myItemInstanceIdOfControl]) then
                    disableResearch = true
                    break -- leave for ... do loop
                else
                    disableResearch = false
				end
			-- All other filters
            else
	        	if(isFilterActivated == true and not isSoulGem and markedItems[j][myItemInstanceIdOfControl]) then
                    disableResearch = true
                    break -- leave for ... do loop
                else
                    disableResearch = false
	            end
			end


        end -- for j = 1, locVars.gFCONumFilters, 1 do

		--Get here after for is "breaked" and item is not researchable
        if disableResearch == true then
				--Change the color of the item to red
        		rowControl:GetNamedChild("Name"):SetColor(0.75, 0, 0, 1)
        end

        -- for all filters: Create/Update the icons
		for i = 1, locVars.gFCONumFilterIcons, 1 do
			CreateMarkerControl(rowControl, i, locVars.gIconWidth, locVars.gIconHeight, MARKER_TEXTURES[settings.icon[i].texture])

			--Pre-Hook the list dialog's rows
			-- The current "rowControl" is a row at the popup dialog ZO_ListDialog1 (crafting station research popup e.g.)
	        -- and we are currently creating textures for filter 1 (to prevent adding the context menu entries 8 times
	        -- for every of the 8 textures)
	        if (i==1) then

		        --Pre-Hook the handler "OnMouseUp" event for the rowControl to disable the researching of the item,
		        --but still enable the right click/context menu:
				--Show context menu at mouse button 2, but keep the normal OnMouseUp handler as well
				ZO_PreHookHandler(rowControl, "OnMouseUp", function(control, button, upInside)
					debugMessage("Clicked: " ..control:GetName() .. ", MouseButton: " .. tostring(button), true)
					--button 1= left mouse button / 2= right mouse button

					--Right click/mouse button 2 context menu hook part:
					if button == 2 and upInside then
						--Add the menu items to the popup dialog row (parent)
			            ClearMenu()
			            for j = 1, locVars.gFCONumFilterIcons, 1 do
							zo_callLater(function() AddMark(control, j, false, true) end, 50)
						end
	                end

					--Disable the researching of this item part:
					if disableResearch == true then
						--Do nothing (true tells the handler function that everything was achieved already in this function
						--and the normal "hooked" functions don't need to be run afterwards
						return true
					else -- if disableResearch == false
		                --Is the row selected?
		            	if LIST_DIALOG.selectedControl ~= nil then
							--Enable the button again, as it will be disabled if the item won't be researchable
			  				WINDOW_MANAGER:GetControlByName("ZO_ListDialog1Button1", ""):SetEnabled(true)
		                end
					end -- if disableResearch == true

				end) -- ZO_PreHookHandler(rowControl, "OnMouseUp"...

			end -- if (i==1) then
		end -- for i = 1, locVars.gFCONumFilterIcons, 1 do

	end -- research list pre-hook


--======== INVENTORY ===========================================================
	--Pre Hook the inventory for prevention methods
	PreHookHandler( "OnEffectivelyShown",      BACKPACK_BAG, FCOItemSaver_OnEffectivelyShown )
--======== CHARACTER ===========================================================
	--Pre Hook the character window for prevention methods
	PreHookHandler( "OnEffectivelyShown",      CHARACTER, FCOItemSaver_CharacterOnEffectivelyShown )

--======== RIGHT CLICK / CONTEXT MENU ==========================================
	--Pre Hook the right click/context menu addition of items
    ZO_PreHook(ZO_InventorySlotActions, "AddSlotAction", preAddSlotAction)

--======== BANK ================================================================
    --Pre Hook the 2 menubar button's (take and deposit) handler at the bank
	ZO_PreHookHandler(BANK_MENUBAR_BUTTON_WITHDRAW, "OnMouseUp", function(control, button, upInside)
        if (button == 1 and upInside and locVars.gLastBankButton~=BANK_MENUBAR_BUTTON_WITHDRAW) then
            locVars.gLastBankButton = BANK_MENUBAR_BUTTON_WITHDRAW
            zo_callLater(function() FCOItemSaver_PreHookButtonHandler(INV_BANK) end, 50)
        end
    end)
	ZO_PreHookHandler(BANK_MENUBAR_BUTTON_DEPOSIT, "OnMouseUp", function(control, button, upInside)
        if (button == 1 and upInside and locVars.gLastBankButton~=BANK_MENUBAR_BUTTON_DEPOSIT) then
            locVars.gLastBankButton = BANK_MENUBAR_BUTTON_DEPOSIT
            zo_callLater(function() FCOItemSaver_PreHookButtonHandler(INV_BAGS) end, 50)
        end
    end)

--======== GUILD BANK ==========================================================
    --Pre Hook the 2 menubar button's (take and deposit) handler at the guild bank
	ZO_PreHookHandler(GUILD_BANK_MENUBAR_BUTTON_WITHDRAW, "OnMouseUp", function(control, button, upInside)
        if (button == 1 and upInside and locVars.gLastGuildBankButton~=GUILD_BANK_MENUBAR_BUTTON_WITHDRAW) then
            locVars.gLastGuildBankButton = GUILD_BANK_MENUBAR_BUTTON_WITHDRAW
            zo_callLater(function() FCOItemSaver_PreHookButtonHandler(INV_GUILDBANK) end, 50)
        end
    end)
	ZO_PreHookHandler(GUILD_BANK_MENUBAR_BUTTON_DEPOSIT, "OnMouseUp", function(control, button, upInside)
        if (button == 1 and upInside and locVars.gLastGuildBankButton~=GUILD_BANK_MENUBAR_BUTTON_DEPOSIT) then
            locVars.gLastGuildBankButton = GUILD_BANK_MENUBAR_BUTTON_DEPOSIT
            zo_callLater(function() FCOItemSaver_PreHookButtonHandler(INV_BAGS) end, 50)
        end
    end)

--======== MAIL ================================================================
	--Register a callback function for the mail scene
	MAIL_SEND_SCENE:RegisterCallback("StateChange", function(oldState, newState)
	    hideContextMenu()

		debugMessage("[MAIL_SEND_SCENE] State: " .. tostring(newState), true)

    	--When the mail send panel is showing up
		if newState == SCENE_SHOWING then
        	--Check if another filter panel was already opened and we are coming form there before the CLOSE EVENT function was called
			if preventerVars.gActiveFilterPanel == true then
				--Set the "No Close Event" flag so the called close event won't override gFilterWhere and update the filter button colors and callback handlers
				preventerVars.gNoCloseEvent = true
            end
			--Check the filter buttons and create them if they are not there. Update the inventory afterwards too
			CheckFilterButtonsAtCurrentPanel(true, INV_MAIL)

		--When the mail send panel is hidden
		elseif newState == SCENE_HIDDEN then
		    --Update the inventory filter buttons
		    updateFilterButtonsInInv(-1)
			--Update the 4 inventory button's color
			UpdateButtonColors(-1, nil, -1, INV_BAGS)
        end
	end)

--======== SMITHING=============================================================
--Prehook the smithing function SetMode() which gets executed as the smithing tabs are changed
	ZO_PreHook(ZO_Smithing, "SetMode", function(smithing_obj, mode)
	    hideContextMenu()

		debugMessage("[ZO_Smithing:SetMode] Mode: " .. tostring(mode), true)
    	--[[ mode could be:
			SMITHING_MODE_REFINMENT
			SMITHING_MODE_CREATION
			SMITHING_MODE_IMPROVEMENT then
			SMITHING_MODE_DECONSTRUCTION
			SMITHING_MODE_RESEARCH
        ]]
        --Deconstruction
        if     mode == SMITHING_MODE_DECONSTRUCTION then
	        FCOItemSaver_PreHookButtonHandler(INV_DECONSTRUCTION)
            zo_callLater(function() FCOItemSaver_OnEffectivelyShown(DECONSTRUCTION_BAG) end, 100)
		--Improvement
        elseif mode == SMITHING_MODE_IMPROVEMENT then
	        FCOItemSaver_PreHookButtonHandler(INV_IMPROVEMENT)
            zo_callLater(function() FCOItemSaver_OnEffectivelyShown(IMPROVEMENT_BAG) end, 100)
        end
        --Go on with original function
        return false
	end)

--======== ENCHANTING===========================================================
--Prehook the enchanting function SetEnchantingMode() which gets executed as the enchanting tabs are changed
	ZO_PreHook(ZO_Enchanting, "SetEnchantingMode", function(enchanting_obj, enchantingMode)
	    hideContextMenu()

		debugMessage("[ZO_Enchanting:SetEnchantingMode] EnchantingMode: " .. tostring(enchantingMode), true)
    	--[[ enchantingMode could be:
			ENCHANTING_MODE_CREATION
			ENCHANTING_MODE_EXTRACTION
        ]]
        --Creation
        if     enchantingMode == ENCHANTING_MODE_CREATION then
	        FCOItemSaver_PreHookButtonHandler(INV_ENCHANTING_CREATION)
            zo_callLater(function() FCOItemSaver_OnEffectivelyShown(ENCHANTING_STATION_BAG) end, 100)
		--Extraction
        elseif enchantingMode == ENCHANTING_MODE_EXTRACTION then
	        FCOItemSaver_PreHookButtonHandler(INV_ENCHANTING_EXTRACTION)
            zo_callLater(function() FCOItemSaver_OnEffectivelyShown(ENCHANTING_STATION_BAG) end, 100)
        end
        --Go on with original function
        return false
	end)

end
--==============================================================================
--===== HOOKS END ==============================================================
--==============================================================================


--Localized texts etc.
local function Localization()
    --Fallback to english
	if (defaultSettings.language ~= 1 and defaultSettings.language ~= 2 and defaultSettings.language ~= 3) then
    	defaultSettings.language = 1
    end
	--Is the standard language english set?
    if (defaultSettings.language == 1 and settings.languageChoosen == false) then
		local lang = GetCVar("language.2")
		--Check for supported languages
		if(lang == "de") then
	    	defaultSettings.language = 2
	    elseif (lang == "en") then
	    	defaultSettings.language = 1
	    elseif (lang == "fr") then
	    	defaultSettings.language = 3
		else
	    	defaultSettings.language = 1
	    end
	end
	debugMessage("[Localization] default settings, language: " .. tostring(defaultSettings.language), false)
    fco_loc	= fco_loc[defaultSettings.language]

    --Prepare the texts for the right click menus
    contextEntries.menu_add_lock_text  	   = fco_loc["rightclick_menu_lock"]
	contextEntries.menu_add_gear1_text	 	   = fco_loc["rightclick_menu_add_gear1"]
	contextEntries.menu_add_research_text     = fco_loc["rightclick_menu_mark_analysis"]
	contextEntries.menu_add_gear2_text	 	   = fco_loc["rightclick_menu_add_gear2"]
	contextEntries.menu_add_sell_text 	 	   = fco_loc["rightclick_menu_mark_sell"]
	contextEntries.menu_add_gear3_text 	   = fco_loc["rightclick_menu_add_gear3"]
	contextEntries.menu_add_gear4_text 	   = fco_loc["rightclick_menu_add_gear4"]
	contextEntries.menu_add_gear5_text 	   = fco_loc["rightclick_menu_add_gear5"]

	contextEntries.menu_add_all_gear1_text	   = fco_loc["rightclick_menu_add_all_gear1"]
	contextEntries.menu_add_all_gear2_text	   = fco_loc["rightclick_menu_add_all_gear2"]
	contextEntries.menu_add_all_gear3_text    = fco_loc["rightclick_menu_add_all_gear3"]
	contextEntries.menu_add_all_gear4_text    = fco_loc["rightclick_menu_add_all_gear4"]
	contextEntries.menu_add_all_gear5_text    = fco_loc["rightclick_menu_add_all_gear5"]

    contextEntries.menu_remove_lock_text  	   = fco_loc["rightclick_menu_unlock"]
	contextEntries.menu_remove_gear1_text	   = fco_loc["rightclick_menu_remove_gear1"]
	contextEntries.menu_remove_research_text  = fco_loc["rightclick_menu_demark_analysis"]
	contextEntries.menu_remove_gear2_text	   = fco_loc["rightclick_menu_remove_gear2"]
	contextEntries.menu_remove_sell_text 	   = fco_loc["rightclick_menu_demark_sell"]
	contextEntries.menu_remove_gear3_text 	   = fco_loc["rightclick_menu_remove_gear3"]
	contextEntries.menu_remove_gear4_text 	   = fco_loc["rightclick_menu_remove_gear4"]
	contextEntries.menu_remove_gear5_text 	   = fco_loc["rightclick_menu_remove_gear5"]

	contextEntries.menu_remove_all_gear1_text = fco_loc["rightclick_menu_remove_all_gear1"]
	contextEntries.menu_remove_all_gear2_text = fco_loc["rightclick_menu_remove_all_gear2"]
	contextEntries.menu_remove_all_gear3_text = fco_loc["rightclick_menu_remove_all_gear3"]
	contextEntries.menu_remove_all_gear4_text = fco_loc["rightclick_menu_remove_all_gear4"]
	contextEntries.menu_remove_all_gear5_text = fco_loc["rightclick_menu_remove_all_gear5"]

	--Overwrite the localized texts for the equipment gears, if changed in the settings
	changeEquipmentGearMenuTexts(-1)
end

--Register the slash commands
local function RegisterSlashCommands()
    -- Register slash commands
	SLASH_COMMANDS["/fcoitemsaver"] = command_handler
	SLASH_COMMANDS["/fcois"] 		= command_handler
end

--Addon loads up
local function FCOItemSaver_Loaded(eventCode, addOnName)
	--Check if addon "Research Assistant" is active
	if(addOnName == "ResearchAssistant") then
        otherAddons.researchAssistantActive = true
        debugMessage("Addon Research Assistant is active", false)
    end
	--Check if addon "InventoryGridView" is active
    if(addOnName == "InventoryGridView") then
    	otherAddons.inventoryGridViewActive = true
    end
	--Check if addon "ChatMerchant" is active
    if(addOnName == "ChatMerchant") then
    	otherAddons.chatMerchantActive = true
    end
	--Is this addon found?
	if(addOnName ~= addonVars.gAddonName) then
        return
    end
	--Unregister this event again so it isn't fired again after this addon has beend reckognized
    EVENT_MANAGER:UnregisterForEvent(addonVars.gAddonName, EVENT_ADD_ON_LOADED)

    debugMessage("[Addon loading begins...]", true)
	addonVars.gAddonLoaded = false

    -- Standard filtering: Inside player inventory
    locVars.gFilterWhere = 1

    --The default values for the language and save mode
    local defaultsSettings = {
        language 	 		    = 1, --Standard: English
        saveMode     		    = 2, --Standard: Account wide settings
    }

    --Pre-set the deafult values
    local defaults = {
        --language 	 		    	= 1, --Standard: English
        --saveMode     		    	= 2, --Standard: Account wide settings
        languageChoosen				= false,
    	markedItems	 		    	= {},
        icon		 		    	= {},
        filterButtonLeft			= {},
        filterButtonTop				= {},
        splitFilters				= true,
        isFilterOn   		   	 	= {false, false, false, false},
        isFilterPanelOn				= {},
		atPanelEnabled				= {},
		isGearEnabled				= {true, true, true, true, true},
        allowResearch 		    	= false,
        allowInventoryFilter    	= true,
        allowVendorFilter	    	= true,
        allowGuildBankFilter    	= true,
        allowBankFilter 	    	= true,
		allowTradinghouseFilter		= true,
		allowTradeFilter	   	 	= true,
        allowMailFilter		    	= true,
        allowEnchantingFilter		= true,
        allowDeconstructionFilter	= true,
        allowImprovementFilter  	= true,
        blockDestroying				= true,
        blockDeconstruction			= true,
        blockImprovement			= true,
        blockSendingByMail			= true,
		blockTrading				= true,
        blockSelling				= true,
        allowSellingForBlockedOrnate = false,
		autoMarkOrnate 		    	= false,
        autoMarkResearch			= false,
        autoMarkAllEquipment		= false,
        autoMarkAllWeapon			= false,
        autoMarkAllJewelry			= false,
		showFilterStatusInChat  	= true,
        showOrnateItemsInChat   	= false,
        showResearchItemsInChat 	= false,
        showFCOISMenuBarButton		= true,
		showFCOISAdditionalInventoriesButton = true,
        showFilterButtonTooltip		= true,
        debug						= false,
        deepDebug					= false,
    }

    for helper_inv = 1, locVars.gFCONumFilterInventoryTypes, 1 do
		--Create 2-dimensional arrays for the filters
	    defaults.isFilterPanelOn[helper_inv] = {false, false, false, false}

	    --Create 2-dimensional array for the "enabled" setings (Filters, Anti-Destroy, Anti-Deconstruction, Anti-Sell, Anti-Trade)
	    defaults.atPanelEnabled[helper_inv]	= {
	        ["filters"] 		 = false,
		    ["anti-destroy"] 	 = false,
		    ["anti-deconstruct"] = false,
			["anti-improve"]	 = false,
			["anti-sell"]	 	 = false,
			["anti-sell-exception"] = false,
			["anti-trade"] 		 = false,
			["anti-mail"] 		 = false,
        }
    end

	--Create 2-dimensional arrays for the icons
	for helper = 1, locVars.gFCONumFilterIcons, 1 do
	    defaults.markedItems[helper] = {}
        defaults.icon[helper] 		 = {}
    end

    defaults.icon[1].color   = {["r"] = 1,["g"] = 0,["b"] = 0,["a"] = 1}
	defaults.icon[1].texture = 1
	defaults.icon[1].size    = locVars.gIconWidth
	defaults.icon[2].color   = {["r"] = 0,["g"] = 1,["b"] = 0,["a"] = 1}
	defaults.icon[2].texture = 2
	defaults.icon[2].size    = locVars.gIconWidth
	defaults.icon[3].color   = {["r"] = 1,["g"] = 1,["b"] = 1,["a"] = 1}
	defaults.icon[3].texture = 3
	defaults.icon[3].size    = locVars.gIconWidth
	defaults.icon[4].color   = {["r"] = 1,["g"] = 0,["b"] = 1,["a"] = 1}
	defaults.icon[4].texture = 2
	defaults.icon[4].size    = locVars.gIconWidth
	defaults.icon[5].color   = {["r"] = 1,["g"] = 1,["b"] = 0,["a"] = 1}
	defaults.icon[5].texture = 4
	defaults.icon[5].size    = locVars.gIconWidth
	defaults.icon[6].color   = {["r"] = 1,["g"] = 1,["b"] = 0,["a"] = 1}
	defaults.icon[6].texture = 2
	defaults.icon[6].size    = locVars.gIconWidth
	defaults.icon[7].color   = {["r"] = 1,["g"] = 1,["b"] = 0,["a"] = 1}
	defaults.icon[7].texture = 2
	defaults.icon[7].size    = locVars.gIconWidth
	defaults.icon[8].color   = {["r"] = 1,["g"] = 1,["b"] = 0,["a"] = 1}
	defaults.icon[8].texture = 2
	defaults.icon[8].size    = locVars.gIconWidth


	--Update the texts depending on locaization
	defaults.icon[2].name    = "Gear 1"
	defaults.icon[4].name    = "Gear 2"
	defaults.icon[6].name    = "Gear 3"
	defaults.icon[7].name    = "Gear 4"
	defaults.icon[8].name    = "Gear 5"

	--Defaults for filter button offsets
	for helperFilterButtons = 1, locVars.gFCONumFilterIcons, 1 do
		defaults.filterButtonTop[helperFilterButtons]  = locVars.gFilterButtonTop
		defaults.filterButtonLeft[helperFilterButtons] = locVars.gFilterButtonLeft[helperFilterButtons]
	end

--=============================================================================================================
--	LOAD USER SETTINGS
--=============================================================================================================
    --Load the user's settings from SavedVariables file -> Account wide of basic version 999 at first
	defaultSettings = ZO_SavedVars:NewAccountWide("FCOItemSaver_Settings", 999, "SettingsForAll", defaultsSettings)

	--Check, by help of basic version 999 settings, if the settings should be loaded for each character or account wide
    --Use the current addon version to read the settings now
	if (defaultSettings.saveMode == 1) then
    	settings = ZO_SavedVars:New("FCOItemSaver_Settings", addonVars.addonVersion , "Settings", defaults )
	elseif (defaultSettings.saveMode == 2) then
		settings = ZO_SavedVars:NewAccountWide("FCOItemSaver_Settings", addonVars.addonVersion, "Settings", defaults)
	else
		settings = ZO_SavedVars:NewAccountWide("FCOItemSaver_Settings", addonVars.addonVersion, "Settings", defaults)
	end
--=============================================================================================================

    --Preset global variable for item destroying
    preventerVars.gAllowDestroyItem = not settings.blockDestroying

    -- Get the marked items for each filter
	for markedItemsId = 1, locVars.gFCONumFilterIcons, 1 do
	    markedItems[markedItemsId] = settings.markedItems[markedItemsId]
    end

	-- Set Localization
    Localization()

	--Update the texts depending on localization
	defaults.icon[2].name    = fco_loc["options_icon2_name"]
	defaults.icon[4].name    = fco_loc["options_icon4_name"]
	defaults.icon[6].name    = fco_loc["options_icon6_name"]
	defaults.icon[7].name    = fco_loc["options_icon7_name"]
	defaults.icon[8].name    = fco_loc["options_icon8_name"]
    if (settings.icon[2].name == "") then
		settings.icon[2].name = defaults.icon[2].name
    end
    if (settings.icon[4].name == "") then
		settings.icon[4].name = defaults.icon[4].name
    end
    if (settings.icon[6].name == "") then
		settings.icon[6].name = defaults.icon[6].name
    end
    if (settings.icon[7].name == "") then
		settings.icon[7].name = defaults.icon[7].name
    end
    if (settings.icon[8].name == "") then
		settings.icon[8].name = defaults.icon[8].name
    end

	--Show the menu
	BuildAddonMenu()

    --Create the icon textures
    CreateTextures(-1)

	--Create the hooks
    CreateHooks()

    --Check inventory for ornate and/or researchable items and mark them to be sold, if activated in the settings
	scanInventory()

	--Build the inventory filter buttons and addd them to the panels
	updateFilterButtonsInInv(-1)

	--Initialize the filters
    EnableFilters(-100)

    -- Register slash commands
    RegisterSlashCommands()

	--Add the additional buttons, controlled by the settings
	AddAdditionalButtons(-1)

    debugMessage("[Addon loading finished. Have fun!]", true)
    addonVars.gAddonLoaded = true
end

-- Register the event "addon loaded" for this addon
local function FCOItemSaver_Initialized()

	EVENT_MANAGER:RegisterForEvent(addonVars.gAddonName, EVENT_ADD_ON_LOADED, FCOItemSaver_Loaded)
	--Register for the zone change/player ready event
	EVENT_MANAGER:RegisterForEvent(addonVars.gAddonName, EVENT_PLAYER_ACTIVATED, FCOItemSaver_Player_Activated)
	--Register for Crafting stations opened & closed
	EVENT_MANAGER:RegisterForEvent(addonVars.gAddonName, EVENT_CRAFTING_STATION_INTERACT, function(eventCode, craftSkill, sameStation) FCOItemSaver_Crafting_Interact(craftSkill, sameStation) end)
	EVENT_MANAGER:RegisterForEvent(addonVars.gAddonName, EVENT_END_CRAFTING_STATION_INTERACT, FCOItemSaver_End_Crafting_Interact)
    --Register for Store opened & closed
    EVENT_MANAGER:RegisterForEvent(addonVars.gAddonName, EVENT_OPEN_STORE, FCOItemSaver_Open_Store)
    EVENT_MANAGER:RegisterForEvent(addonVars.gAddonName, EVENT_CLOSE_STORE, FCOItemSaver_Close_Store)
	--Register for Trading house (guild store) opened & closed
	EVENT_MANAGER:RegisterForEvent(addonVars.gAddonName, EVENT_OPEN_TRADING_HOUSE, FCOItemSaver_Open_Trading_House)
	EVENT_MANAGER:RegisterForEvent(addonVars.gAddonName, EVENT_CLOSE_TRADING_HOUSE, FCOItemSaver_Close_Trading_House)
    --Register for Guild Bank opened & closed
	EVENT_MANAGER:RegisterForEvent(addonVars.gAddonName, EVENT_OPEN_GUILD_BANK, FCOItemSaver_Open_Guild_Bank)
	EVENT_MANAGER:RegisterForEvent(addonVars.gAddonName, EVENT_CLOSE_GUILD_BANK, FCOItemSaver_Close_Guild_Bank)
    --Register for Player's Bank opened & closed
	EVENT_MANAGER:RegisterForEvent(addonVars.gAddonName, EVENT_OPEN_BANK, FCOItemSaver_Open_Player_Bank)
	EVENT_MANAGER:RegisterForEvent(addonVars.gAddonName, EVENT_CLOSE_BANK, FCOItemSaver_Close_Player_Bank)
    --Register for Trade panel opened & closed
	EVENT_MANAGER:RegisterForEvent(addonVars.gAddonName, EVENT_TRADE_INVITE_ACCEPTED, FCOItemSaver_Open_Trade_Panel)
	EVENT_MANAGER:RegisterForEvent(addonVars.gAddonName, EVENT_TRADE_CANCELED, FCOItemSaver_Close_Trade_Panel)
	EVENT_MANAGER:RegisterForEvent(addonVars.gAddonName, EVENT_TRADE_SUCCEEDED, FCOItemSaver_Close_Trade_Panel)
	EVENT_MANAGER:RegisterForEvent(addonVars.gAddonName, EVENT_TRADE_FAILED, FCOItemSaver_Close_Trade_Panel)
    --Register for player inventory slot update
	EVENT_MANAGER:RegisterForEvent(addonVars.gAddonName, EVENT_INVENTORY_SINGLE_SLOT_UPDATE, FCOItemSaver_Inv_Update)
	--Events for destruction & destroy prevention
	EVENT_MANAGER:RegisterForEvent( addonVars.gAddonName, EVENT_INVENTORY_SLOT_LOCKED, FCOItemSaver_OnInventorySlotLocked )
	EVENT_MANAGER:RegisterForEvent( addonVars.gAddonName, EVENT_INVENTORY_SLOT_UNLOCKED, FCOItemSaver_OnInventorySlotUnLocked )
	EVENT_MANAGER:RegisterForEvent( addonVars.gAddonName, EVENT_MOUSE_REQUEST_DESTROY_ITEM, FCOItemSaver_OnMouseRequestDestroyItem )
	--Event if an action layer changes
	EVENT_MANAGER:RegisterForEvent(addonVars.gAddonName, EVENT_ACTION_LAYER_POPPED, FCOItemsaver_OnActionLayerPopped)
end

--==========================================================================================================================================
--==========================================================================================================================================
--==========================================================================================================================================
-- 															GLOBAL FUNCTIONS
--==========================================================================================================================================
--==========================================================================================================================================
--==========================================================================================================================================

--Global function to return boolean value, if an item is filtered
--The item's instance ID and the filter number are needed to check, if the item is marked. Filter can be 1, 2, 3, 4 or -1 for all
--Instance: gives the current item's instance ID, determined by function GetItemInstanceId(bagId, slotIndex)
--filterId: Specifies the filter for which the item is marked. The variable locVars.gFCONumFilters gives the maximum filter number.
--			Special: Filter 2 controls icons 2, 4, 6, 7 and 8 (Gear sets)
--					 Filter 4 controls icon 5
--filterPanelId: The panel where the filter is activated. Possible values are:
--(Attention: Only the ones marked with * are currntly supported!):
--INV_BAGS 				= 1 *
--INV_BANK 				= 2 *
--INV_GUILDBANK 		= 3 *
--INV_STORE 			= 4 *
--INV_DECONSTRUCTION  	= 5 *
--INV_GUILDSTORE 	 	= 6 *
--INV_MAIL 				= 7 *
--INV_TRADE 			= 8 *
--INV_ENCHANTING_EXTRACTION	= 9 *
--INV_IMPROVEMENT 		= 10 *
--INV_ENCHANTING_CREATION	= 11 *
function FCOIsFiltered(instance, filterId, filterPanelId)
	if (instance ~= nil and filterId ~= nil) then
        local offs  = 0
        --Workaround for empty filterPanelId
    	if (filterPanelId == nil or filterPanelId == '') then
            --Fallback solution: FilterPanelId will be deconstruction panel
        	filterPanelId = INV_DECONSTRUCTION
        end

	    local filterStatusVar = {}
		--Create 2-dimensional arrays for the filters
	    for h_inv = 1, locVars.gFCONumFilterInventoryTypes, 1 do
		    filterStatusVar[h_inv] = {false, false, false, false}
	    end

	    --Check all filters, silently (chat output: disabled)
	    filterStatusVar = filterStatus(-1, true, true)

		if (filterId ~= -1) then
			--Workaround for filterID 4 and icon 5
			if filterId == 4 then
				offs = 1
			else
				offs = 0
			end
			filterId = filterId + offs
			if (filterId == 2) then
				if (markedItems[2][SignItemId(instance)] == true or
					markedItems[4][SignItemId(instance)] == true or
					markedItems[6][SignItemId(instance)] == true or
					markedItems[7][SignItemId(instance)] == true or
					markedItems[8][SignItemId(instance)] == true    ) then
					--Split filtes, or old behaviour?
                    if (settings.splitFilters == true) then
						--Is the deconstruction filter activated?
	                    if (filterStatusVar[filterPanelId][2] == true) then
		                    return true
	                    end
                    else
	                    if (filterStatusVar[1][2] == true) then
		                    return true
	                    end
                   	end
				end
			else
				if (markedItems[filterId] ~= nil and markedItems[filterId][SignItemId(instance)]) then
					--Split filtes, or old behaviour?
                    if (settings.splitFilters == true) then
						--Is the deconstruction filter activated?
	                    if (filterStatusVar[filterPanelId][filterId-offs] == true) then
		                    return true
	                    end
                    else
	                    if (filterStatusVar[1][filterId-offs] == true) then
		                    return true
	                    end
                   	end
                end
			end
        else
			--Check for all filters if the item is marked. return true, if any filter applies
			for filtId = 1, locVars.gFCONumFilters, 1 do
				--Workaround for filterID 4 and icon 5
				if filtId == 4 then
					offs = 1
				else
					offs = 0
				end
				filtId = filtId + offs
            	if (filtId == 2) then
					if (markedItems[2][SignItemId(instance)] == true or
                    	markedItems[4][SignItemId(instance)] == true or
                    	markedItems[6][SignItemId(instance)] == true or
                    	markedItems[7][SignItemId(instance)] == true or
                    	markedItems[8][SignItemId(instance)] == true    ) then
						--Split filtes, or old behaviour?
	                    if (settings.splitFilters == true) then
							--Is the deconstruction filter activated?
		                    if (filterStatusVar[filterPanelId][2] == true) then
			                    return true
		                    end
	                    else
		                    if (filterStatusVar[1][2] == true) then
			                    return true
		                    end
	                   	end
	                end
                else
                	if (markedItems[filtId][SignItemId(instance)] == true) then
						--Split filtes, or old behaviour?
	                    if (settings.splitFilters == true) then
							--Is the deconstruction filter activated?
		                    if (filterStatusVar[filterPanelId][filtId-offs] == true) then
			                    return true
		                    end
	                    else
		                    if (filterStatusVar[1][filtId-offs] == true) then
			                    return true
		                    end
	                   	end
	                end
                end
			end
            return false
        end
    else
        --Current function parameters are missing. Return nil
    	return nil
    end
end

--Global function to return boolean value, if an item is marked
--The item's instance ID and the icon number are needed to check, if the item is marked. Icon can be 1, 2, 3, 4, 5, 6, 7, 8 or -1 for all
--Instance: gives the current item's instance ID, determined by function GetItemInstanceId(bagId, slotIndex)
--iconId: Specifies the icon with which the item is marked. The variable locVars.gFCONumFilterIcons gives the maximum icon number.
--			Special: Icons 2, 4, 6, 7, 8 belong to filter 2 (Gear set filter for 5 different gears)
--					 Icon 5 belongs to filter 4
function FCOIsMarked(instance, iconId)
	if (instance ~= nil and iconId ~= nil) then
		if (iconId ~= -1) then
			return markedItems[iconId] ~= nil and markedItems[iconId][SignItemId(instance)]
        else
			--Check for all icons if the item is marked. return true, if any icon is set
			for icoId = 1, locVars.gFCONumFilterIcons, 1 do
				if (markedItems[icoId][SignItemId(instance)] == true) then
                	return true
                end
			end
            return false
        end
    else
        --Current function parameters are missing. Return nil
    	return nil
    end
end

--------------------------------------------------------------------------------
--- Call the start function for this addon to register events etc.
--------------------------------------------------------------------------------
FCOItemSaver_Initialized()
